#include "Circle.h"
using namespace std;

static int solve (DOUBLE a, DOUBLE b,
		  DOUBLE cx, DOUBLE cy, DOUBLE cr,
		  DOUBLE *rx, DOUBLE *ry) {
  // y = a x + b
  // (x - X)^2 + (y - Y)^2 = r^2
  // (x - X)^2 + (a x + b - Y)^2 = r^2
  // (1 + a^2) x^2 + (-2 X + 2 a (b - Y)) x + X^2 + (b - Y)^2 - r^2

  DOUBLE b_Y = b - cy;
  DOUBLE A = 1 + a * a;
  DOUBLE B = -2 * cx + 2 * a * b_Y;
  DOUBLE C = cx * cx + b_Y * b_Y - cr * cr;

  DOUBLE rad = B * B - 4 * A * C;
  if (rad <= 0)
    return 0;
  
  DOUBLE srad = sqrt(rad);
  rx[0] = (-B - srad) / (2 * A);
  rx[1] = (-B + srad) / (2 * A);
  ry[0] = a * rx[0] + b;
  ry[1] = a * rx[1] + b;
  assert(rx[0] < 0 || rx[0] == 0 || rx[0] > 0);
  assert(ry[0] < 0 || ry[0] == 0 || ry[0] > 0);
  assert(rx[1] < 0 || rx[1] == 0 || rx[1] > 0);
  assert(ry[1] < 0 || ry[1] == 0 || ry[1] > 0);
  return 2;
}  

#ifdef BLEEN
static int solve (DOUBLE x1, DOUBLE y1, DOUBLE r1,
		  DOUBLE x2, DOUBLE y2, DOUBLE r2,
		  DOUBLE *rx, DOUBLE *ry) {
  // (x - X1)^2 + (y - Y1)^2 = R1^2
  // (x - X2)^2 + (y - Y2)^2 = R2^2
  // -2 (X1 - X2) x + X1^2 - X2^2 + -2 (Y1 - Y2) y + Y1^2 - Y2^2 = R1^2 - R2^2
  // -2 (X1 - X2) x - 2 (Y1 - Y2) y = R1^2 - R2^2 - X1^2 + X2^2 - Y1^2 + Y2^2
  DOUBLE ax = -2 * (x1 - x2);
  DOUBLE ay = -2 * (y1 - y2);
  DOUBLE c = (r1 * r1 - r2 * r2) - (x1 * x1 - x2 * x2) - (y1 * y1 - y2 * y2);

  DOUBLE a = -ax / ay;
  DOUBLE b = c / ay;

  return solve (a, b, x1, y1, r1, rx, ry);
}
#endif

static int solve (DOUBLE x1, DOUBLE y1, DOUBLE r1,
		  DOUBLE x2, DOUBLE y2, DOUBLE r2,
		  DOUBLE *rx, DOUBLE *ry) {
  if (r1 > r2)
    return solve(x2, y2, r2, x1, y1, r1, rx, ry);

  DOUBLE x12 = x2 - x1;
  DOUBLE y12 = y2 - y1;
  DOUBLE dd = x12 * x12 + y12 * y12;
  DOUBLE d = sqrt(dd);
  if (d >= r1 + r2)
    return 0;

  DOUBLE X1 = 0.5 * (d + (r1 - r2) * (r1 + r2) / d);
  if (X1 >= r1 || -X1 >= r1)
    return 0;

  DOUBLE Y = sqrt(r1 * r1 - X1 * X1);
  if (Y == 0)
    return 0;

  DOUBLE ux = x12 / d;
  DOUBLE uy = y12 / d;
  DOUBLE x1X1ux = x1 + X1 * ux;
  DOUBLE y1X1uy = y1 + X1 * uy;
  DOUBLE Yux = Y * ux;
  DOUBLE Yuy = Y * uy;

  rx[0] = x1X1ux + Yuy; // x1 + X1 * ux + Y * uy
  ry[0] = y1X1uy - Yux; // y1 + X1 * uy - Y * ux
  rx[1] = x1X1ux - Yuy; // x1 + X1 * ux - Y * uy
  ry[1] = y1X1uy + Yux; // y1 + X1 * uy + Y * ux;
  assert(rx[0] < 0 || rx[0] == 0 || rx[0] > 0);
  assert(ry[0] < 0 || ry[0] == 0 || ry[0] > 0);
  assert(rx[1] < 0 || rx[1] == 0 || rx[1] > 0);
  assert(ry[1] < 0 || ry[1] == 0 || ry[1] > 0);
  return 2;
}

static SetP<Circle> circleSet;

Circle *Circle::alloc (Point c, DOUBLE r) {
  if (c.x - r == c.x + r)
    cout << "this is it" << endl;
  Circle circle(c, r);
  Circle *p = circleSet.find(&circle);
  if (p == 0) {
    p = new Circle(c, r);
    assert(p == circleSet.insert(p));
  }
  return p;
}

void Circle::free () {
  if (&lower == (Function *) 0x570006 || &upper == (Function *) 0x570006)
    cout << "this is it" << endl;
  assert(this == circleSet.find(this));
  circleSet.remove();
  delete this;
}

int Circle::intersect (Line *that, Point roots[2]) {
  if (this->r == 0)
    return 0;
  Point u = that->getMid() - c;
  Point v = that->getDiff();
  double udv = u * v;
  double v2 = v * v;
  double t = -udv / v2;
  Point w = u + v * t;
  double w2 = w * w;
  double r2 = r * r;
  if (w2 >= r2)
    return 0;
  t = sqrt((r2 - w2) / v2);
  roots[0] = c + w - v * t;
  roots[1] = c + w + v * t;
  int n = 0;
  for (int i = 0; i < 2; i++) {
    that->pushToBox(roots[i]);
    if (that->inBBox(roots[i]))
      roots[n++] = roots[i];
  }

  static double maxCircle1;
  for (int i = 0; i < n; i++) {
    Point p = roots[i];
    Point q = that->nearestPoint(p);
    Point pq = q - p;
    double dist = sqrt(pq * pq);
    if (maxCircle1 < dist) {
      maxCircle1 = dist;
      cout << "maxCircle1 " << maxCircle1 << endl;
    }

    Point cp = p - c;
    double lcpl = sqrt(cp * cp);
    dist = fabs(lcpl - r);
    if (maxCircle1 < dist) {
      maxCircle1 = dist;
      cout << "maxCircle1 " << maxCircle1 << endl;
    }
  }



  return n;
}

int Circle::intersect (Circle *that, Point roots[2]) {
  if (this->r > that->r)
    return that->intersect(this, roots);
  if (this->r == 0)
    return 0;
  assert(this != that);

  Point c01 = that->c - this->c;
  double c012 = c01 * c01;
  double r02 = this->r * this->r;
  double r12 = that->r * that->r;
  Point w = c01 / 2 + c01 * ((r02 - r12) / (2 * c012));
  double w2 = w * w;
  Point v = c01.rot90();
  if (w2 >= r02)
    return 0;
  double t = sqrt((r02 - w2) / c012);
  roots[0] = c + w - v * t;
  roots[1] = c + w + v * t;

  static double maxCircle2;
  Circle *circs[2] = { this, that };
  for (int h = 0; h < 2; h++) {
    for (int i = 0; i < 2; i++) {
      Circle *circ = circs[h];
      Point cp = roots[i] - circ->c;
      double lcpl = sqrt(cp * cp);
      double dist = fabs(lcpl - circ->r);
      if (maxCircle2 < dist) {
	maxCircle2 = dist;
	cout << "maxCircle2 " << maxCircle2 << endl;
      }
    }
  }

  return 2;
}

Circle *Circle::transform (const Transformation &tr, bool ymonotone) {
  Point cT = tr(c);
  double rT = tr.getR().length() * r;
  Circle circle(cT, rT);
  Circle *p = circleSet.find(&circle);
  if (p == 0) {
    p = new Circle(cT, rT);
    assert(p == circleSet.insert(p));
  }
  return p;
}

DOUBLE Circle::getY (bool height, DOUBLE x) const {
  DOUBLE dx = x - c.x;
  DOUBLE dy2 = r * r - dx * dx;
  if (dy2 < 0)
    return c.y;
  else if (height)
    return c.y + sqrt(dy2);
  else
    return c.y - sqrt(dy2);
}

DOUBLE Circle::distance (Point p) const {
  return fabs((p - c).length() - r);
}

DOUBLE Circle::getDY (bool height, DOUBLE x) const {
  static double maxCDY;
  double cdy = 0;
  if (x < tail.x) {
    cdy = tail.x - x;
    x = tail.x;
  }
  else if (x > head.x) {
    cdy = x - head.x;
    x = head.x;
  }
  if (cdy > maxCDY) {
    maxCDY = cdy;
    std::cout << "maxCDY " << maxCDY << std::endl;
  }

  DOUBLE dx = x - c.x;
  DOUBLE r2mdx2 = r * r - dx * dx;
  DOUBLE rad = r2mdx2 < 0 ? 0 : sqrt(r2mdx2);
  DOUBLE ret;

  if (rad == 0)
    if (dx < 0)
      ret = Function::FINFINITY;
    else
      ret = -Function::FINFINITY;
  else
    return ret = dx / rad;
  
  if (height)
    return ret;
  else
    return -ret;
}

void Cbranch::free () {
  assert(this == &circle->lower || this == &circle->upper);
  Cbranch *other = this == &circle->lower ? &circle->upper : &circle->lower;
  if (other->getRef() == 0)
    circle->free();
}

DOUBLE Cbranch::getMinX () const {
  return circle->tail.x;
}

DOUBLE Cbranch::getMaxX () const {
  return circle->head.x;
}

DOUBLE Cbranch::getY (DOUBLE x) const {
  return circle->getY(height, x);
}

DOUBLE Cbranch::getDY (DOUBLE x) const {
  return circle->getDY(height, x);
}

DOUBLE Cbranch::distance (Point p) const {
  // return circle->distance(p);
  if (fabs(p.y - circle->c.y) < 1e-6 * circle->r ||
      (p.y > circle->c.y - circle->r * 1e-6) == height)
    return circle->distance(p);
  else
    return fabs(p.y - getY(p.x));
}

RootList *Cbranch::makeRootList (Function *that) {
  bool isLine = that->getType() == LINE;
  if (!isLine)
    assert(that->getType() == LOWER_CIRCLE || that->getType() == UPPER_CIRCLE);
  Line *thatL = isLine ? (Line *) that : 0;
  Circle *thisC = circle;
  Circle *thatC = !isLine ? ((Cbranch *) that)->circle : 0;
  
  vector<Point> roots;
  if (thisC != thatC) {
    Point allroots[2];
    int nroots = isLine ?
      circle->intersect(thatL, allroots) : circle->intersect(thatC, allroots);
    for (int i = 0; i < nroots; i++) {
      Point p = allroots[i];
      if ((!height && p.y < thisC->c.y || height && p.y > thisC->c.y) &&
	  (that->getType() == LINE ||
	   that->getType() == LOWER_CIRCLE && p.y < thatC->c.y ||
	   that->getType() == UPPER_CIRCLE && p.y > thatC->c.y))
	roots.push_back(p);
    }
  }

  if (this < that)
    return new RootList(this, that, roots);
  else {
    assert(isLine);
    return new RootList(that, this, roots);
  }
}

void Cbranch::transform (const Transformation &tr, vector<Segment> &out, bool ymonotone) {
  Circle *circleT = circle->transform(tr);
  Point cT = circleT->c;
  Point u = tr.getR() * circle->r, v;
  if (u.y != 0) {
    bool mid = ((u.y > 0) ^ height);
    v.x = (2 * mid - 1) * circle->r;
  }
  bool side = u.y == 0 ? height ^ (u.x < 0) : u.y < 0;

  if (u.y == 0)
    out.push_back(Segment(cT - u, cT + u, circleT->getBranch(side)));
  else {
    out.push_back(Segment(cT - u, cT + v, circleT->getBranch(side)));
    out.push_back(Segment(cT + v, cT + u, circleT->getBranch(!side)));
  }
}

Point Cbranch::nearestPoint (Point p) const {
  Point u = p - circle->c;
  Point q = circle->c + u * (circle->r / u.length());
  if ((q.y >= circle->c.y) == height)
    return q;

  Point qt = circle->getTail() - q;
  Point qh = circle->getHead() - q;
  if (qt * qt < qh * qh)
    return circle->getTail();
  else
    return circle->getHead();
}
