#include "Line.h"

class Circle;

class Cbranch : public Function {
public:
  Cbranch (Circle *circle, bool height)
    : circle(circle), height(height) {}

  void free ();
  
  int getType () const { return height ? UPPER_CIRCLE : LOWER_CIRCLE; }

  DOUBLE getMinX () const;

  DOUBLE getMaxX () const;

  DOUBLE getY (DOUBLE x) const;

  DOUBLE getDY (DOUBLE x) const;

  DOUBLE distance (Point p) const;

  RootList *makeRootList (Function *that);

  void transform (const Transformation &tr, vector<Segment> &out, bool ymonotone = false);

  void transform (Segment &seg, const Transformation &tr,
		  vector<Segment> &out, bool ymonotone = false) {
    assert(0);
  }

  Circle *transform (const Transformation &tr, bool ymonotone = false);

  Point nearestPoint (Point p) const;

  Circle *const circle;
  const bool height;
};

class Circle {
public:
  static Circle *alloc (Point c, DOUBLE r);
  void free ();
  ~Circle () {}

  bool lessThan (const Circle *that) const {
    if (this->c.x < that->c.x)
      return true;
    else if (this->c.x > that->c.x)
      return false;
    else if (this->c.y < that->c.y)
      return true;
    else if (this->c.y > that->c.y)
      return false;
    else
      return this->r < that->r;
  }

  Point getTail() { return tail; }
  Point getHead() { return head; }

  Cbranch *getBranch (bool height) { return height ? &upper : &lower; }
  DOUBLE getY (bool height, DOUBLE x) const;
  DOUBLE getDY (bool height, DOUBLE x) const;
  DOUBLE distance (Point p) const;

  int intersect (Line *that, Point roots[2]);
  int intersect (Circle *that, Point roots[2]);
  Circle *transform (const Transformation &tr, bool ymonotone = false);

  const Point c;
  const DOUBLE r;
  Cbranch lower, upper;

private:
  Circle (Point c, DOUBLE r) : c(c), r(r), lower(this, 0), upper(this, 1) {
    assert(r > 0);
    tail.y = head.y = c.y;
    tail.x = c.x - r;
    head.x = c.x + r;
  }

  friend class Cbranch;
  Point tail, head;
};
