char copyright2[] = "copyright (c) 2008 Victor J. Milenkovic, not for commercial use without permission";

#include "Embedding.h"
#include "Circle.h"
#include "curve.h"
#include <iostream>
#include <fstream>
#include <list>
#include <map>
using namespace std;

int timeArrangeTotal, timeEmbedTotal;

const double Embedding::TELOMERE_LENGTH = 1.0 / (1 << 12);
int Embedding::nEmbedding;
int Embedding::Edge::nEdge;
int Embedding::Edge::nextSSN;

Embedding::Vert Embedding::Vert::vert(Point(0, 0), true);

const double Embedding::BBOX = 5.0;
const double Embedding::MU = Embedding::BBOX / (1 << 26) / (1 << 26);

inline int degree2fit (int degree) {
  return (degree+1)*(degree+2)/2-1;
}

inline double random (double minX, double maxX) {
  return minX + (maxX - minX) * random() / RAND_MAX;
}

Polynomial fitPoints (int d, const vector<Point> &points);
bool fitPoints (int d, const vector<Point> &points, vector<Segment> &out);

Polynomial fitPoints (int d, const vector<Point> &points) {
  int n = degree2fit(d);
  assert(points.size() == n);
  double points2[2 * n];
  for (int i = 0; i < n; i++) {
    // assert(Embedding::inBBox(points[i]));
    points2[2 * i] = points[i].x;
    points2[2 * i + 1] = points[i].y;
  }
  Polynomial p;
  fit_curve(d, points2, p);
  return p;
}

bool fitPoints (int d, const vector<Point> &points, vector<Segment> &out) {
  static int counter;
  counter++;

  Polynomial p(fitPoints(d, points));
  vector<Segment> segs;
  double BBOX = Embedding::BBOX;
  double ibbox[] = { -BBOX, BBOX, -BBOX, BBOX };
  Acurve *c = new Acurve(p, ibbox);
  c->form_branches(false);
  c->segments(out, -BBOX - 1, -BBOX, BBOX, BBOX + 1);
  return true;
}

Abranch *makeImplicit (Line *line) {
  Point tail = line->getTail();
  Point head = line->getHead();
  Point diff = head - tail;
  if (diff.length() < 1e-3) {
    diff = diff.unit();
    tail = tail - diff;
    head = head + diff;
  }
  vector<Point> points;
  points.push_back(tail);
  points.push_back(head);
  Polynomial p(fitPoints(1, points));
  double BBOX = Embedding::BBOX;
  double ibbox[] = { -BBOX, BBOX, -BBOX, BBOX };
  Acurve *c = new Acurve(p, ibbox);
  c->form_branches(false);
  assert(c->branches.size() == 1);
  return c->branches[0];
}

void makeImplicit (vector<Segment> &segs) {
  for (int i = 0; i < segs.size(); i++) {
    Function *f = segs[i].getF();
    if (f == 0 || f->getType() != Function::LINE)
      continue;
    Segment &seg = segs[i];
    seg = Segment(seg.getTail(), seg.getHead(), makeImplicit((Line *) f));
  }
}      

int Embedding::SweepNode::compare (Treap *that_) {
  SweepNode *that = dynamic_cast<SweepNode *>(that_);
  if (this->edge->height >= 0 && that->edge->height >= 0 &&
      this->edge->id == that->edge->id)
    return this->edge->height - that->edge->height;
  if (edge->f != that->edge->f) {
    double x = max(edge->tail()->p.x, that->edge->tail()->p.x);
    return edge->f->compare(that->edge->f, x, 1);
  }
  if (edge->id != that->edge->id)
    return edge->id - that->edge->id;
  if (this->edge->height != that->edge->height)
    return this->edge->height - that->edge->height;
  assert(0);
  return 0;
}

Embedding::Vert *Embedding::addVert (Point p) {
  Vert key(p, true);
  Verti it = verts.find(&key);
  if (it != verts.end())
    return *it;

  Vert *vert = new Vert(p);
  verts.insert(vert);
  return vert;
}

void Embedding::Edge::dumpPolys (Function *f, Function *g) {
  cout << "-7 7 -7 7 0.1" << endl;
  cout << 2 << endl;
  cout << dynamic_cast<Abranch *>(f)->curve->f << endl;
  cout << dynamic_cast<Abranch *>(g)->curve->f << endl;
}

double Embedding::Edge::implicitDistance (Function *f, Point p) {
  return f->distance(p);

  Abranch *b = dynamic_cast<Abranch *>(f);
  Acurve *c = b->curve;

  if (b->tail->type == Split) {
    int lohi = b == b->tail->branches[1];
    assert(b == b->tail->branches[lohi]);
    Biftype tMinMax = lohi == 0 ? Min : Max;
    int iMinMax;
    for (iMinMax = 0; iMinMax < c->bifs.size(); iMinMax++) {
      Bifurcation *bMinMax = c->bifs[iMinMax];
      if (bMinMax->point.x > b->tail->point.x &&
	  bMinMax->point.x < b->head->point.x &&
	  bMinMax->type == tMinMax)
	break;
    }
    Bifurcation *bMinMax;
  }
}
    
Embedding::Edge *Embedding::addEdge (Point tail, Point head, Ptr<Function> f,
				     int id, int mult, int info) {
  Edge *edge = new Edge(addVert(tail), addVert(head), f, id, mult, info);
  if (f == 0)
    edgesV.push_back(edge);
  else {
    numAdded++;
    edges.push_back(edge);
    edge->height = -numAdded;
  }
  return edge;
}

void Embedding::checkMults () {
  return;
  map<Vert *, int> mults;

  for (int i = 0; i < edges.size(); i++) {
    Edge *edge = edges[i];
    mults[edge->tail()] += edge->mult;
    mults[edge->head()] -= edge->mult;
  }

  for (int i = 0; i < edgesV.size(); i++) {
    Edge *edge = edgesV[i];
    mults[edge->tail()] += edge->mult;
    mults[edge->head()] -= edge->mult;
  }
    
  for (Verti it = verts.begin(); it != verts.end(); ++it)
    assert(mults[*it] == 0);
}

void Embedding::arrange1d (vector<Embedding::Edge *> &edges) {
  vector<Edge *> in = edges;
  edges.clear();
  sort(in.begin(), in.end(), Edge::compareF);
  Verts vertsF;

  int j = 0;
  for (int i = 0; i < in.size(); i = j) {
    Ptr<Function> f = in[i]->f;

    vertsF.clear();

    if (in[i]->isVertical()) {
      double x = in[i]->tail()->p.x;
      Verti it = verts.find(in[i]->tail());
      assert(it != verts.end());
      for (; it != verts.begin() && (*it)->p.x == x; --it);
      if ((*it)->p.x != x)
	++it;
      for (; it != verts.end() && (*it)->p.x == x; ++it)
	vertsF.insert(*it);
    }

    for (j = i; j < in.size() && in[i]->equalF(in[j]); j++) {
      vertsF.insert(in[j]->tail());
      vertsF.insert(in[j]->head());
    }
    
    for (Verti it = vertsF.begin(); it != vertsF.end(); it++)
      (*it)->edge[0][0] = (*it)->edge[0][1] = 0;

    int jj = 0;
    for (int ii = i; ii < j; ii = jj) {
      int info = in[ii]->info;
      int id = in[ii]->id;
      for (jj = ii; jj < j && (ignoreIDs || in[jj]->id == id); jj++) {
	Edge *edge = in[jj];
	if (!ignoreIDs && !sumInfo)
	  if (edge->info != info)
	    /* cout << "this is it" << endl */;
	// assert(edge->info == info);
	for (int th = 0; th <= 1; th++) {
	  Vert *vert = edge->getVert(th);
	  edge->edge[0][th] = vert->edge[0][th];
	  vert->edge[0][th] = edge;
	}
      }

      int mult = 0;
      int sum_info = 0;
      Vert *prev = 0;
      for (Verti it = vertsF.begin(); it != vertsF.end(); it++) {
	Vert *vert = *it;
	if (mult != 0) {
	  if (f != 0 && prev->p.x == vert->p.x) {
	    cout << "arrange1d warning " << vert->p - prev->p << endl;
	    addEdge(prev->p, vert->p, 0, sumInfo ? 0 : id, mult, 0);
	  }
	  else
	    addEdge(prev->p, vert->p, f, id, mult, sumInfo ? sum_info : info);
	}
	prev = vert;
	for (Edge *edge = vert->edge[0][1]; edge != 0;
	     edge = edge->edge[0][1]) {
	  mult -= edge->mult;
	  if (sumInfo)
	    sum_info -= edge->info;
	}
	vert->edge[0][1] = 0;
	for (Edge *edge = vert->edge[0][0]; edge != 0;
	     edge = edge->edge[0][0]) {
	  mult += edge->mult;
	  if (sumInfo)
	    sum_info += edge->info;
	  if (ignoreIDs) {
	    id = edge->id;
	    info = edge->info;
	  }
	}
	vert->edge[0][0] = 0;
      }
    }
  }

  for (int i = 0; i < in.size(); i++)
    delete in[i];
}

class HLine {
public:
  double y;
  Acurve *curve;
  HLine *next;

  HLine (double y, Acurve *curve, HLine *next)
    : y(y), curve(curve), next(next) {}

  static HLine *hlines;

  static Abranch *getHLine (double y) {
    for (HLine *hline = hlines; hline != 0; hline = hline->next)
      if (hline->y == y)
	return hline->curve->branches[0];

    double a[3] = {-y, 0.0, 1.0};
    Polynomial p = linear_polynomial(2, a);
    double BBOX = Embedding::BBOX;
    double ibbox[] = { -2 * BBOX, 2 * BBOX, -2 * BBOX, 2 * BBOX };
    Acurve *curve = new Acurve(p, ibbox);
    curve->form_branches(false);
    assert(curve->branches.size() == 1);
    // assert(curve->f.isHorizontal());
    hlines = new HLine(y, curve, hlines);

    return curve->branches[0];
  }
};

HLine *HLine::hlines = 0;

bool isShallow (Function *f, double x) {
  return f->getType() == 2 || f->getType() == -2 || fabs(f->getDY(x)) < 1e6;
}

void Embedding::arrange (vector<Event *> &events) {
  if (verts.size() == 0)
    return;

  // arrange1d(edges);

  // create telomeres
  for (Verti it = verts.begin(); it != verts.end(); it++)
    (*it)->edge[1][0] = 0;

  set<Vert *, Vert::CompareYX> tails;
  for (int i = 0; i < edges.size(); i++)
    tails.insert(edges[i]->tail());

  vector<Edge *> telos;
  double minX = (*verts.begin())->p.x;
  double maxX = (*verts.rbegin())->p.x;
  Vert *prev = 0;

  for (set<Vert *, Vert::CompareYX>::iterator it = tails.begin();
       it != tails.end(); ++it) {
    Vert *head = *it;
    double x = head->p.x - TELOMERE_LENGTH, y = head->p.y;
    Vert *tail = 0;
    if (prev != 0 && prev->p.y == y && prev->p.x >= x)
      tail = prev;
    else
      tail = addVert(Point(x, y));
    prev = head;
    Function *line = 0;
    if (isImplicit())
      line = HLine::getHLine(y);
    else
      line = Line::alloc(Point(minX - TELOMERE_LENGTH, y), Point(maxX, y));
    Edge *telo = new Edge(tail, head, line, -1, 0);
    telos.push_back(telo);
    head->edge[1][0] = telo;
  }

  Treap *sweep = 0;
  Queue queue;
  int maxID = 0;

  for (int i = 0; i < edges.size(); i++) {
    Edge *edge = edges[i];
    assert(edge->tail()->edge[1][0] != 0);
    queue.push(new Event(Event::CONTINUE, edge->tail()->p, edge, 0, 0));
    queue.push(new Event(Event::REMOVE, edge->head()->p, edge, 0, 0));
    if (edge->id > maxID)
      maxID = edge->id;
  }

  Treap *sweepID[maxID+1];
  for (int i = 0; i <= maxID; i++)
    sweepID[i] = 0;

  for (int i = 0; i < telos.size(); i++) {
    Edge *edge = telos[i];
    if (edge->tail()->edge[1][0] == 0)
      queue.push(new Event(Event::INSERT, edge->tail()->p, edge, 0, 0));
    else
      queue.push(new Event(Event::CONTINUE, edge->tail()->p, edge, 0, 0));
    queue.push(new Event(Event::TREMOVE, edge->head()->p, edge, 0, 0));
  }

  while (!queue.empty()) {
    Event *event = queue.top();
    queue.pop();

    Edge *a = event->a;
    Edge *b = event->b;
    Edge *c = event->c;
    Point p = event->p;
    Event::Type type = event->type;

    if (type == Event::INSERT) {
      assert(c->id == -1);
      SweepNode *node = new SweepNode(c);
      if (sweep == 0)
	sweep = node;
      else {
	sweep->insert(node);
	sweep = node->rebalance();
	Edge *next = node->nextEdge();
	Edge *prev = node->prevEdge();
	if (next != 0)
	  queue.push(new Event(Event::CHECK, p, 0, c, next));
	if (prev != 0)
	  queue.push(new Event(Event::CHECK, p, 0, prev, c));

#ifdef BLEEN
	if (isShallow(c->f, p.x)) {
	  while (prev != 0 && isShallow(prev->f, p.x))
	    prev = prev->node->prevEdge();
	  assert(prev == 0 || prev->tail()->p.y < c->tail()->p.y + 1e-6);
	  while (next != 0 && isShallow(next->f, p.x))
	    next = next->node->nextEdge();
	  assert(next == 0 || next->tail()->p.y > c->tail()->p.y - 1e-6);
	}
#endif
      }

      delete event;
    }
    else if (type == Event::CONTINUE) {
      Edge *prevID = 0, *nextID = 0;
      if (c->height >= 0) {
	SweepNode *nodeID = new SweepNode(c);
	c->nodeID() = nodeID;
	c->node = 0;
	if (sweepID[c->id] == 0)
	  sweepID[c->id] = nodeID;
	else {
	  sweepID[c->id]->insert(nodeID);
	  sweepID[c->id] = nodeID->rebalance();
	}
	prevID = nodeID->prevEdge();
	nextID = nodeID->nextEdge();
      }

      SweepNode *node = new SweepNode(c);
      Edge *next = 0, *prev = 0;
      if (sweep == 0)
	sweep = node;
      else {
	Edge *telo = c->tail()->edge[1][0];
	next = telo->edge[1][0];
	prev = telo->edge[1][1];

	// next < prevID
	if (next != 0 && prevID != 0 &&
	    next->node->compareTree(prevID->node) <= 0) {
	  prev = prevID;
	  next = prevID->node->nextEdge();
	}
	// prev > nextID
	else if (prev != 0 && nextID != 0 &&
		 prev->node->compareTree(nextID->node) >= 0) {
	  next = nextID;
	  prev = nextID->node->prevEdge();
	}
	else {
	  // prevID > prev
	  if (prevID != 0 &&
	      (prev == 0 || prevID->node->compareTree(prev->node) > 0))
	    prev = prevID;
	  // nextID < next
	  if (nextID != 0 &&
	      (next == 0 || nextID->node->compareTree(next->node) < 0))
	    next = nextID;
	}

	if (next != 0)
	  next->node->setMax();
	if (prev != 0)
	  prev->node->setMin();

	sweep->insert(node);

	if (next != 0)
	  next->node->clearMinMax();
	if (prev != 0)
	  prev->node->clearMinMax();

	sweep = node->rebalance();

	next = node->nextEdge();
	prev = node->prevEdge();
	if (next != 0)
	  queue.push(new Event(Event::CHECK, p, 0, c, next));
	if (prev != 0)
	  queue.push(new Event(Event::CHECK, p, 0, prev, c));

#ifdef BLEEN
	Edge *savePrev = prev, *saveNext = next;
	if (isShallow(c->f, p.x)) {
	  while (prev != 0 && !isShallow(prev->f, p.x))
	    prev = prev->node->prevEdge();
	  assert(prev == 0 || prev->tail()->p.y < c->tail()->p.y + 1e-6);
	  while (next != 0 && !isShallow(next->f, p.x))
	    next = next->node->nextEdge();
	  assert(next == 0 || next->tail()->p.y > c->tail()->p.y - 1e-6);
	}
	prev = savePrev;
	next = saveNext;
#endif
      }

      if (c->id != -1) {
	while (prev != 0 && prev->id == -1)
	  prev = prev->node->prevEdge();
	while (next != 0 && next->id == -1)
	  next = next->node->nextEdge();
	event->a = prev;
	event->b = next;
	event->type = Event::INSERT;
	events.push_back(event);
      }
      else
	delete event;
    }
    else if (type == Event::REMOVE || type == Event::TREMOVE) {
      SweepNode *node = c->node;
      Edge *prev = node->prevEdge();
      Edge *next =  node->nextEdge();
      if (prev != 0 && next != 0) {
#ifdef BLEEN
	// debug
	double prevY = prev->f->getY(p.x);
	double nextY = next->f->getY(p.x);
	double cY = c->f->getY(p.x);
#endif
	queue.push(new Event(Event::CHECK, p, 0, prev, next));
      }
#ifdef BLEEN
      while (prev != 0 && prev->id == -1)
	prev = prev->node->prevEdge();
      while (next != 0 && next->id == -1)
	next = next->node->nextEdge();
#endif
      c->edge[1][0] = next;
      c->edge[1][1] = prev;
      sweep = node->remove();
      delete node;

      if (c->id != -1) {
	event->a = prev;
	event->b = next;
	events.push_back(event);
      }
      else
	delete event;

      if (c->height >= 0) {
	c->node = c->nodeID();
	c->nodeID() = 0;
	sweepID[c->id] = c->node->remove();
	delete c->node;
      }
    }
    else if (type == Event::SWAP) {
      if (a->node->nextEdge() != b) {
	delete event;
	continue;
      }
      if (p.x != a->f->nearestRoot(b->f, p.x))
	p.y = a->bestY(b, p.x);
      event->p = p;
      SweepNode *nodeA = a->node;
      SweepNode *nodeB = b->node;
      a->node = nodeB;
      b->node = nodeA;
      nodeA->edge = b;
      nodeB->edge = a;
      Edge *prev = nodeA->prevEdge();
      Edge *next = nodeB->nextEdge();
      queue.push(new Event(Event::CHECK, p, 0, b, a));
      if (prev != 0)
	queue.push(new Event(Event::CHECK, p, 0, prev, b));
      if (next != 0)
	queue.push(new Event(Event::CHECK, p, 0, a, next));

      if (a->id != -1 && b->id != -1)
	events.push_back(event);
      else
	delete event;
    }
    else if (type == Event::CHECK) {
      if (a->id >= 0 && a->height >= 0 &&
	  b->id >= 0 && b->height >= 0 &&
	  a->id == b->id) {
	assert(a->height < b->height);
	delete event;
	continue;
      }

      if (a->node == 0 || a->node->nextEdge() != b) {
	delete event;
	continue;
      }

#ifdef BLEEN
      // debug
      double maxX = max(a->tail()->p.x, b->tail()->p.x);
      double ay = a->f->getY(maxX);
      double by = b->f->getY(maxX);
#endif

      p = a->nextRoot(b, p.x);
      if (p.x != Function::FINFINITY)
	queue.push(new Event(Event::SWAP, p, 0, a, b));
      
      delete event;
    }
  }

  for (int i = 0; i < telos.size(); i++)
    delete telos[i];
}

void Embedding::splitEdges (vector<Event *> &events) {
  vector<Edge *> in = edges;
  edges.clear();

  for (int i = 0; i < in.size(); i++) {
    Edge *edge = in[i];
    edge->subEdge() = 0;
    edge->nextSweep() = 0;
    edge->prevSweep() = 0;
    edge->subEdge() = addEdge(edge->tail()->p, edge->head()->p, edge->f,
			      edge->id, edge->mult, edge->info);
  }    

  int j;
  for (int i = 0; i < events.size(); i = j) {
    for (j = i + 1; j < events.size() && events[i]->p.x == events[j]->p.x;
	 j++);

    // Set above/below head for subedge of edges removed (incoming)
    // or swapped and through.
    for (int ii = i; ii < j; ii++) {
      Event *event = events[ii];
      if (event->type == Event::REMOVE) {
	Edge *c = event->c;
	c->setEdgeAboveBelowHead();
      }
      else if (event->type == Event::SWAP) {
	Point p = event->p;
	Edge *a = event->a;
	Edge *b = event->b;
	assert(a->subEdge()->head()->p.x > p.x);
	// Make sure it is through.
	if (a->subEdge()->tail()->p.x < p.x)
	  a->setEdgeAboveBelowHead();
	assert(b->subEdge()->head()->p.x > p.x);
	// Make sure it is through.
	if (b->subEdge()->tail()->p.x < p.x)
	  b->setEdgeAboveBelowHead();
      }
    }

    // process events
    for (int ii = i; ii < j; ii++) {
      Event *event = events[ii];
      if (event->type == Event::REMOVE) {
	Edge *c = event->c;
	if (c->nextSweep() != 0) {
	  assert(c->nextSweep()->prevSweep() == c);
	  c->nextSweep()->prevSweep() = c->prevSweep();
	}
	if (c->prevSweep() != 0) {
	  assert(c->prevSweep()->nextSweep() == c);
	  c->prevSweep()->nextSweep() = c->nextSweep();
	}
	c->nextSweep() == c->prevSweep() == 0;
      }
      else if (event->type == Event::INSERT) {
	Edge *a = event->a;
	Edge *b = event->b;
	Edge *c = event->c;
	c->prevSweep() = a;
	if (a != 0) {
	  assert(a->nextSweep() == b);
	  a->nextSweep() = c;
	}
	c->nextSweep() = b;
	if (b != 0) {
	  assert(b->prevSweep() == a);
	  b->prevSweep() = c;
	}
      }
      else if (event->type == Event::SWAP) {
	Edge *a = event->a;
	Edge *b = event->b;
	Point p = event->p;
	Vert *v = addVert(p);
	assert(a->nextSweep() == b && b->prevSweep() == a);
	if (a->prevSweep() != 0)
	  a->prevSweep()->nextSweep() = b;
	if (b->nextSweep() != 0)
	  b->nextSweep()->prevSweep() = a;
	a->nextSweep() = b->nextSweep();
	b->prevSweep() = a->prevSweep();
	a->prevSweep() = b;
	b->nextSweep() = a;

	Edge *&aSub = a->subEdge();
	if (aSub->tail() != v) {
	  aSub->setHead(v);
	  aSub = addEdge(p, a->head()->p, a->f, a->id, a->mult, a->info);
	}

	Edge *&bSub = b->subEdge();
	if (bSub->tail() != v) {
	  bSub->setHead(v);
	  bSub = addEdge(p, b->head()->p, b->f, b->id, b->mult, b->info);
	}
      }
    }

    // Set above/below tail for subedge of edges inserted or swapped
    for (int ii = i; ii < j; ii++) {
      Event *event = events[ii];
      if (event->type == Event::INSERT)
	event->c->setEdgeAboveBelowTail();
      else if (event->type == Event::SWAP) {
	assert(event->a->head()->p.x > event->p.x);
	event->a->setEdgeAboveBelowTail();
	assert(event->b->head()->p.x > event->p.x);
	event->b->setEdgeAboveBelowTail();
      }
    }
  }

  for (int i = 0; i < in.size(); i++)
    delete in[i];

  int ne = 0;
  for (int i = 0; i < edges.size(); i++) {
    Edge *edge = edges[i];
    if (edge->tail()->p.x == edge->head()->p.x) {
      assert(edge->tail()->p.y != edge->head()->p.y);
      if (edge->tail()->p.y > edge->head()->p.y) {
	Vert *temp = edge->tail();
	edge->setTail(edge->head());
	edge->setHead(temp);
	edge->mult = -edge->mult;
      }
      edge->f = 0;
      edgesV.push_back(edge);
    }
    else
      edges[ne++] = edge;
  }
  edges.resize(ne);
}

void setOrder (int nEdges, int *outLimit, int *outEdges, int iEdge,
	       int *order, int &n) {
  int lo = iEdge == 0 ? 0 : outLimit[iEdge-1];
  int hi = outLimit[iEdge];

  for (int i = lo; i < hi; i++) {
    int jEdge = outEdges[i];
    if (order[jEdge] == -1)
      setOrder(nEdges, outLimit, outEdges, jEdge, order, n);
  }

  order[iEdge] = --n;
}

void Embedding::setHeights () {
  int nEdges = edges.size();

  // edges[i]->height will temporarily store index of edge
  // outDegree[i] is number of edges directly above edges[i]
  int outDegree[nEdges];
  for (int i = 0; i < nEdges; i++) {
    edges[i]->height = i;
    outDegree[i] = 0;
  }
  
  for (int i = 0; i < nEdges; i++) {
    Edge *edge = edges[i];
    if (edge->aboveTail() != 0)
      outDegree[i]++;
    if (edge->aboveHead() != 0)
      outDegree[i]++;
    if (edge->belowTail() != 0)
      outDegree[edge->belowTail()->height]++;
    if (edge->belowHead() != 0)
      outDegree[edge->belowHead()->height]++;
  }

  int outLimit[nEdges];
  outLimit[0] = 0;
  for (int i = 1; i < nEdges; i++)
    outLimit[i] = outLimit[i-1] + outDegree[i-1];

  int totalDegree = outLimit[nEdges-1] + outDegree[nEdges-1];
  int outEdges[totalDegree];
  for (int i = 0; i < nEdges; i++) {
    Edge *edge = edges[i];
    if (edge->aboveTail() != 0)
      outEdges[outLimit[i]++] = edge->aboveTail()->height;
    if (edge->aboveHead() != 0)
      outEdges[outLimit[i]++] = edge->aboveHead()->height;
    if (edge->belowTail() != 0)
      outEdges[outLimit[edge->belowTail()->height]++] = edge->height;
    if (edge->belowHead() != 0)
      outEdges[outLimit[edge->belowHead()->height]++] = edge->height;
  }
  
  int order[nEdges];
  for (int i = 0; i < nEdges; i++)
    order[i] = -1;

  int n = nEdges;
  for (int i = 0; i < nEdges; i++)
    if (order[i] == -1)
      setOrder(nEdges, outLimit, outEdges, i, order, n);

  Edge *sorted[nEdges];
  for (int i = 0; i < nEdges; i++)
    sorted[order[i]] = edges[i];
  for (int i = 0; i < nEdges; i++)
    (edges[i] = sorted[i])->height = 0;

  for (int i = 0; i < nEdges; i++) {
    Edge *edge = edges[i];
    if (edge->belowTail() != 0)
      edge->height = max(edge->height, edge->belowTail()->height + 1);
    if (edge->belowHead() != 0)
      edge->height = max(edge->height, edge->belowHead()->height + 1);
    if (edge->aboveTail() != 0)
      edge->aboveTail()->height =
	max(edge->height + 1, edge->aboveTail()->height);
    if (edge->aboveHead() != 0)
      edge->aboveHead()->height =
	max(edge->height + 1, edge->aboveHead()->height);
  }
}

void Embedding::splitBetween (Edge *edges[2], SweepNode *sweepTHS[3],
			      vector<Edge *> edgesTHV[3]) {
  if (sweepTHS[2] == 0)
    return;

  double x;
  if (!edgesTHV[0].empty())
    x = edgesTHV[0][0]->tail()->p.x;
  else if (!edgesTHV[1].empty())
    x = edgesTHV[1][0]->head()->p.x;
  else
    assert(0);

  Edge *prev[2] = { 0, 0 };
  Edge *next[2] = { 0, 0 };

  for (int du = 0; du <= 1; du++)
    sweepTHS[2]->locate(edges[du], prev[du], next[du]);

  for (Edge *edge = next[0]; edge != next[1];) {
    Edge *edge0 = edge;
    edge = edge->node->nextEdge();
    sweepTHS[2] = edge0->node->remove();

    Vert *vert = addVert(Point(x, edge0->getY(x)));
    Edge *edge1 = addEdge(edge0->tail()->p, vert->p, edge0->f,
			  edge0->id, edge0->mult, edge0->info);
    edge1->height = edge0->height;
    edge0->setTail(vert);
    edge1->node = new HeightNode(edge1);
    edgesTHV[0].push_back(edge0);
    edgesTHV[1].push_back(edge1);
  }
}

void Embedding::pinch (Edge *edge, int du, int th,
		       SweepNode *sweepTHS[3], vector<Edge *> edgesTHV[3]) {
  if (sweepTHS[2] == 0)
    return;
  if (sweepTHS[th] == 0)
    return;
  Edge *through[2];
  sweepTHS[2]->locate(edge, through[0], through[1]);
  if (through[du] == 0)
    return;
  Edge *incident[2];
  sweepTHS[th]->locate(through[du], incident[0], incident[1]);
  if (incident[du] == 0)
    return;
  Vert *vert = edge->tail();
  if (vert->p.x != incident[du]->getVert(th)->p.x)
    vert = edge->head();
  int sgn = 2 * du - 1;
  if (sgn * vert->p.y < sgn * incident[du]->getVert(th)->p.y)
    return;

  Edge *edges[2];
  if (du == 0) {
    edges[0] = incident[0];
    edges[1] = edge;
  }
  else {
    edges[0] = edge;
    edges[1] = incident[1];
  }
  splitBetween(edges, sweepTHS, edgesTHV);
}  

#define USE_DY
//#define USE_DYD

void Embedding::embed (Edge *edge, int th, int du) {
  static int count = 0;
  count++;
  // if (count == 78061)
  // cout << "this is it" << endl;

  Vert *vert = edge->getVert(th);
  Point p = vert->p;
  double dyd = 1;
#ifdef USE_DY
#ifdef USE_DYD
  double dy = fabs(edge->getDY(vert->p.x, dyd));
#else
  double dy = fabs(edge->getDY(vert->p.x));
#endif
#endif
  int sgn = 2 * du - 1;
  Edge *begin = edge->node->stepEdge(du);
  Edge *end;
  for (end = begin; end != 0; end = end->node->stepEdge(du)) {
    Point q = end->getVert(th)->p;
    if (sgn * p.y <= sgn * q.y)
      break;
#ifdef USE_DY
#ifdef USE_DYD
    double dy2 = fabs(end->getDY(p.x, dyd));
    if (dy <= dy2) {
#else
    double dy2 = fabs(end->getDY(p.x));
    if (dy >= dy2) {
#endif
      begin = edge;
      vert = end->getVert(th);
      break;
    }
#ifdef BLEEN
    }
#endif
#else
    double edgePtoEnd = end->f->distance(p);
    double endPtoEdge = edge->f->distance(end->getVert(th)->p);
    if (edgePtoEnd > endPtoEdge) {
      begin = edge;
      vert = end->getVert(th);
      break;
    }
#endif
  }

  // cout << setprecision(20);
  bool violation = false;
  for (Edge *e = begin; e != end; e = e->node->stepEdge(du)) {
    if (e->getVert(th) == vert)
      continue;
    violation = true;
    addEdge(vert->p, e->getVert(th)->p, 0,
	    e->id, (2 * th - 1) * e->mult, e->info);
    // cout << "ev " << vert->p.x << " " << e->getVert(th)->p.y << " "
    //	 << vert->p.y << endl;
    e->setVert(th, vert);
  }

  if (violation) {
    // cout << "embedding violation" << endl;
    Timing::embeddingViolations++;
  }
}

void Embedding::buildLists (bool doArrange1D) {
  if (doArrange1D)
    arrange1d(edgesV);

  // Store all edges in linked lists attached to incident vertices.
  for (Verti iv = verts.begin(); iv != verts.end(); ++iv) {
    Vert *vert = *iv;
    vert->edge[0][0] = vert->edge[0][1] =
      vert->edge[1][0] = vert->edge[1][1] =
      vert->edge[2][0] = vert->edge[2][1] = 0;
  }
  for (int i = 0; i < edges.size(); i++) {
    Edge *edge = edges[i];
    assert(edge->tail()->p.x < edge->head()->p.x);
    for (int th = 0; th <= 1; th++) {
      assert(verts.find(edge->getVert(th)) != verts.end());
      edge->edge[th][0] = edge->getVert(th)->edge[th][0];
      edge->getVert(th)->edge[th][0] = edge;
    }
  }

  for (int i = 0; i < edgesV.size(); i++) {
    Edge *edge = edgesV[i];
    Edge *&edge0 = edge->tail()->edge[2][0];
    Edge *&edge1 = edge->head()->edge[2][1];
    if (edge0 == 0) {
      assert(edge1 == 0);
      edge0 = edge;
      edge1 = edge;
      edge->edge[0][0] = 0;
      edge->edge[1][0] = 0;
      edge->edge[0][1] = 0;
      edge->edge[1][1] = 0;
    }
    else {
      assert(edge1 != 0);
      edge->edge[0][0] = 0;
      edge->edge[1][0] = 0;
      edge->edge[0][1] = edge0;
      edge->edge[1][1] = edge0;
      edge0->edge[0][0] = edge;
      edge0->edge[1][0] = edge;
      edge0 = edge;
    }
  }

  for (Verti it = verts.begin(); it != verts.end();) {
    Vert *vert = *it;
    Verti jt = it++;

    if (vert->edge[0][0] == 0 && vert->edge[1][0] == 0)
      assert((vert->edge[2][0] == 0) == (vert->edge[2][1] == 0));

    if (vert->edge[0][0] || vert->edge[1][0] ||
	vert->edge[2][0] || vert->edge[2][1])
      continue;
    verts.erase(jt);
    delete vert;
  }

#ifdef MERGE_VERTICALS
  map<int, Edge *> id2edge;
  for (Verti it = verts.begin(); it != verts.end();) {
    Verti jt = it++;
    Vert *vert = *jt;
    if (vert->edge[0][0] || vert->edge[1][0])
      continue;
    verts.erase(jt);
    if (vert->edge[2][0] == 0 || vert->edge[2][1] == 0) {
      assert(vert->edge[2][0] == 0 && vert->edge[2][1] == 0);
      delete vert;
      continue;
    }

    Vert *vert0 = vert->edge[2][1]->tail();
    id2edge.clear();
    for (Edge *edge = vert->edge[2][0]; edge != 0;
	 edge = edge->edge[0][1]) {
      id2edge[edge->id] = edge;
      edge->tail() = vert0;
    }
    vert0->edge[2][0] = vert->edge[2][0];

    for (Edge *edge = vert->edge[2][1]; edge != 0;
	 edge = edge->edge[1][0]) {
      Edge *edge2 = id2edge[edge->id];
      assert(edge2 != 0);
      assert(edge2->mult == edge->mult);
      assert(edge2->info == edge->info);
      edge2->head() = 0;
    }
    delete vert;
  }

  int nv = 0;
  for (int i = 0; i < edgesV.size(); i++)
    if (edgesV[i]->head() == 0)
      delete edgesV[i];
    else
      edgesV[nv++] = edgesV[i];
  edgesV.resize(nv);
#endif


  if (!ignoreIDs)
    return;

  // Merge vertical line segments.
  for (Verti it = verts.begin(); it != verts.end();) {
    Verti saveIT = it;
    Vert *vert = *it++;
    if (vert->edge[0][0] != 0 || vert->edge[1][0] != 0)
      continue;

    Edge *edge[2] = { vert->edge[2][0], vert->edge[2][1] };
    if (edge[0] == 0 && edge[1] == 0) {
      verts.erase(saveIT);
      delete vert;
      continue;
    }
      
    assert(edge[0] != 0 && edge[1] != 0);
    assert(edge[0]->edge[0][0] == 0 && edge[0]->edge[0][1] == 0);
    assert(edge[1]->edge[0][0] == 0 && edge[1]->edge[0][1] == 0);
    assert(edge[0]->tail()->p.x == edge[0]->head()->p.x);
    assert(edge[1]->tail()->p.x == edge[1]->head()->p.x);

    int i = (edge[0]->head()->p.y - edge[0]->tail()->p.y <
	     edge[1]->head()->p.y - edge[1]->tail()->p.y);

    assert(edge[0]->getVert(0) == vert && edge[1]->getVert(1) == vert);
    edge[i]->id = edge[!i]->id;
    edge[i]->info = edge[!i]->info;
    Vert *verti = edge[!i]->getVert(i);
    edge[i]->setVert(i, verti);
    assert(edge[i]->tail()->p.y < edge[i]->head()->p.y);
    verti->edge[2][i] = edge[i];
    verts.erase(saveIT);
    edge[!i]->setVert(!i, verti);
    delete vert;
  }

  int ne = 0;
  for (int i = 0; i < edgesV.size(); i++) {
    Edge *edge = edgesV[i];
    if (edge->tail() == edge->head())
      delete edge;
    else
      edgesV[ne++] = edge;
  }
  edgesV.resize(ne);
}

void Embedding::embed () {
  Timing::begin(timeEmbedTotal);

  buildLists();

  // Sweep lists of outgoing, incoming, and ``through'' (sweep) edges.
  SweepNode *sweepTHS[3] = { 0, 0, 0 };

  // Vector of outgoing, incoming, and vertical edges.
  vector<Edge *> edgesTHV[3];

  for (Verti iv = verts.begin(); iv != verts.end(); iv++) {
    Vert *vert = *iv;
    assert(vert->edge[0][0] || vert->edge[1][0] ||
	   vert->edge[2][0] || vert->edge[2][1]);
  }

  Verti jv;
  for (Verti iv = verts.begin(); iv != verts.end(); iv = jv) {
    // Determine range of vertices with same x.
    double x = (*iv)->p.x;
    for (jv = iv; jv != verts.end() && (*jv)->p.x == x; ++jv);
    
    // Get incoming, outgoing, and vertical edges for this x.
    for (Verti iiv = iv; iiv != jv; ++iiv) {
      Vert *vert = *iiv;
      for (int thv = 0; thv < 3; thv++) {
	int th0 = thv == 2 ? 0 : thv;
	for (Edge *edge = vert->edge[thv][0]; edge != 0;
	     edge = edge->edge[th0][thv == 2])
	  edgesTHV[thv].push_back(edge);
      }
    }

    // Remove incoming edges from sweep list.
    for (int i = 0; i < edgesTHV[1].size(); i++) {
      Edge *edge = edgesTHV[1][i];
      sweepTHS[2] = edge->node->remove();
    }

    // Give nodes to outgoing edges, but don't insert them yet.
    for (int i = 0; i < edgesTHV[0].size(); i++)
      new HeightNode(edgesTHV[0][i]);

    // Vertical edges split sweep edges.
#ifdef OLD_SPLIT
    for (int i = 0; i < edgesTHV[2].size(); i++) {
      Edge *edgeV = edgesTHV[2][i];

      // Find a non-vertical edge incident on each endpoint.
      Edge *edgesN[2] = { 0, 0 };
      for (int du = 0; du <= 1; du++) {
	Vert *vert = edgeV->getVert(du);
	for (int th = 0; th <= 1; th++)
	  if (vert->edge[th][0] != 0)
	    edgesN[du] = vert->edge[th][0];
	assert(edgesN[du] != 0);
      }

      // Cut all through edges between these non-vertical edges.
      if (edgesN[0]->height > edgesN[1]->height) {
	Edge *temp = edgesN[0];
	edgesN[0] = edgesN[1];
	edgesN[1] = temp;
      }
      splitBetween(edgesN, sweepTHS, edgesTHV);
    }
#else
    Verti jjv;
    for (Verti iiv = iv; iiv != jv; iiv = jjv) {
      Vert *high = *iiv;
      Edge *edgesN[] = { 0, 0 };
      for (jjv = iiv; jjv != jv && (*jjv)->p.y <= high->p.y; ++jjv) {
	Vert *vert = *jjv;
	for (Edge *edge = vert->edge[2][0]; edge != 0; edge = edge->edge[1][1])
	  if (edge->head()->p.y > high->p.y)
	    high = edge->head();
	for (int th = 0; th <= 1; th++)
	  for (Edge *edge = vert->edge[th][0]; edge != 0;
	       edge = edge->edge[th][0]) {
	    if (edgesN[0] == 0 || edge->height < edgesN[0]->height)
	      edgesN[0] = edge;
	    if (edgesN[1] == 0 || edge->height > edgesN[1]->height)
	      edgesN[1] = edge;
	  }
      }
      if (edgesN[0] != 0 && edgesN[0]->height < edgesN[1]->height)
	splitBetween(edgesN, sweepTHS, edgesTHV);
    }
#endif

    // Embed all incoming and outgoing edges.
    for (int iTH[2] = { 0, 0 };
	 iTH[0] < edgesTHV[0].size() || iTH[1] < edgesTHV[1].size();) {
      int th = iTH[0] < edgesTHV[0].size() ? 0 : 1;
      Edge *edge = edgesTHV[th][iTH[th]++];
      SweepNode *node = edge->node;
      
      if (sweepTHS[th] == 0)
	sweepTHS[th] = node;
      else {
	sweepTHS[th]->insert(node);
	sweepTHS[th] = node->rebalance();
      }

      Timing::embeddedEndpoints++;

      embed(edge, th, 0);
      embed(edge, th, 1);
      pinch(edge, 0, 0, sweepTHS, edgesTHV);
      pinch(edge, 0, 1, sweepTHS, edgesTHV);
      pinch(edge, 1, 0, sweepTHS, edgesTHV);
      pinch(edge, 1, 1, sweepTHS, edgesTHV);
    }

    // Take incoming out of incoming list.
    for (int i = 0; i < edgesTHV[1].size(); i++) {
      SweepNode *node = edgesTHV[1][i]->node;
      sweepTHS[1] = node->remove();
      delete node;
    }

    // Put outgoing edges into the sweep list.
    for (int i = 0; i < edgesTHV[0].size(); i++) {
      SweepNode *node = edgesTHV[0][i]->node;
      sweepTHS[0] = node->remove();
      if (sweepTHS[2]) {
	sweepTHS[2]->insert(node);
	sweepTHS[2] = node->rebalance();
      }
      else
	sweepTHS[2] = node;
    }

    assert(sweepTHS[0] == 0);
    assert(sweepTHS[1] == 0);
    for (int thv = 0; thv < 3; thv++)
      edgesTHV[thv].clear();
  }

  arrange1d(edgesV);

  Timing::end(timeEmbedTotal);
}

int Embedding::upfind (int up) {
  if (up == -1 || loops[up].cell == up)
    return up;
  return loops[up].cell = upfind(loops[up].cell);
}

void Embedding::build (bool dontArrange1D) {
  buildLists(!dontArrange1D);

  SweepNode *sweep = 0;

  // Vectors of outgoing and incoming edges.
  vector<Edge *> edgesTH[2];

  Verti jv;
  for (Verti iv = verts.begin(); iv != verts.end(); iv = jv) {
    // Determine range of vertices with same x.
    double x = (*iv)->p.x;
    for (jv = iv; jv != verts.end() && (*jv)->p.x == x; ++jv);
    
    // Get incoming and outgoing edges for this x.
    for (Verti iiv = iv; iiv != jv; ++iiv) {
      Vert *vert = *iiv;
      for (int th = 0; th <= 1; th++) {
	for (Edge *edge = vert->edge[th][0]; edge != 0;
	     edge = edge->edge[th][0])
	  edgesTH[th].push_back(edge);
	vert->edge[th][0] = vert->edge[th][1] = 0;
      }
    }

    // Set aboveHead and belowHead for incoming edges.
    for (int i = 0; i < edgesTH[1].size(); i++) {
      Edge *edge = edgesTH[1][i];
      Vert *vert = edge->head();
      Edge *&edge0 = edge->head()->edge[1][0];
      Edge *&edge1 = edge->head()->edge[1][1];
      if (edge0 == 0 || edge0->height < edge->height)
	edge0 = edge;
      if (edge1 == 0 || edge1->height > edge->height)
	edge1 = edge;
      edge->aboveHead() = edge->node->nextEdge();
      edge->belowHead() = edge->node->prevEdge();
    }

    // Remove incoming edges from sweep list.
    for (int i = 0; i < edgesTH[1].size(); i++) {
      Edge *edge = edgesTH[1][i];
      sweep = edge->node->remove();
      delete edge->node;
    }

    // Insert outgoing edges into the sweep list.
    for (int i = 0; i < edgesTH[0].size(); i++) {
      SweepNode *node = new HeightNode(edgesTH[0][i]);
      if (sweep == 0)
	sweep = node;
      else {
	sweep->insert(node);
	sweep = node->rebalance();
      }
    } 

    // Set aboveTail and belowTail for outgoing edges.
    for (int i = 0; i < edgesTH[0].size(); i++) {
      Edge *edge = edgesTH[0][i];
      Edge *&edge0 = edge->tail()->edge[0][0];
      Edge *&edge1 = edge->tail()->edge[0][1];
      if (edge0 == 0 || edge0->height < edge->height)
	edge0 = edge;
      if (edge1 == 0 || edge1->height > edge->height)
	edge1 = edge;
      edge->aboveTail() = edge->node->nextEdge();
      edge->belowTail() = edge->node->prevEdge();
    }

    edgesTH[0].clear();
    edgesTH[1].clear();
  }

  // Delete orphan vertices.
  for (jv = verts.begin(); jv != verts.end();) {
    Verti jjv = jv++;
    Vert *vert = *jjv;
    if (vert->edge[0][0] == 0 && vert->edge[0][1] == 0 &&
	vert->edge[1][0] == 0 && vert->edge[1][1] == 0 &&
	vert->edge[2][0] == 0 && vert->edge[2][1] == 0) {
      verts.erase(jjv);
      delete vert;
    }
  }

  // make loops
  loops.clear();

  for (int i = 0; i < edges.size() + edgesV.size(); i++) {
    Edge *edge = i < edges.size() ? edges[i] : edgesV[i - edges.size()];
    edge->loop[0] = edge->loop[1] = -1;
  }

  for (int i = 0; i < edges.size() + edgesV.size(); i++) {
    Edge *edge = i < edges.size() ? edges[i] : edgesV[i - edges.size()];
    for (int down = 0; down <= 1; down++) {
      if (edge->loop[down] != -1)
	continue;

      int iloop = loops.size();
      DEdge first(edge, down), curr = first;
      do {
	curr.loop() = iloop;
      } while (++curr != first);

      loops.push_back(Loop(first));
    }
  }

  // make cells
  cells.clear();

  int nloops = loops.size();
  for (int i = 0; i < nloops; i++)
    loops[i].cell = i;

  for (int i = 0; i < edges.size(); i++) {
    Edge *edge = edges[i];
    if (!edge->isVertical())
      for (int down = 0; down <= 1; down++)
	for (int oi = 0; oi <= 1; oi++) {
	  DEdge d1(edge, down);
	  DEdge d2 = d1.face(oi);
	  
	  int root1 = upfind(d1.loop());
	  int root2 = d2.null() ? -1 : upfind(d2.loop());
	  
	  if (root1 < root2)
	    loops[root2].cell = root1;
	  else if (root1 > root2)
	    loops[root1].cell = root2;
	}
  }

  for (int i = 0; i < nloops; i++)
    upfind(i);

  int ncells = 1;
  for (int i = 0; i < nloops; i++) {
    int &cell = loops[i].cell;
    if (cell == -1)
      cell = 0;
    else if (cell == i)
      cell = ncells++;
    else
      cell = loops[cell].cell;
  }

  cells.resize(ncells);
  for (int i = 0; i < nloops; i++)
    cells[loops[i].cell].loops.push_back(i);

  // set depths
  list<Cell *> cqueue;
  cells[0].depth = 0;
  cqueue.push_back(&cells[0]);
  while (!cqueue.empty()) {
    Cell *cell = cqueue.front();
    cqueue.pop_front();

    for (int i = 0; i < cell->loops.size(); i++) {
      DEdge first = loops[cell->loops[i]].d, curr = first;
      do {
	Cell *cell2 = &cells[loops[curr.loop2()].cell];
	if (cell2->depth == -1) {
	  cell2->depth = cell->depth + 1;
	  int loop0 = cell2->loops[0];
	  cell2->loops[0] = curr.loop2();
	  for (int j = 1; j < cell2->loops.size(); j++)
	    if (cell2->loops[j] == curr.loop2()) {
	      cell2->loops[j] = loop0;
	      break;
	    }
	  cqueue.push_back(cell2);
	}
      } while (++curr != first);
    }
  }

  if (!(2 * cells.size() - loops.size() - edges.size() - edgesV.size() + verts.size() == 2)) {
    set<Edge *, Edge::CompareH> sweep;
    vector<Edge *> edgesTH[2];
    vector<Vert *> vertsX;
    int nedges = 0, nedgesV = 0;
    
    for (Verti it = verts.begin(); it != verts.end();) {
      Vert *vert = *it;
      it++;
      if (vert->edge[2][0] != 0) {
	Edge *edgeV = vert->edge[2][0];
	if (edgeV->edge[0][1] != 0) {
	  assert(edgeV->edge[0][1]->edge[0][1] == 0);
	  // assert(loopcount[edgeV->loop[1]] == 2);
	}

	assert(edgeV->edge[0][0] == 0);
	nedgesV++;
	for (Edge *edge = edgeV->edge[0][1]; edge != 0;
	     edge = edge->edge[0][1]) {
	  assert(edge->head() == edgeV->head());
	  nedgesV++;
	}
	assert(edgeV->head() == *it);
      }
      
      for (int th = 1; th >= 0; th--)
	for (Edge *edge = vert->edge[th][1];
	     edge != 0 && edge->getVert(th) == vert;
	     edge = edge->edge[th][0]) {
	  if (edgesTH[th].size() > 0)
	    assert(edge->height > edgesTH[th].back()->height);
	  edgesTH[th].push_back(edge);
	}
      vertsX.push_back(vert);
      if (it != verts.end() && (*it)->p.x == vert->p.x)
	continue;

      for (int i = 0; i < edgesTH[1].size(); i++) {
	Edge *edge = edgesTH[1][i];
	set<Edge *, Edge::CompareH>::iterator it = sweep.find(edge);
	assert(it != sweep.end() && *it == edge);
	Edge *prev = 0, *next = 0;
	if (it != sweep.begin()) {
	  it--;
	  prev = *it;
	  assert(loops[prev->loop[0]].cell == loops[edge->loop[1]].cell);
	  it++;
	}
	it++;
	if (it != sweep.end()) {
	  next = *it;
	  assert(loops[edge->loop[0]].cell == loops[next->loop[1]].cell);
	  it--;
	}
	assert(edge->edge[1][0] == next && edge->edge[1][1] == prev);
      }
	  
      for (int i = 0; i < edgesTH[1].size(); i++)
	sweep.erase(edgesTH[1][i]);

      for (int i = 0; i < vertsX.size(); i++) {
	Vert *vert = vertsX[i];

	bool foundLB = false;
	set<Edge *, Edge::CompareH>::iterator lb;
	for (int th = 0; th <= 1; th++)
	  for (int ud = 0; ud <= 1; ud++) {
	    Edge *edge = vert->edge[th][ud];
	    if (edge == 0)
	      continue;
	    set<Edge *, Edge::CompareH>::iterator it = sweep.lower_bound(edge);
	    if (foundLB)
	      assert(lb == it);
	    else {
	      lb = it;
	      foundLB = true;
	    }
	  }
	assert(foundLB);

	if (vert->edge[2][0] != 0) {
	  Vert *vert2 = vert->edge[2][0]->head();

	  Edge *edge0 = vert->edge[0][0];
	  if (edge0 == 0)
	    edge0 = vert->edge[1][0];
	  assert(edge0 != 0);

	  Edge *edge1 = vert2->edge[0][0];
	  if (edge1 == 0)
	    edge1 = vert2->edge[1][0];
	  assert(edge1 != 0);
	  
	  for (set<Edge *, Edge::CompareH>::iterator it = sweep.begin();
	       it != sweep.end(); it++)
	    assert((*it)->height < edge0->height ||
		   (*it)->height > edge1->height);

	  for (int th = 0; th <= 1; th++)
	    for (int ud = 0; ud <= 1; ud++) {
	      Edge *edge = vert->edge[th][ud];
	      if (edge == 0)
		continue;
	      set<Edge *, Edge::CompareH>::iterator it = sweep.lower_bound(edge);
	      assert(lb == it);
	    }
	}
      }

      for (int i = 0; i < edgesTH[0].size(); i++)
	sweep.insert(edgesTH[0][i]);

      for (int i = 0; i < edgesTH[0].size(); i++) {
	Edge *edge = edgesTH[0][i];
	set<Edge *, Edge::CompareH>::iterator it = sweep.find(edge);
	assert(it != sweep.end() && *it == edge);
	Edge *prev = 0, *next = 0;
	if (it != sweep.begin()) {
	  it--;
	  prev = *it;
	  it++;
	}
	it++;
	if (it != sweep.end()) {
	  next = *it;
	  it--;
	}
	assert(edge->edge[0][0] == next && edge->edge[0][1] == prev);
      }

      nedges += edgesTH[0].size();
      edgesTH[0].clear();
      edgesTH[1].clear();
      vertsX.clear();
    }
    assert(nedges == edges.size());
    assert(nedgesV == edgesV.size());

#ifdef BLEEN
    ofstream out;
    out.open("edges.txt");
    out << setprecision(20);
    map<Function *, int> f2i;
    int fcount = 1;
    for (int i = 0; i < edges.size(); i++) {
      Edge *edge = edges[i];
      Function *f = edge->f;
      if (f2i[f] == 0)
	f2i[f] = fcount++;
      out << edge->height << " "
	  << edge->id << " "
	  << edge->mult << " "
	  << f2i[f] << " "
	  << edge->tail()->p << " "
	  << edge->head()->p << endl;
    }
    for (int i = 0; i < edgesV.size(); i++) {
      Edge *edge = edgesV[i];
      out << edge->height << " "
	  << edge->id << " "
	  << edge->mult << " "
	  << 0 << " "
	  << edge->tail()->p << " "
	  << edge->head()->p << endl;
    }
    out.close();

    int neV = 0;
    for (int i = 0; i < edgesV.size(); i++) {
      Edge *edge = edgesV[i];
      if (edge->edge[0][0] != 0) {
	assert(edge->tail()->edge[2][0] != edge);
	assert(edge->head()->edge[2][1] == edge);
      }
      else
	edgesV[neV++] = edge;
    }
    edgesV.resize(neV);
      
    if (neV > 67) {
      edgesV.resize(67);
      for (int i = 0; i < 66; i++)
	edgesV.erase(edgesV.begin());
      /* double badX = edgesV[0]->tail()->p.x;
      int ne = 0;
      for (int i = 0; i < edges.size(); i++) {
	Edge *edge = edges[i];
	if (edge->tail()->p.x <= badX &&
	    edge->head()->p.x >= badX)
	  edges[ne++] = edge;
      }
      edges.resize(ne);*/
      build();
    }
    
    dontArrange1D = 1;

#ifdef BLEEN
    edgesV.clear();
    for (Verti it = verts.begin(); it != verts.end(); it++) {
      Vert *vert = *it;
      vert->edge[2][0] = 0;
      vert->edge[2][1] = 0;
    }
    build();
#endif

    for (int i = 0; i < edges.size(); i++)
      edges[i]->prev[0] = edges[i]->prev[1] = 0;
    for (int i = 0; i < edgesV.size(); i++)
      edgesV[i]->prev[0] = edgesV[i]->prev[1] = 0;
    for (int i = 0; i < edges.size(); i++)
      for (int sw = 0; sw <=1; sw++) {
	DEdge d(edges[i], sw);
	++d;
	assert(d.edge()->prev[d.down()] == 0);
	d.edge()->prev[d.down()] = edges[i];
      }
    for (int i = 0; i < edgesV.size(); i++)
      for (int sw = 0; sw <=1; sw++) {
	DEdge d(edgesV[i], sw);
	++d;
	assert(d.edge()->prev[d.down()] == 0);
	d.edge()->prev[d.down()] = edgesV[i];
      }

    vector<int> loopcomp(loops.size());
    for (int i = 0; i < loops.size(); i++)
      loopcomp[i] = i;
    for (int i = 0; i < edges.size(); i++) {
      Edge *edge = edges[i];
      if (loopcomp[edge->loop[0]] != loopcomp[edge->loop[1]]) {
	int iloop = min(loopcomp[edge->loop[0]], loopcomp[edge->loop[1]]);
	int jloop = max(loopcomp[edge->loop[0]], loopcomp[edge->loop[1]]);
	for (int j = 0; j < loops.size(); j++)
	  if (loopcomp[j] == jloop)
	    loopcomp[j] = iloop;
      }
    }
    for (int i = 0; i < edgesV.size(); i++) {
      Edge *edge = edgesV[i];
      if (loopcomp[edge->loop[0]] != loopcomp[edge->loop[1]]) {
	int iloop = min(loopcomp[edge->loop[0]], loopcomp[edge->loop[1]]);
	int jloop = max(loopcomp[edge->loop[0]], loopcomp[edge->loop[1]]);
	for (int j = 0; j < loops.size(); j++)
	  if (loopcomp[j] == jloop)
	    loopcomp[j] = iloop;
      }
    }

    int nvertsT = 0;
    int nloopsT = 0;
    int nedgesT = 0;
    
    for (int i = 0; i < loops.size(); i++) {
      if (loopcomp[i] != i) {
	assert(loopcomp[i] < i);
	continue;
      }

      int nverts = 0;
      int nloops = 0;
      int nedges = 0;

      for (Verti it = verts.begin(); it != verts.end(); it++) {
	Vert *vert = *it;
	Edge *edge = vert->edge[0][0];
	if (edge == 0)
	  edge = vert->edge[1][0];
	assert(edge != 0);
	if (loopcomp[edge->loop[0]] == loopcomp[i])
	  nverts++;
      }

      for (int j = 0; j < edges.size(); j++)
	if (loopcomp[edges[j]->loop[0]] == loopcomp[i])
	  nedges++;
      
      for (int j = 0; j < edgesV.size(); j++)
	if (loopcomp[edgesV[j]->loop[0]] == loopcomp[i])
	  nedges++;
      
      for (int j = i; j < loops.size(); j++)
	if (loopcomp[j] == loopcomp[i])
	  nloops++;
      
      if (nloops - nedges + nverts != 2)
	cout << nloops - nedges + nverts << endl;
      nloopsT += nloops;
      nedgesT += nedges;
      nvertsT += nverts;
    }      

    vector<int> comps;
    for (int i = 0; i < loops.size(); i++)
      if (loopcomp[i] == i)
	comps.push_back(i);
    for (int i = 0; i < loops.size(); i++) {
      int comp = loopcomp[i];
      int j;
      for (j = 0; j < comps.size() && comps[j] != comp; j++);
      assert(j < comps.size());
      loopcomp[i] = j;
    }
    
    set<int> outside;
    vector<int> loops1, loops2;
    for (int i = 0; i < edges.size(); i++) {
      Edge *edge = edges[i];
      for (int ud = 0; ud <= 1; ud++) {
	DEdge d(edge, ud);
	int loop = d.loop();
	for (int th = 0; th <= 1; th++) {
	  DEdge d2 = d.face(th);
	  if (d2.null()) {
	    outside.insert(loop);
	    continue;
	  }
	  int loop2 = d2.loop();
	  if (loopcomp[loop] == loopcomp[loop2])
	    continue;
	  int loop1 = loop;
	  if (loopcomp[loop1] > loopcomp[loop2]) {
	    loop1 = loop2;
	    loop2 = loop;
	  }
	  int j;
	  for (j = 0; j < loops1.size(); j++)
	    if (loopcomp[loops1[j]] == loopcomp[loop1] &&
		loopcomp[loops2[j]] == loopcomp[loop2])
	      break;
	  if (j < loops1.size())
	    assert(loops1[j] == loop1 && loops2[j] == loop2);
	  else {
	    loops1.push_back(loop1);
	    loops2.push_back(loop2);
	  }
	}
      }
    }

    for (set<int>::iterator it1 = outside.begin();
	 it1 != outside.end(); it1++) {
      int loop1 = *it1;
      for (set<int>::iterator it2 = outside.begin();
	   it2 != outside.end(); it2++) {
      int loop2 = *it2;
      if (loopcomp[loop1] >= loopcomp[loop2])
	continue;
	  int j;
	  for (j = 0; j < loops1.size(); j++)
	    if (loopcomp[loops1[j]] == loopcomp[loop1] &&
		loopcomp[loops2[j]] == loopcomp[loop2])
	      break;
	  if (j < loops1.size())
	    assert(loops1[j] == loop1 && loops2[j] == loop2);
	  else {
	    loops1.push_back(loop1);
	    loops2.push_back(loop2);
	  }
      }
    }

    vector<Vert *> bverts(comps.size());
    for (Verti it = verts.begin(); it != verts.end(); it++) {
      Vert *vert = *it;
      Edge *edge = vert->edge[0][1];
      if (edge == 0)
	continue;
      Vert *&bvert = bverts[loopcomp[edge->loop[1]]];
      if (bvert == 0 ||
	  vert->p.x < bvert->p.x ||
	  (vert->p.x == bvert->p.x && vert->p.y < bvert->p.y))
	bvert = vert;
    }

    vector<int> oloops(comps.size());
    for (int i = 0; i < comps.size(); i++)
      oloops[i] = bverts[i]->edge[0][1]->loop[1];

    vector<int> cloops(comps.size());
    for (int i = 0; i < comps.size(); i++)
      cloops[i] = -2;
    int ncloops = 0, ncloops2 = 0;
    do {
      ncloops = ncloops2;
      for (int i = 0; i < comps.size(); i++) {
	DEdge start = loops[oloops[i]].d;
	DEdge d = start;
	do {
	  for (int th = 0; th <= 1; th++) {
	    DEdge d2 = d.face(th);
	    if (d2.null()) {
	      assert(cloops[i] < 0);
	      cloops[i] = -1;
	      continue;
	    }
	    int loop2 = d2.loop();
	    int comp2 = loopcomp[loop2];
	    if (loop2 == oloops[comp2]) {
	      if (cloops[i] < -1 && cloops[comp2] >= -1)
		cloops[i] = cloops[comp2];
	      else if (cloops[i] >= -1 && cloops[comp2] < -1)
		cloops[comp2] = cloops[i];
	      else if (cloops[i] >= -1 && cloops[comp2] >= -1)
		assert(cloops[i] == cloops[comp2]);
	    }
	    else {
	      if (cloops[i] < -1)
		cloops[i] = loop2;
	      else
		assert(cloops[i] == loop2);
	    }
	  }
	  ++d;
	} while (d != start);
      }

      int ncloops2 = 0;
      for (int i = 0; i < comps.size(); i++)
	if (cloops[i] >= -1)
	  ncloops2++;
    } while (ncloops2 > ncloops);

    cout << "cloops" << endl;
    for (int i = 0; i < cloops.size(); i++)
      cout << cloops[i] << endl;

    for (int i = 0; i < comps.size(); i++)
      for (int j = i+1; j < comps.size(); j++)
	if (cloops[i] == cloops[j])
	  assert(loops[oloops[i]].cell == loops[oloops[j]].cell);
	else
	  assert(loops[oloops[i]].cell != loops[oloops[j]].cell);

    Verts verts2;
    for (int i = 0; i < edges.size(); i++)
      for (int th = 0; th <= 1; th++)
	verts2.insert(edges[i]->getVert(th));
    for (int i = 0; i < edgesV.size(); i++)
      for (int th = 0; th <= 1; th++)
	verts2.insert(edgesV[i]->getVert(th));
    assert(verts2.size() == verts.size());

    set<int> iloops;
    set<int> icells;
    for (int i = 0; i < edges.size(); i++)
      for (int sw = 0; sw <= 1; sw++) {
	DEdge dedge(edges[i], sw);
	iloops.insert(dedge.loop());
	icells.insert(loops[dedge.loop()].cell);
	int iloop = dedge.loop();
	++dedge;
	assert(dedge.loop() == iloop);
      }
    for (int i = 0; i < edgesV.size(); i++)
      for (int sw = 0; sw <= 1; sw++) {
	DEdge dedge(edgesV[i], sw);
	iloops.insert(dedge.loop());
	icells.insert(loops[dedge.loop()].cell);
	int iloop = dedge.loop();
	++dedge;
	assert(dedge.loop() == iloop);
      }
    assert(loops.size() == iloops.size());
    assert(cells.size() == icells.size());

    vector<int> loopcount(loops.size());
    for (int i = 0; i < loops.size(); i++)
      loopcount[i] = 0;
    for (int i = 0; i < edges.size(); i++) {
      loopcount[edges[i]->loop[0]]++;
      loopcount[edges[i]->loop[1]]++;
    }
    for (int i = 0; i < edgesV.size(); i++) {
      loopcount[edgesV[i]->loop[0]]++;
      loopcount[edgesV[i]->loop[1]]++;
    }

    set<Edge *, Edge::CompareH> sweep;
    vector<Edge *> edgesTH[2];
    int nedges = 0, nedgesV = 0;
    for (Verti it = verts.begin(); it != verts.end();) {
      Vert *vert = *it;
      it++;
      if (vert->edge[2][0] != 0) {
	Edge *edgeV = vert->edge[2][0];
	if (edgeV->edge[0][1] != 0) {
	  assert(edgeV->edge[0][1]->edge[0][1] == 0);
	  assert(loopcount[edgeV->loop[1]] == 2);
	}

	assert(edgeV->edge[0][0] == 0);
	nedgesV++;
	for (Edge *edge = edgeV->edge[0][1]; edge != 0;
	     edge = edge->edge[0][1]) {
	  assert(edge->head() == edgeV->head());
	  nedgesV++;
	}
	assert(edgeV->head() == *it);
      }
      
      for (int th = 1; th >= 0; th--)
	for (Edge *edge = vert->edge[th][1];
	     edge != 0 && edge->getVert(th) == vert;
	     edge = edge->edge[th][0]) {
	  if (edgesTH[th].size() > 0)
	    assert(edge->height > edgesTH[th].back()->height);
	  edgesTH[th].push_back(edge);
	}
      if (it != verts.end() && (*it)->p.x == vert->p.x)
	continue;

      for (int i = 0; i < edgesTH[1].size(); i++) {
	Edge *edge = edgesTH[1][i];
	set<Edge *, Edge::CompareH>::iterator it = sweep.find(edge);
	assert(it != sweep.end() && *it == edge);
	Edge *prev = 0, *next = 0;
	if (it != sweep.begin()) {
	  it--;
	  prev = *it;
	  assert(loops[prev->loop[0]].cell == loops[edge->loop[1]].cell);
	  it++;
	}
	it++;
	if (it != sweep.end()) {
	  next = *it;
	  assert(loops[edge->loop[0]].cell == loops[next->loop[1]].cell);
	  it--;
	}
	assert(edge->edge[1][0] == next && edge->edge[1][1] == prev);
      }
	  
      for (int i = 0; i < edgesTH[1].size(); i++)
	sweep.erase(edgesTH[1][i]);

      bool foundLB = false;
      set<Edge *, Edge::CompareH>::iterator lb;
      for (int th = 0; th <= 1; th++)
	for (int ud = 0; ud <= 1; ud++) {
	  Edge *edge = vert->edge[th][ud];
	  if (edge == 0)
	    continue;
	  set<Edge *, Edge::CompareH>::iterator it = sweep.lower_bound(edge);
	  if (foundLB)
	    assert(lb == it);
	  else {
	    lb = it;
	    foundLB = true;
	  }
	}
      assert(foundLB);

      if (vert->edge[2][0] != 0) {
	Vert *vert2 = vert->edge[2][0]->head();

	Edge *edge0 = vert->edge[0][0];
	if (edge0 == 0)
	  edge0 = vert->edge[1][0];
	assert(edge0 != 0);

	Edge *edge1 = vert2->edge[0][0];
	if (edge1 == 0)
	  edge1 = vert2->edge[1][0];
	assert(edge1 != 0);

	for (set<Edge *, Edge::CompareH>::iterator it = sweep.begin();
	     it != sweep.end(); it++)
	  assert((*it)->height < edge0->height ||
		 (*it)->height > edge1->height);

	for (int th = 0; th <= 1; th++)
	  for (int ud = 0; ud <= 1; ud++) {
	    Edge *edge = vert->edge[th][ud];
	    if (edge == 0)
	      continue;
	    set<Edge *, Edge::CompareH>::iterator it = sweep.lower_bound(edge);
	    assert(lb == it);
	  }
      }

      for (int i = 0; i < edgesTH[0].size(); i++)
	sweep.insert(edgesTH[0][i]);

      for (int i = 0; i < edgesTH[0].size(); i++) {
	Edge *edge = edgesTH[0][i];
	set<Edge *, Edge::CompareH>::iterator it = sweep.find(edge);
	assert(it != sweep.end() && *it == edge);
	Edge *prev = 0, *next = 0;
	if (it != sweep.begin()) {
	  it--;
	  prev = *it;
	  it++;
	}
	it++;
	if (it != sweep.end()) {
	  next = *it;
	  it--;
	}
	assert(edge->edge[0][0] == next && edge->edge[0][1] == prev);
      }

      nedges += edgesTH[0].size();
      edgesTH[0].clear();
      edgesTH[1].clear();
    }
    assert(nedges == edges.size());
    assert(nedgesV == edgesV.size());

    map<double, set<Edge *, Edge::CompareH> > tailedges, headedges;
    for (int i = 0; i < edges.size(); i++) {
      Edge *edge = edges[i];
      assert(edge->mult == 1 || edge->mult == -1);
      tailedges[edge->tail()->p.x].insert(edge);
      headedges[edge->head()->p.x].insert(edge);
    }

    for (int i = 0; i < edges.size(); i++) {
      Edge *edge = edges[i];
      set<Edge *, Edge::CompareH>::iterator it;
      set<Edge *, Edge::CompareH>  &tailset = tailedges[edge->tail()->p.x];
      it = tailset.find(edge);
      it++;
      if (it != tailset.end()) {
	Edge *edge2 = *it;
	assert(edge->tail()->p.x == edge2->tail()->p.x);
	assert(edge->tail()->p.y <= edge2->tail()->p.y);
	assert(edge->mult = -(*it)->mult);
	assert(edge->edge[0][0] == edge2);
	assert(edge2->edge[0][1] == edge);
      }
      set<Edge *, Edge::CompareH>  &headset = headedges[edge->head()->p.x];
      it = headset.find(edge);
      it++;
      if (it != headset.end()) {
	Edge *edge2 = *it;
	assert(edge->head()->p.x == edge2->head()->p.x);
	assert(edge->head()->p.y <= edge2->head()->p.y);
	assert(edge->mult = -(*it)->mult);
	assert(edge->edge[1][0] == edge2);
	assert(edge2->edge[1][1] == edge);
      }
    }
    
    Edge *es[edges.size() + edgesV.size()];
    Edge **es_ = es;
    int n = 0;
    for (int i = 0; i < edges.size(); i++)
      es[n++] = edges[i];
    for (int i = 0; i < edgesV.size(); i++)
      es[n++] = edgesV[i];
    for (int i = 0; i < n; i++) {
      if (es[i]->isVertical())
	continue;
      for (int j = 0; j < n; j++) {
	if (es[j]->isVertical())
	continue;
	for (int oi = 0; oi <= 1; oi++)
	  if (es[i]->getVert(oi)->p.x == es[j]->getVert(oi)->p.x)
	    if (es[i]->height < es[j]->height)
	      assert(es[i]->getVert(oi)->p.y <= es[j]->getVert(oi)->p.y);
	    else
	      assert(es[i]->getVert(oi)->p.y >= es[j]->getVert(oi)->p.y);
      }
    }
    for (int i = 0; i < n; i++) {
      if (!es[i]->isVertical())
	continue;
      for (int j = 0; j < n; j++) {
	if (es[j]->isVertical())
	continue;
	for (int oi = 0; oi <= 1; oi++)
	  if (es[i]->getVert(oi)->p.x == es[j]->getVert(oi)->p.x)
	    assert(es[j]->getVert(oi)->p.y <= es[i]->tail()->p.y ||
		   es[j]->getVert(oi)->p.y >= es[i]->head()->p.y);

	if (es[i]->tail()->p.x >= es[j]->tail()->p.x)
	  continue;
	if (es[i]->head()->p.x <= es[j]->tail()->p.x)
	  continue;

	for (int oi = 0; oi <= 1; oi++) {
	  Vert *vert = es[i]->getVert(oi);
	  assert(vert->edge[0][0] != 0 || vert->edge[1][0] != 0);
	  for (int rl = 0; rl <= 1; rl++)
	    for (Edge *edge = vert->edge[rl][0]; edge != 0;
		 edge = edge->edge[rl][1]) {
	      assert((1-2*oi) * (es[j]->height - edge->height) > 0);
	      if (edge == vert->edge[rl][1])
		break;
	    }
	}
      }
    }

    for (int i = 0; i < n; i++)
      for (int oi = 0; oi <= 1; oi++) {
	Vert *vert = es[i]->getVert(oi);
	for (int j = 0; j < n; j++) {
	  if (es[j]->isVertical())
	    continue;
	  Edge *edge = es[j];
	  if (edge->tail()->p.x >= vert->p.x ||
	      edge->head()->p.x <= vert->p.x)
	    continue;

	  int sgn = 0;
	  for (int rl = 0; rl <= 1; rl++)
	    if (vert->edge[rl][0] != 0)
	      sgn = edge->height - vert->edge[rl][0]->height > 0 ? 1 : -1;

	  for (int rl = 0; rl <= 1; rl++)
	    for (Edge *e = vert->edge[rl][0]; e != 0; e = e->edge[rl][1]) {
	      assert(sgn * (edge->height - e->height) > 0);
	      if (e == vert->edge[rl][1])
		break;
	    }
	}
      }
#endif
  }

#ifdef MERGE_VERTICALS
  for (Verti it = verts.begin(); it != verts.end(); ++it) {
    Vert *vert = *it;
    assert(vert->edge[0][0] != 0 ||
	   vert->edge[1][0] != 0);
  }
#endif

  assert(2 * cells.size() - loops.size() - edges.size() - edgesV.size() + verts.size() == 2);
}

void Embedding::setCellState (vector<int> &cn, const CellTest &test, int iCell) {
  Cell &cell = cells[iCell];
  assert(cell.state == UNKNOWN);
  cells[iCell].state = test(cn) ? INSIDE : OUTSIDE;

  for (int i = 0; i < cell.loops.size(); i++) {
    DEdge first = loops[cell.loops[i]].d, curr = first;
    do {
      int iCell2 = loops[curr.loop2()].cell;
      if (cells[iCell2].state == UNKNOWN) {
	cn[curr.id()] -= curr.mult();
	setCellState(cn, test, iCell2);
	cn[curr.id()] += curr.mult();
      }
    } while (++curr != first);
  }
}

void Embedding::getBoundary () {
  bool unbounded = cells[0].state == INSIDE;

  int ne = 0;
  for (int i = 0; i < edges.size(); i++) {
    Edge *edge = edges[i];
    if (cells[loops[edge->loop[0]].cell].state !=
	cells[loops[edge->loop[1]].cell].state) {
      edge->mult = cells[loops[edge->loop[0]].cell].state == INSIDE ? 1 : -1;
      edge->id = 0;
      // edge->info = 0;
      edges[ne++] = edge;
    }
    else
      delete edge;
  }
  edges.resize(ne);

  ne = 0;
  for (int i = 0; i < edgesV.size(); i++) {
    Edge *edge = edgesV[i];
    if (cells[loops[edge->loop[0]].cell].state !=
	cells[loops[edge->loop[1]].cell].state) {
      edge->mult = cells[loops[edge->loop[0]].cell].state == INSIDE ? 1 : -1;
      edge->id = 0;
      // edge->info = 0;
      edgesV[ne++] = edge;
    }
    else
      delete edge;
  }
  edgesV.resize(ne);
  
  build();

  CellState even = unbounded ? INSIDE : OUTSIDE;
  CellState odd = unbounded ? OUTSIDE : INSIDE;
  for (int i = 0; i < cells.size(); i++)
    if (cells[i].depth % 2 == 0)
      cells[i].state = even;
    else
      cells[i].state = odd;
}

void Embedding::arrange () {
  Timing::begin(timeArrangeTotal);
  vector<Event *> events;

  if (sumInfo)
    arrangeM(events);
  else
    arrange(events);

  splitEdges(events);

  for (int i = 0; i < events.size(); i++)
    delete events[i];
  setHeights();
  Timing::end(timeArrangeTotal);

  if (sumInfo) {
    for (int i = 0; i < edgesV.size(); i++) {
      Edge *edge = edgesV[i];
      edge->id = edge->info = 0;
    }
    for (int i = 0; i < edges.size(); i++) {
      Edge *edge = edges[i];
      edge->info = edge->id;
      edge->id = 0;
    }
  }

  embed();
  mergeEdgesV ();

#ifdef BLEEN
  int edgesSize = edges.size();
  int edgesVSize = edgesV.size();
  embed();
  assert(edgesSize == edges.size());
  assert(edgesVSize == edgesV.size());
#endif

  build();
}

void Embedding::mergeEdgesV () {
  buildLists();

  for (Verti it = verts.begin(); it != verts.end();) {
    Vert *vert = *it;
    Verti jt = it++;
    if (vert->edge[0][0] != 0 || vert->edge[1][0] != 0)
      continue;
    verts.erase(jt);

    if (vert->edge[2][0] == 0) {
      assert(vert->edge[2][1] == 0);
      delete vert;
      continue;
    }

    for (Edge *edge0 = vert->edge[2][0]; edge0 != 0;
	 edge0 = edge0->edge[1][1]) {
      bool found = false;
      for (Edge *edge1 = vert->edge[2][1]; !found && edge1 != 0;
	   edge1 = edge1->edge[0][0])
	if (edge0->id == edge1->id && edge0->mult == edge1->mult)
	  found = true;
      assert(found);
    }

    assert(vert->edge[2][1] != 0);
    Vert *vert0 = vert->edge[2][1]->tail();
    int n0 = 0, n1 = 0;
    for (Edge *edge0 = vert->edge[2][0]; edge0 != 0;
	 edge0 = edge0->edge[1][1]) {
      n0++;
      edge0->setTail(vert0);
    }
    for (Edge *edge1 = vert->edge[2][1]; edge1 != 0;
	 edge1 = edge1->edge[0][0]) {
      n1++;
      edge1->setHead(edge1->tail());
    }
    assert(n0 == n1);
    vert0->edge[2][0] = vert->edge[2][0];
    delete vert;
  }

  int ne = 0;
  for (int i = 0; i < edgesV.size(); i++) {
    Edge *edge = edgesV[i];
    if (edge->head() == edge->tail())
      delete edge;
    else
      edgesV[ne++] = edge;
  }
  edgesV.resize(ne);
}

inline Embedding::Edge *forward (Embedding::Edge *edge) {
  return edge->tail() == edge->head() ? edge->edge[0][0] : edge;
}

void Embedding::mergeEdges () {
  bool unbounded = cells[0].state == INSIDE;

  for (Verti iv = verts.begin(); iv != verts.end();) {
    Vert *vert = *iv;
    if (vert->edge[2][0] != 0 || vert->edge[2][1] != 0) {
      ++iv;
      continue;
    }
    for (int hl = 0; hl <= 1; hl++)
      while (vert->edge[1][hl] != 0 && vert->edge[0][hl] != 0 &&
	     vert->edge[1][hl]->f == vert->edge[0][hl]->f) {
	Edge *left = vert->edge[1][hl];
	Edge *right = vert->edge[0][hl];
	assert(left->tail() != left->head());
	assert(right->tail() != right->head());

	// remove left from vert
	if (vert->edge[1][!hl] == left)
	  vert->edge[1][0] = vert->edge[1][1] = 0;
	else
	  vert->edge[1][hl] = left->edge[1][!hl];

	// remove right from vert
	if (vert->edge[0][!hl] == right)
	  vert->edge[0][0] = vert->edge[0][1] = 0;
	else
	  vert->edge[0][hl] = right->edge[0][!hl];

	// ``move'' vert to right->head()
	Vert *head = right->head();
	left->setHead(head);
	right->setTail(head);

	// replace right by left in head
	for (int ud = 0; ud <= 1; ud++) {
	  left->edge[1][ud] = right->edge[1][ud];
	  if (head->edge[1][ud] == right)
	    head->edge[1][ud] = left;
	  else
	    left->edge[1][ud]->edge[1][!ud] = left;
	}

	// set forwarding address
	right->edge[0][0] = left;
      }
    
    if (vert->edge[0][0] == 0 && vert->edge[1][0] == 0) {
      assert(vert->edge[0][1] == 0 && vert->edge[1][1] == 0);
      Verti temp = iv;
      ++iv;
      verts.erase(temp);
      delete vert;
    }
    else
      ++iv;
  }

  for (int i = 0; i < edges.size(); i++) {
    Edge *edge = edges[i];
    for (int th = 0; th <= 1; th++)
      for (int ud = 0; ud <= 1; ud++) {
	Edge *e = edge->edge[th][ud];
	if (e != 0 && e->tail() == e->head())
	  edge->edge[th][ud] = e->edge[0][0];
      }
  }

  int ne = 0;
  for (int i = 0; i < edges.size(); i++) {
    Edge *edge = edges[i];
    if (edge->tail() == edge->head())
      delete edge;
    else
      edges[ne++] = edge;
  }
  edges.resize(ne);

  setHeights();
  build();

  CellState even = unbounded ? INSIDE : OUTSIDE;
  CellState odd = unbounded ? OUTSIDE : INSIDE;
  for (int i = 0; i < cells.size(); i++)
    if (cells[i].depth % 2 == 0)
      cells[i].state = even;
    else
      cells[i].state = odd;
}  

Embedding::Embedding (vector<Segment> &segs, bool oddDepthIsInside)
  : numAdded(0), ignoreIDs(false), sumInfo(0) {
  // cout << "Embedding segs " << this << endl;

  for (int i = 0; i < segs.size(); i++) {
    Segment &seg = segs[i];
    addEdge(seg.getTail(), seg.getHead(), seg.getF());
  }

  arrange();

  int ne = 0;
  for (int i = 0; i < edges.size(); i++) {
    Edge *edge = edges[i];
    if (cells[loops[edge->loop[0]].cell].depth !=
	cells[loops[edge->loop[1]].cell].depth)
      edges[ne++] = edge;
    else
      delete edge;
  }
  edges.resize(ne);

  ne = 0;
  for (int i = 0; i < edgesV.size(); i++) {
    Edge *edge = edgesV[i];
    if (cells[loops[edge->loop[0]].cell].depth !=
	cells[loops[edge->loop[1]].cell].depth)
      edgesV[ne++] = edge;
    else
      delete edge;
  }
  edgesV.resize(ne);

  build();

  if (oddDepthIsInside) {
    for (int i = 0; i < cells.size(); i++)
      if (cells[i].depth % 2 == 0)
	cells[i].state = OUTSIDE;
      else
	cells[i].state = INSIDE;
  }
  else {
    Edge *eMax = 0;
    for (int i = 0; i < edges.size(); i++) {
      Edge *edge = edges[i];
      if (edge->aboveTail() && edge->belowTail() &&
	  edge->aboveHead() && edge->belowHead())
	continue;
      if (eMax && (edge->head()->p.x - edge->tail()->p.x <=
		   eMax->head()->p.x - eMax->tail()->p.x))
	continue;
      eMax = edge;
    }
    assert(eMax->mult != 0);
    
    bool unbounded =
      (eMax->belowTail() == 0 || eMax->belowHead() == 0) && eMax->mult < 0 ||
      (eMax->aboveTail() == 0 || eMax->aboveHead() == 0) && eMax->mult > 0;
    
    vector<int> cn(1);
    cn[0] = unbounded;
    setCellState(cn, GTZero());
  }

  getBoundary();
}

Embedding::Embedding (vector<Embedding *> &in, vector<int> &cn,
		      const CellTest &test) 
  : numAdded(0), ignoreIDs(false), sumInfo(0) {
  // cout << "Embedding cn " << this << endl;

  static int timeOverlayTotal;
  int timeArrange = -timeArrangeTotal;
  int timeEmbed = -timeEmbedTotal;
  int timeOverlay = -timeOverlayTotal;
  int timeLion = 0;

  Timing::begin(timeOverlayTotal);

  for (int i = 0; i < in.size(); i++) {
    vector<Edge *> &edges = in[i]->edges;
    for (int j = 0; j < edges.size(); j++) {
      Edge *e = edges[j];
      Edge *edge = addEdge(e->tail()->p, e->head()->p, e->f,
			   i, e->mult, e->info);
      edge->height = e->height;
    }
    vector<Edge *> &edgesV = in[i]->edgesV;
    for (int j = 0; j < edgesV.size(); j++) {
      Edge *e = edgesV[j];
      Edge *edge = addEdge(e->tail()->p, e->head()->p, e->f,
			   i, e->mult, e->info);
    }
  }

  arrange();
  setCellState(cn, test);
  getBoundary();

  Timing::end(timeOverlayTotal);
#ifdef EMBED_TIMING
  cout << "timeArrangeTotal " << timeArrangeTotal << endl;
  cout << "timeEmbedTotal " << timeEmbedTotal << endl;
  cout << "timeOverlayTotal " << timeOverlayTotal << endl;
  timeArrange += timeArrangeTotal;
  cout << "timeArrange " << timeArrange << endl;
  timeEmbed += timeEmbedTotal;
  cout << "timeEmbed " << timeEmbed << endl;
  timeOverlay += timeOverlayTotal;
  cout << "timeOverlay " << timeOverlay << endl;
  cout << "timeLion " << timeLion << endl;
#endif
}

Embedding::Embedding (Embedding *that, const Transformation &tr)
  : numAdded(0), ignoreIDs(false), sumInfo(0) {
  // cout << "Embedding tr " << this << endl;

  static int timeTransformTotal;
  static int timeTurnTotal;
  static int timeInconTotal;
  int timeTransform = -timeTransformTotal;
  int timeTurn = -timeTurnTotal;
  int timeIncon = -timeInconTotal;
  int timeEmbed = -timeEmbedTotal;

  Timing::begin(timeTransformTotal);

  if (tr.getR().y == 0) {
    vector<Segment> segs;

    int maxHeight = 0;
    for (int i = 0; i < that->edges.size(); i++) {
      Edge *e = that->edges[i];
      if (e->height > maxHeight)
	maxHeight = e->height;
    }

    for (int i = 0; i < that->edges.size(); i++) {
      Edge *e = that->edges[i];
      Segment s(e->tail()->p, e->head()->p, e->f);
      segs.clear();
      s.transform(tr, segs);
      if (isImplicit())
	makeImplicit(segs);
      if (segs.size() == 0)
	continue;
      if (segs.size() == 1 && segs[0].getTail() == segs[0].getHead())
	continue;

      Ptr<Function> f = 0;
      double maxLength = 0;
      double length;
      for (int j = 0; j < segs.size(); j++) {
	Segment &seg = segs[j];
	if (seg.getF() != 0 &&
	    (length = seg.getTail().distance(seg.getHead())) > maxLength) {
	  maxLength = length;
	  f = seg.getF();
	}
      }
      // assert(!(f == 0));
      Edge *edge = addEdge(segs.front().getTail(), segs.back().getHead(), f,
			   e->id, e->mult, e->info);
      if (tr.getR().x > 0)
	edge->height = e->height;
      else
	edge->height = maxHeight - e->height;
    }
    
    for (int i = 0; i < that->edgesV.size(); i++) {
      Edge *e = that->edgesV[i];
      Point tail = tr(e->tail()->p);
      Point head = tr(e->head()->p);
      if (head == tail)
	continue;
      addEdge(tail, head, 0, e->id, e->mult, e->info);
    }
  }
  else {
    Embedding orig;
    orig.ignoreIDs = true;
    for (int i = 0; i < that->edges.size(); i++) {
      Edge *e = that->edges[i];
      Edge *edge = orig.addEdge(e->tail()->p, e->head()->p, e->f, i, e->mult, 0);
      edge->height = e->height;
    }
    for (int i = 0; i < that->edgesV.size(); i++) {
      Edge *e = that->edgesV[i];
      orig.addEdge(e->tail()->p, e->head()->p, 0,
		   that->edges.size() + i, e->mult, -1);
    }
    

    buildLists();

    vector<Ptr<Function> > funcs;
    // funcs[edge->info] is rotated function
    Timing::begin(timeTurnTotal);
    orig.splitAtTurns(tr, funcs);
    Timing::end(timeTurnTotal);
    
    Point u = tr.getR().conjugate();
    if (u.y > 0)
      u = -u;
    
    Timing::begin(timeInconTotal);
    bool split[2] = { true, true };
    for (int i = 0; split[0] || split[1]; i++) {
      int du = i % 2;
      split[du] = orig.splitAtInconsistencies(u, du);
      assert(i < 100);
    }
    Timing::end(timeInconTotal);
    
    for (int i = 0; i < orig.edges.size() + orig.edgesV.size(); i++) {
      Edge *eOrig = i < orig.edges.size() ?
	orig.edges[i] : orig.edgesV[i - orig.edges.size()];
      Edge *eThat = eOrig->id < that->edges.size() ?
	that->edges[eOrig->id] : that->edgesV[eOrig->id - that->edges.size()];
      Point tail = tr(eOrig->tail()->p);
      Point head = tr(eOrig->head()->p);
      if (tail == head)
	;
      else if (tail.x == head.x)
	addEdge(tail, head, 0, i, eOrig->mult, eThat->info);
      else if (eOrig->f == 0 || funcs.at(eOrig->info) == 0) {
	Line *line = Line::alloc(tail, head);
	Ptr<Function> tmp = line;
	Ptr<Function> f = line;
	if (isImplicit())
	  f = makeImplicit(line);
	addEdge(tail, head, f, i, eOrig->mult, eThat->info);
      }
      else
	addEdge(tail, head, funcs[eOrig->info], i, eOrig->mult, eThat->info);
    }
    
    vector<Event *> events;
    arrange(tr, &orig, events);
    splitEdges(events);
    for (int i = 0; i < events.size(); i++)
      delete events[i];
    setHeights();
    
    vector<Edge *> *edgesNV[] = { &edges, &edgesV };
    for (int nv = 0; nv <= 1; nv++) {
      vector<Edge *> &edges = *edgesNV[nv];
      // arrange1d(edges);
      for (int i = 0; i < edges.size(); i++)
	edges[i]->id = 0;
#ifdef BLEEN
      set<Edge *, Edge::CompareF> edgeSet;
      int ne = 0;
      for (int i = 0; i < edges.size(); i++) {
	Edge *edge = edges[i];
	set<Edge *, Edge::CompareF>::iterator it = edgeSet.find(edge);
	if (it == edgeSet.end()) {
	  edgeSet.insert(edge);
	  edges[ne++] = edge;
	  continue;
	}
	(*it)->info = (*it)->info | edge->info;
	delete edge;
      }
      edges.resize(ne);
#endif
    }
  }
  embed();
  build();
  bool unbounded = (that->cells[0].state == INSIDE);
  for (int i = 0; i < cells.size(); i++)
    if (cells[i].depth % 2 == unbounded)
      cells[i].state = OUTSIDE;
    else
      cells[i].state = INSIDE;

  Timing::end(timeTransformTotal);
#ifdef EMBED_TIMING
  cout << "timeTurnTotal " << timeTurnTotal << endl;
  cout << "timeInconTotal " << timeInconTotal << endl;
  cout << "timeEmbedTotal " << timeEmbedTotal << endl;
  cout << "timeTransformTotal " << timeTransformTotal << endl;
  timeTransform += timeTransformTotal;
  cout << "timeTransform " << timeTransform << endl;
  timeTurn += timeTurnTotal;
  cout << "timeTurn " << timeTurn << endl;
  timeIncon += timeInconTotal;
  cout << "timeIncon " << timeIncon << endl;
  timeEmbed += timeEmbedTotal;
  cout << "timeEmbed " << timeEmbed << endl;
#endif
}

void Embedding::splitAtTurns (const Transformation &tr,
			      vector<Ptr<Function> > &funcs) {
  Transformation trInv = tr.inverse();
  vector<Segment> segs, segs2;
  vector<Edge *> *edgesNV[] = { &edges, &edgesV };
  for (int h = 0; h < 1; h++) {
    int size = edgesNV[h]->size();
    for (int i = 0; i < size; i++) {
      Edge *edge = (*edgesNV[h])[i];
      if (h == 0)
	assert(edge->tail()->p.x < edge->head()->p.x);
      else
	assert(edge->tail()->p.x == edge->head()->p.x);

      Segment seg(edge->tail()->p, edge->head()->p, edge->f);
      segs.clear();
#ifdef BLEEN
      double dist2tail = 0;
      double dist2head = 0;
      if (edge->f != 0 && edge->f->getType() == Function::IMPLICIT) {
	dist2tail = edge->f->distance(edge->tail()->p);
	dist2head = edge->f->distance(edge->head()->p);
      }
#endif
      seg.transform(tr, segs);
      edge->info = -1;
      if (segs.size() == 0)
	continue;

      assert(segs.front().getTail() == tr(edge->tail()->p));
      assert(segs.back().getHead() == tr(edge->head()->p));
      for (int i = 1; i < segs.size(); i++)
	assert(segs[i-1].getHead() == segs[i].getTail());

      if (isImplicit())
	makeImplicit(segs);

      assert(segs.front().getTail() == tr(edge->tail()->p));
      assert(segs.back().getHead() == tr(edge->head()->p));
      for (int i = 1; i < segs.size(); i++)
	assert(segs[i-1].getHead() == segs[i].getTail());

      Edge *edge2 = 0;
      for (int i = 0; i < segs.size() && edge2 != edge; i++) {
	Point p = segs[i].getHead();
	if (i < segs.size() - 1) {
	  assert(p == segs[i+1].getTail());
	  p = trInv(p);
	}
	else
	  p = edge->head()->p;
	if (p.x < edge->tail()->p.x)
	  p.x = edge->tail()->p.x;
	if (p.x > edge->head()->p.x)
	  p.x = edge->head()->p.x;
	if (p == edge->tail()->p)
	  continue;
	if (p == edge->head()->p)
	  edge2 = edge;
	else if (edge->tail()->p.x != p.x) {
	  edge2 = addEdge(edge->tail()->p, p, edge->f,
			  edge->id, edge->mult, edge->info);
	  edge2->height = edge->height;
	}
	else
	  edge2 = addEdge(edge->tail()->p, p, 0,
			  edge->id, edge->mult, edge->info);

	if (p != edge->head()->p) {
	  edge->setTail(addVert(p));
#ifdef BLEEN
	  double distp = edge->f->distance(p);
	  if (distp > 1e-4) {
	    double distq = segs[0].getF()->distance(segs[0].getHead());
	    Abranch *f = dynamic_cast<Abranch *>((Function *)edge->f);
	    Abranch *g = dynamic_cast<Abranch *>((Function *)segs[0].getF());
	    cout << f->curve->f << endl;
	    cout << g->curve->f << endl;
	    Acurve *c = f->curve;
	    vector<Segment> segs;
	    for (int i = 1; i < c->zones[0].size(); i += 2)
	      segs.push_back(Segment(c->bifs[-c->zones[0][i-1]-1]->point,
				     c->bifs[-c->zones[0][i+1]-1]->point,
				     c->branches[c->zones[0][i]-1]));
	    segdump("zone.fig", segs);
	    cout << "this is it" << endl;
	  }
#endif
	}

	if (edge->tail()->p.x == edge->head()->p.x)
	  edge->f = 0;
	edge2->info = funcs.size();
	funcs.push_back(segs[i].getF());
	Point tailTR = tr(edge2->tail()->p);
	Point headTR = tr(edge2->head()->p);
	if (tailTR.x != headTR.x && funcs.back() == 0) {
	  Segment s(edge2->tail()->p, edge2->head()->p, edge2->f);
	  segs2.clear();
	  s.transform(tr, segs2);
	  if (isImplicit())
	    makeImplicit(segs);
	  assert(segs2.size() > 0);
	  funcs.back() = segs2[0].getF();
	  assert(funcs.back() != 0);
	  if (p == edge->head()->p)
	    break;
	}
	else if (tailTR.x == headTR.x && funcs.back() != 0)
	  funcs.back() = 0;
      }
    }
  }

  int ne = 0;
  for (int i = 0; i < edges.size(); i++) {
    Edge *edge = edges[i];
    if (edge->tail()->p.x == edge->head()->p.x)
      edgesV.push_back(edge);
    else
      edges[ne++] = edge;
  }
  edges.resize(ne);    
}

bool Embedding::splitAtInconsistencies (Point u, int du) {
  if (du)
    u = -u;

  buildLists();
  
  SweepNode *sweep = 0;

  // Vectors of outgoing and incoming edges.
  vector<Edge *> edgesTH[2];
  vector<Vert *> vertsX;
  map<Edge *, set<double> > e2x;

  Verti jv;
  for (Verti iv = verts.begin(); iv != verts.end(); iv = jv) {
    // Determine range of vertices with same x.
    double x = (*iv)->p.x;
    for (jv = iv; jv != verts.end() && (*jv)->p.x == x; ++jv)
      vertsX.push_back(*jv);
    
    // Get incoming and outgoing edges for this x.
    for (int i = 0; i < vertsX.size(); i++) {
      Vert *vert = vertsX[i];
      for (int th = 0; th <= 1; th++) {
	for (Edge *edge = vert->edge[th][0]; edge != 0;
	     edge = edge->edge[th][0])
	  if (u * edge->tail()->p < u * edge->head()->p)
	    edgesTH[th].push_back(edge);
	  else
	    new HeightNode(edge);
      }
    }

    for (int i = 0; i < edgesTH[1].size(); i++) {
      Edge *edge = edgesTH[1][i];
      if (u * edge->above()->p >= u * edge->head()->p) {
	assert(edge->tail()->p.x < edge->above()->p.x &&
	       edge->head()->p.x > edge->above()->p.x);
	e2x[edge].insert(edge->above()->p.x);
      }
      sweep = edge->node->remove();
    }

    for (int i = 0; i < edgesTH[0].size(); i++)
      new HeightNode(edgesTH[0][i]);

    for (int i = 0; i < vertsX.size(); i++) {
      Vert *vert = vertsX[i];
      Edge *edge = vert->edge[0][0] ? vert->edge[0][0] : vert->edge[1][0];
      // assert(edge != 0);
      if (edge == 0)
	continue;
      Edge *edgeDU[2] = { 0, 0 };
      if (sweep != 0)
	sweep->locate(edge, edgeDU[0], edgeDU[1]);
      while (edgeDU[du] != 0 && u * vert->p > u * edgeDU[du]->above()->p) {
	edgeDU[du]->above() = vert;
	edgeDU[du] = edgeDU[du]->node->stepEdge(du);
      }
      while (edgeDU[!du] != 0 && u * edgeDU[!du]->above()->p >= u * vert->p) {
	e2x[edgeDU[!du]].insert(vert->p.x);
	edgeDU[!du] = edgeDU[!du]->node->stepEdge(!du);
      }
    }

    for (int i = 0; i < edgesTH[1].size(); i++)
      delete edgesTH[1][i]->node;

    for (int i = 0; i < edgesTH[0].size(); i++) {
      Edge *edge = edgesTH[0][i];
      if (sweep == 0)
	sweep = edge->node;
      else {
	sweep->insert(edge->node);
	sweep = edge->node->rebalance();
      }
      edge->above() = edge->tail();
    }

    // Delete nodes of edges with wrong slope.
    for (int i = 0; i < vertsX.size(); i++) {
      Vert *vert = vertsX[i];
      for (int th = 0; th <= 1; th++) {
	for (Edge *edge = vert->edge[th][0]; edge != 0;
	     edge = edge->edge[th][0])
	  if (!(u * edge->tail()->p < u * edge->head()->p))
	    delete edge->node;
      }
    }
    edgesTH[0].clear();
    edgesTH[1].clear();
    vertsX.clear();
  }

  for (map<Edge *, set<double> >::iterator it = e2x.begin();
	 it != e2x.end(); ++it) {
    Edge *edge = (*it).first;
    set<double> &sx = (*it).second;
    for (set<double>::iterator it = sx.begin(); it != sx.end(); ++it) {


      Timing::splitAtInconsistencies++;
      // cout << "splitAtInconsistency" << endl;

      double x = *it;
      Edge *e = addEdge(edge->tail()->p, Point(x, edge->getY(x)), edge->f,
			edge->id, edge->mult, edge->info);
      e->height = edge->height;
      edge->setTail(e->head());
    }
  }

  if (e2x.empty())
    return false;

  embed();
  return true;
}

void Embedding::arrange (const Transformation &tr, Embedding *orig,
			 vector<Event *> &events) {
  assert(events.empty());
  orig->buildLists();
  RotateNode::ry = -tr.getR().y;
  SweepNode *sweep = 0;

  for (int i = 0; i < edges.size(); i++) {
    Edge *edge = edges[i];
    events.push_back(new Event(Event::INSERT, edge->tail()->p, edge, 0, 0));
    events.push_back(new Event(Event::REMOVE, edge->head()->p, edge, 0, 0));
  }
  sort(events.begin(), events.end(), Event::Compare());
  reverse(events.begin(), events.end());

  for (int i = 0; i < events.size(); i++) {
    Event *event = events[i];
    Edge *c = event->c;
    Point p = event->p;
    Event::Type type = event->type;

    if (type == Event::REMOVE) {
      event->a = c->node->prevEdge();
      event->b = c->node->nextEdge();

#ifdef BLEEN
      // debug
      double cy = c->f->getY(p.x);
      if (event->a != 0) {
	double ay = event->a->f->getY(p.x);
	double acy = ay - cy;
	static double maxACY;
	if (maxACY < acy) {
	  maxACY = acy;
	  cout << "maxACY " << maxACY << endl;
	}
      }
      if (event->b != 0) {
	double by = event->b->f->getY(p.x);
	double cby = cy - by;
	static double maxCBY;
	if (maxCBY < cby) {
	  maxCBY = cby;
	  cout << "maxCBY " << maxCBY << endl;
	}
      }
#endif

      sweep = c->node->remove();
      delete c->node;
    }
    else {
      new RotateNode(c);
      c->subEdge() = c->id < orig->edges.size() ?
	orig->edges[c->id] : orig->edgesV[c->id - orig->edges.size()];
      if (sweep == 0)
	sweep = c->node;
      else {
	sweep->insert(c->node);
	sweep = c->node->rebalance();
      }
      event->a = c->node->prevEdge();
      event->b = c->node->nextEdge();

#ifdef BLEEN
      // debug
      double cy = c->f->getY(p.x);
      if (event->a != 0) {
	double ay = event->a->f->getY(p.x);
	double acy = ay - cy;
	static double maxACY;
	if (maxACY < acy) {
	  maxACY = acy;
	  cout << "maxACY " << maxACY << endl;
	}
      }
      if (event->b != 0) {
	double by = event->b->f->getY(p.x);
	double cby = cy - by;
	static double maxCBY;
	if (maxCBY < cby) {
	  maxCBY = cby;
	  cout << "maxCBY " << maxCBY << endl;
	}
      }
#endif
    }
  }
}

double Embedding::RotateNode::ry;

int Embedding::RotateNode::compare (Treap *that_) {
  RotateNode *that = dynamic_cast<RotateNode *>(that_);
  Edge *thisEdge = this->edge, *thatEdge = that->edge;
  Edge *thisOrig = thisEdge->subEdge(), *thatOrig = thatEdge->subEdge();

  assert(thisEdge->mult == thisOrig->mult ||
	 thisEdge->mult == -thisOrig->mult);
  assert(thatEdge->mult == thatOrig->mult ||
	 thatEdge->mult == -thatOrig->mult);

  if (thisEdge->tail()->p == thatEdge->tail()->p &&
      thisEdge->head()->p == thatEdge->head()->p)
    if (thatEdge->mult == thatOrig->mult)
      return thisOrig->height - thatOrig->height;
    else
      return thatOrig->height - thisOrig->height;

  if (thisEdge->tail()->p.x < thatEdge->tail()->p.x ||
      thisEdge->tail()->p == thatEdge->tail()->p &&
      thisEdge->head()->p.x > thatEdge->head()->p.x)
    return -that->compare(this);

  int iVert = thisEdge->tail()->p == thatEdge->tail()->p;
  int sw = thisEdge->mult != thisOrig->mult;
  Vert *vert = thisOrig->getVert(iVert ^ sw);

  if (thisOrig->tail()->p.x == thisOrig->head()->p.x) {
    if (vert->edge[0][0] != 0)
      thisOrig = vert->edge[0][0];
    else if (vert->edge[1][0] != 0)
      thisOrig = vert->edge[1][0];
    else
      assert(0);
  }
  Vert *vert2 = thisOrig->getVert(vert == thisOrig->tail());
  assert(vert != vert2);

  if (vert->p.x < thatOrig->tail()->p.x ||
      vert->p.x == thatOrig->tail()->p.x &&
      vert2->p.x < thatOrig->tail()->p.x)
    return ry < 0 ? -1 : 1;
  if (vert->p.x > thatOrig->head()->p.x ||
      vert->p.x == thatOrig->head()->p.x &&
      vert2->p.x > thatOrig->head()->p.x)
    return ry < 0 ? 1 : -1;

  if (thatEdge->mult == thatOrig->mult)
    return thisOrig->height - thatOrig->height;
  else
    return thatOrig->height - thisOrig->height;
}

void Embedding::draw (const char *filename) {
#ifdef NODRAW
  return;
#endif
  cout << "start draw " << filename << endl;
  cout << getNumVerts() << " verts "
       << getNumEdges() << " edges " << endl;
  ofstream out;
  out.open(filename);
  out << "#FIG 3.2" << endl
      << "Landscape" << endl
      << "Center" << endl
      << "Inches" << endl
      << "Letter  " << endl
      << "100.00" << endl
      << "Single" << endl
      << "-2" << endl
      << "1200 2" << endl;

  draw(out);
  out.close();
  cout << "end draw " << filename << endl;
}  

void Embedding::draw (const char *filename,
		      const vector<Embedding *> &regions) {
  cout << "start draw " << filename << endl;
  ofstream out;
  out.open(filename);
  out << "#FIG 3.2" << endl
      << "Landscape" << endl
      << "Center" << endl
      << "Inches" << endl
      << "Letter  " << endl
      << "100.00" << endl
      << "Single" << endl
      << "-2" << endl
      << "1200 2" << endl;

  edgedraw(out);
  for (int i = 0; i < regions.size(); i++)
    regions[i]->draw(out);
  out.close();
  cout << "end draw " << filename << endl;
}  

void Embedding::draw (ostream &out) {
  vector<Point> points;

  int maxDepth = 0;
  for (int i = 0; i < cells.size(); i++)
    if (cells[i].depth > maxDepth)
      maxDepth = cells[i].depth;

  for (int i = 0; i < cells.size(); i++) {
    Point minP, maxP;
    bool firstP = true;
    vector<int> &iloops = cells[i].loops;
    for (int j = 0; j < iloops.size(); j++) {
      DEdge first = loops[iloops[j]].d;
      DEdge es = first;
      do {
	++es;
	Point head = (es.head()->p.conjugate() * 1200).rint();
	if (firstP) {
	  minP = head;
	  maxP = head;
	  firstP = false;
	}
	if (minP.x > head.x) minP.x = head.x;
	if (minP.y > head.y) minP.y = head.y;
	if (maxP.x < head.x) maxP.x = head.x;
	if (maxP.y < head.y) maxP.y = head.y;
      } while (es != first);
    }

    int depth = cells[i].depth;
    out << "6 " << minP << " " << maxP << endl;

    for (int j = 0; j < iloops.size(); j++) {
      DEdge first = loops[iloops[j]].d;
      DEdge es = first;
      int n = 0;
      do {
	++es;
	n++;
      } while (es != first);
      
#ifndef NO_SPLINE
      n = 0;
      Point prev = first.tail()->p;
      vector<Point> points;
      do {
	Point next = es.head()->p;
	points.clear();
	approximate(es.edge()->f, prev, next, points);
	for (int i = 1; i < points.size(); i++)
	  n++;
	prev = next;
	++es;
      } while (es != first);

      if (depth % 2 == 0)
	out << "2 3 0 1 0 7 " << maxDepth - depth
	    << " 0 20 0.000 0 0 -1 0 0 " << n + 1 << endl;
      else
	out << "2 3 0 1 0 7 " << maxDepth - depth
	    << " 0 0 0.000 0 0 -1 0 0 " << n + 1 << endl;
	
      prev = first.tail()->p;
      out << "	" << " " << IP(prev.conjugate() * 1200);
      do {
	Point next = es.head()->p;
	points.clear();
	approximate(es.edge()->f, prev, next, points);
	for (int i = 1; i < points.size(); i++)
	  out << " " << IP(points[i].conjugate() * 1200);
	prev = next;
	++es;
      } while (es != first);
      out << endl;
#else
      do {
	Point tail = es.tail()->p;
	Point head = es.head()->p;
	Function *f = es.edge()->f;
	points.clear();
	approximate(f, tail, head, points);
	if (points.size() == 2)
	  drawLine(out, depth, tail, head);
	else
	  drawSpline(out, depth, points);
	++es;
      } while (es != first);
#endif

    }
    out << "-6" << endl;
  }
}

void Embedding::edgedraw (const char *filename) {
#ifdef NODRAW
  return;
#endif
  ofstream out;
  out.open(filename);
  out << "#FIG 3.2" << endl
      << "Landscape" << endl
      << "Center" << endl
      << "Inches" << endl
      << "Letter  " << endl
      << "100.00" << endl
      << "Single" << endl
      << "-2" << endl
      << "1200 2" << endl;
  edgedraw(out);
  out.close();
}

void Embedding::edgedraw (ostream &out) {
  // map<int, int> cct2i;
  // int icct = 0;
  vector<Point> points;
  vector<Edge *> *edgesNV[] = { &edges, &edgesV };
  for (int nv = 0; nv <= 1; nv++) {
    vector<Edge *> &edges = *edgesNV[nv];
    for (int i = 0; i < edges.size(); i++) {
      Edge *edge = edges[i];
      points.clear();
      approximate(edge->f, edge->tail()->p, edge->head()->p, points);
      int depth = edge->info;
#ifdef BLEEN
      if (false && sumInfo) {
	if (cct2i.find(depth) == cct2i.end()) {
	  cct2i[depth] = ++icct;
	  cout << icct << " "
	       << depth << " "
	       << (depth/5)/emb[1]->chains.size() << " "
	       << (depth/5)%emb[1]->chains.size() << " "
	       << depth % 5 << endl;
	  depth = icct;
	}
	else
	  depth = cct2i[depth];
      }
#endif

      depth = 50;
      assert(depth < 999);

      out << "2 1 0 1 0 7 " << depth << " 0 -1 0.000 0 0 -1 0 0 " << points.size()
	  << endl << "\t";
      for (int j = 0; j < points.size(); j++)
	out << " " << IP(points[j].conjugate() * 1200);
      out << endl;
    }
  }
}

void Embedding::getBBox (Point bbox[2]) {
  bbox[0].x = bbox[0].y = Function::FINFINITY;
  bbox[1].x = bbox[1].y = -Function::FINFINITY;

  for (Verti it = verts.begin(); it != verts.end(); ++it) {
    Point p = (*it)->p;
    if (bbox[0].x > p.x)
      bbox[0].x = p.x;
    if (bbox[0].y > p.y)
      bbox[0].y = p.y;
    if (bbox[1].x < p.x)
      bbox[1].x = p.x;
    if (bbox[1].y < p.y)
      bbox[1].y = p.y;
  }
}

void Embedding::complement () {
  for (int i = 0; i < edges.size(); i++)
    edges[i]->mult = -edges[i]->mult;
  for (int i = 0; i < edgesV.size(); i++)
    edgesV[i]->mult = -edgesV[i]->mult;

  for (int i = 0; i < cells.size(); i++) {
    CellState &state = cells[i].state;
    assert(state != UNKNOWN);
    state = state == OUTSIDE ? INSIDE : OUTSIDE;
  }
}

void Embedding::Loop::edgedraw (const char *filename) {
  Embedding emb;
  DEdge dedge = d;
  do {
    emb.addEdge(dedge.tail()->p, dedge.head()->p, dedge.edge()->f);
  } while(++dedge != d);
  emb.edgedraw(filename);
}

void Embedding::Event::report () {
  cout << "this is it" << endl;
}

