#include "Line.h"

using namespace std;

SetP<Line> lineSet;

Line *Line::alloc (const Point &tail, const Point &head) {
  Line line(tail, head);
  Line *p = lineSet.find(&line);
  if (p == 0) {
    p = new Line(line);
    assert(p == lineSet.insert(p));
  }
  return p;
}

Line *Line::alloc (const Point &tail, const Point &head, const Point &diff) {
  Line line(tail, head, diff);
  Line *p = lineSet.find(&line);
  if (p == 0) {
    p = new Line(line);
    assert(p == lineSet.insert(p));
  }
  return p;
}

Line *Line::alloc (istream &in) {
  Line line(in);
  Line *p = lineSet.find(&line);
  if (p == 0) {
    p = new Line(line);
    assert(p == lineSet.insert(p));
  }
  return p;
}

void Line::write (ostream &out) {
  writePoint(out, tail);
  writePoint(out, head);
  writePoint(out, diff);
}

Line::Line (istream &in) {
  readPoint(in, tail);
  readPoint(in, head);
  readPoint(in, diff);
}

Line::Line (const Line &that)
  : tail(that.tail), head(that.head), diff(that.diff) {}

void Line::free () {
  if (this == (Function *) 0x570006)
    cout << "this is it" << endl;
  assert(this == lineSet.find(this));
  lineSet.remove();
  delete this;
}

Line::Line (const Point &tail, const Point &head) {
  if (tail.x < head.x || tail.x == head.x && tail.y < head.y) {
    this->tail = tail;
    this->head = head;
    this->diff = head - tail;
  }
  else {
    this->tail = head;
    this->head = tail;
    this->diff = tail - head;
  }
}

Line::Line (const Point &tail, const Point &head, const Point &diff) {
  if (tail.x < head.x || tail.x == head.x && tail.y < head.y) {
    this->tail = tail;
    this->head = head;
    this->diff = diff;
  }
  else {
    this->tail = head;
    this->head = tail;
    this->diff = -diff;
  }
}

Line::Line (const Line *line, const Transformation &tr) {
  Point tail = tr(line->tail);
  Point head = tr(line->head);
  Point diff = tr.getR()(line->diff);

  if (tail.x < head.x || tail.x == head.x && tail.y < head.y) {
    this->tail = tail;
    this->head = head;
    this->diff = diff;
  }
  else {
    this->tail = head;
    this->head = tail;
    this->diff = -diff;
  }
}

DOUBLE Line::getY (DOUBLE x) const {
  double y = Function::FINFINITY;
  if (x < (tail.x + head.x) / 2) {
    if (diff.x == 0)
      return tail.y;
    y = tail.y + ((x - tail.x) / diff.x) * diff.y;
  }
  else {
    if (diff.x == 0)
      return head.y;
    y = head.y + ((x - head.x) / diff.x) * diff.y;
  }

  if (y < tail.y && y < head.y)
    return std::min(tail.y, head.y);
  if (y > tail.y && y > head.y)
    return std::max(tail.y, head.y);
  return y;
}

DOUBLE Line::distance (Point p) const {
  if (p.x - tail.x < head.x - p.x)
    return fabs((p - tail) % diff) / diff.length();
  else
    return fabs((p - head) % diff) / diff.length();

  assert(diff * (head - tail) > 0);
  Point tp = p - tail;
  double tpDdiff = tp * diff;
  if (tpDdiff < 0)
    return tp.length();
  Point hp = p - head;
  double hpDdiff = hp * diff;
  if (hpDdiff > 0)
    return hp.length();
  if (tpDdiff < -hpDdiff)
    return fabs(tp % diff) / diff.length();
  else
    return fabs(hp % diff) / diff.length();
}

static int solve (DOUBLE a0, DOUBLE b0, DOUBLE a1, DOUBLE b1, DOUBLE *r) {
  if (a0 - a1 < 1e-12 && a1 - a0 < 1e-12)
    return 0;
  if (a0 == a1)
    return 0;

  r[0] = (b0 - b1) / (a1 - a0);
  assert(r[0] < 0 || r[0] == 0 || r[0] > 0);

  if (r[0] == 2)
    cout << "solve:  this is it" << endl;
  return 1;
}

RootList *Line::makeRootList (Function *that) {
  if (that->getType() != LINE)
    return that->makeRootList(this);

  Line *l0 = this;
  Line *l1 = (Line *) that;

  Point u0 = (l0->tail + l0->head) / 2;
  Point v0 = l0->diff;
  Point u1 = (l1->tail + l1->head) / 2;
  Point v1 = l1->diff;

  vector<Point> roots;

  double v0xv1 = v0 % v1;
  if (v0xv1 == 0)
    return new RootList(this, that, roots);

  Point u01 = u1 - u0;
  double u01xv1 = u01 % v1;
  double u01xv0 = u01 % v0;

  Point p;
  if (u01xv1 * u01xv1 * (v0 * v0) < u01xv0 * u01xv0 * (v1 * v1)) {
    double t = u01xv1 / v0xv1;
    p = u0 + v0 * t;
  }
  else {
    double t = u01xv0 / v0xv1;
    p = u1 + v1 * t;
  }

  l0->pushToBox(p);
  l1->pushToBox(p);

  if (l0->inBBox(p) && l1->inBBox(p)) {
    roots.push_back(p);

    static double maxLine;
    Line *lines[2] = { this, l1 };
    for (int i = 0; i < 2; i++) {
      Point q = lines[i]->nearestPoint(p);
      Point pq = q - p;
      double dist = sqrt(pq * pq);
      if (maxLine < dist) {
	maxLine = dist;
	cout << "maxLine " << maxLine << endl;
      }
    }
  }

  return new RootList(this, that, roots);
}

void Line::pushToBox (Point &p) {
  if (p.x <= tail.x || p.x >= head.x)
    return;

  if (p.y < tail.y && p.y < head.y)
    p.y = min(tail.y, head.y);
  else if (p.y > tail.y && p.y > head.y)
    p.y = max(tail.y, head.y);
}

Line *Line::transform (const Transformation &tr, bool ymonotone) {
  if (tr.getT() == Point(0, 0) && tr.getR() == Point(1, 0))
    return this;

  Line line(this, tr);
  Line *p = lineSet.find(&line);
  if (p == 0) {
    p = new Line(this, tr);
    assert(p == lineSet.insert(p));
  }
  return p;
}

void Line::transform (const Transformation &tr, vector<Segment> &out, bool ymonotone) {
  Line *lineT = transform(tr);
  out.push_back(Segment(lineT->tail, lineT->head, lineT));
#ifdef BLEEN
  if (lineT->tail == tr(tail))
    out.push_back(Segment(lineT->tail, lineT->head, lineT));
  else
    out.push_back(Segment(lineT->head, lineT->tail, lineT));
#endif
}

Point Line::nearestPoint (Point p) const {
  Point th = diff;
  double tpth = (p - tail) * th;
  double hpht = (head - p) * th;
  if (tpth < 0)
    return tail;
  if (hpht < 0)
    return head;

  double thth = th * th;
  if (tpth < hpht)
    return tail + th * (tpth / thth);
  else
    return head - th * (hpht / thth);
}
