#ifndef LINE_h
#define LINE_h
#include "Segment.h"

class Line : public Function {
public:
  static Line *alloc (const Point &tail, const Point &head);
  static Line *alloc (const Point &tail, const Point &head, const Point &diff);
  static Line *alloc (istream &in);
  void free ();
  void write (ostream &out);

  bool lessThan (const Line *that) const {
    if (this->tail.x != that->tail.x)
      return this->tail.x < that->tail.x;
    if (this->tail.y != that->tail.y)
      return this->tail.y < that->tail.y;
    if (this->head.x != that->head.x)
      return this->head.x < that->head.x;
    if (this->head.y != that->head.y)
      return this->head.y < that->head.y;
    if (this->diff.x != that->diff.x)
      return this->diff.x < that->diff.x;
    if (this->diff.y != that->diff.y)
      return this->diff.y < that->diff.y;
    return false;
  }

  Point getTail () { return tail; }
  Point getHead () { return head; }
  Point getDiff () { return diff; }
  Point getMid () { return (tail + head) / 2; }

  int getType () const { return LINE; }

  DOUBLE getMinX () const { return tail.x; }

  DOUBLE getMaxX () const { return head.x; }

  DOUBLE getY (DOUBLE x) const;

  DOUBLE getDY (DOUBLE x) const {
    if (diff.x != 0)
      return diff.y / diff.x;
    return Function::FINFINITY;
  }

  DOUBLE distance (Point p) const;

  RootList *makeRootList (Function *that);
  void transform (const Transformation &tr, vector<Segment> &out,
		  bool ymonotone = false);
  void transform (Segment &seg, const Transformation &tr,
		  vector<Segment> &out, bool ymonotone = false) {
    assert(0);
  }
  Line *transform (const Transformation &tr, bool ymonotone = false);
  Point nearestPoint (Point p) const;

  bool inBBox (Point p) {
    return (p.x >= tail.x && p.x <= head.x &&
	    (p.y >= tail.y || p.y >= head.y) &&
	    (p.y <= tail.y || p.y <= head.y));
  }

  void pushToBox (Point &p);

  Ptr<Line> up;

private:
  Line (const Point &tail, const Point &head);
  Line (const Point &tail, const Point &head, const Point &diff);
  Line (const Line *line, const Transformation &tr);
  Line (istream &in);
  Line (const Line &that);
  void operator= (const Line &that) { assert(0); }

  friend class Circle;
  Point tail, head, diff;
};
#endif
