char copyright4[] = "copyright (c) 2008 Victor J. Milenkovic, not for commercial use without permission";

#include "Segment.h"
#include "Line.h"
#include <fstream>
using std::ofstream;

#ifdef SEGMENT_TEST
#include "poly.h"
#include "curve.h"
#include "TCurve.h"
#endif

#define LAZY_ROOTLIST

using std::cout;
using std::endl;

const DOUBLE Transformation::TINY = 1e-12;
const DOUBLE Function::FINFINITY = 1e12 * DBL_MAX;
const int Function::IMPLICIT = 0;
const int Function::LINE = 1;
const int Function::UPPER_CIRCLE = 2;
const int Function::LOWER_CIRCLE = -2;
const int Function::TELOMERE_CURVE = -1;

#ifdef BLEEN
namespace Timing {
  int transform, find, rootlist;
  int sum, sumloops, sumarrange, fit, pointInside;
  int numRootList;
}    
#endif

namespace Timing {
  int sum, sumloops, sumarrange, fit, pointInside, ncells, nbubbles, adapt;
  int arrange, rootlist, dump, transform, find, y, compare;
  int numRootList, numLink, numExtraRoots, newroot, N;
  int numInsertByRoot, degree;
  int rootCount[10];
  int numFG, timeFG, numEvil, timeXY, numXY;
  int rootHist[1024];
  int numrealroots, timerealroots;
  int embeddedEndpoints, embeddingViolations, splitAtInconsistencies;
}    

void Function::incref () {
  refcount++;
}

void Function::decref () {
  if (--refcount == 0)
    free();
}

Segment::Segment (const Point &tail, const Point &head, Function *f_)
  : tail(tail), head(head), f(f_) {
  if (tail.x == head.x)
    f = 0;
  if (f != 0) {
    assert(tail.x != head.x);
    f->check(tail);
    f->check(head);
    f->incref();
  }
  else
    assert(tail.x == head.x);
}

void Segment::transform (const Transformation &tr, vector<Segment> &segs) {
  if (f == 0) {
    assert(tail.x == head.x);
    Point tailT = tr(tail);
    Point headT = tr(head);
    if (tailT.x == headT.x)
      segs.push_back(Segment(tailT, headT, 0));
    else
      segs.push_back(Segment(tailT, headT, Line::alloc(tailT, headT)));
    return;
  }
  
  if (f->getType() == Function::LINE) {
    Point tailT = tr(tail);
    Point headT = tr(head);
    if (tailT.x == headT.x) {
      segs.push_back(Segment(tailT, headT, 0));
      return;
    }
    vector<Segment> segsT;
    f->transform(tr, segsT);
    assert(segsT.size() == 1);
    if (segsT[0].f != 0)
      segs.push_back(Segment(tailT, headT, segsT[0].f));
    else
      segs.push_back(Segment(tailT, headT, Line::alloc(tailT, headT)));
    return;
  }

  Transformation trinv = tr.inverse();
  DOUBLE minX = getMinX();
  DOUBLE maxX = getMaxX();
  vector<Segment> out;

  Timing::begin(Timing::transform);
  if (f->getType() == Function::IMPLICIT) {
    f->transform(*this, tr, segs);
    Timing::end(Timing::transform);
    return;
  }

  f->transform(tr, out);
  Timing::end(Timing::transform);

  int i0 = 0, i1 = out.size() - 1;
  for (int i = 1; i < out.size(); i++) {
    Point p = trinv(out[i].tail);
    if (p.x <= minX)
      i0 = i;
    if (p.x >= maxX) {
      i1 = i - 1;
      break;
    }
  }

  int n = 0;
  for (int i = i0; i <= i1; i++)
    out[n++] = out[i];
  while (out.size() > n)
    out.pop_back();

  Function *fs[out.size()];
  for (int i = 0; i < out.size(); i++) {
    fs[i] = out[i].f;
    if (fs[i] != 0)
      fs[i]->incref();
  }

  {
    Point tailTR = tr(getMin());
    if (tailTR.x == out[0].head.x && fs[0] != 0)
      out[0] = Segment(tailTR, out[0].head, 0);
    else if (tailTR.x != out[0].head.x && fs[0] == 0)
      out[0] = Segment(tailTR, out[0].head,
		       Line::alloc(tailTR, out[0].head));
    else
      out[0] = Segment(tr(getMin()), out[0].head, fs[0]);
  }
  {
    Point headTR = tr(getMax());
    if (headTR.x == out[n-1].tail.x && fs[n-1] != 0)
      out[n-1] = Segment(out[n-1].tail, tr(getMax()), 0);
    else if (headTR.x != out[n-1].tail.x && fs[n-1] == 0)
      out[n-1] = Segment(out[n-1].tail, tr(getMax()),
			 Line::alloc(out[n-1].tail, tr(getMax())));
    else
      out[n-1] = Segment(out[n-1].tail, tr(getMax()), fs[n-1]);
  }

  for (int i = 0; i < out.size(); i++)
    if (fs[i] != 0)
      fs[i]->decref();

  if (tail.x > head.x) {
    int n2 = n / 2;
    for (int i = 0; i < n2; i++) {
      int j = n - 1 - i;
      Segment temp = out[i];
      out[i] = out[j];
      out[j] = temp;
    }
    
    for (int i = 0; i < n; i++)
      out[i] = out[i].reverse();
  }

  for (int i = 0; i < n; i++)
    segs.push_back(out[i]);
}

void report (DOUBLE maxfg) {
  maxfg = maxfg * maxfg;
}

void report2 (DOUBLE maxfg) {
  maxfg = maxfg * maxfg;
}

void dumpFG (Function *f, Function *g) {
#ifdef SEGMENT_TEST
  Acurve *fc = ((Abranch *)f)->curve;
  Acurve *gc = ((Abranch *)g)->curve;
  vector<Segment> segs;
  fc->segments(segs, -2, -1, 1, 2);
  gc->segments(segs, -2, -1, 1, 2);
  
  Acurve *curveA, *curveB;
  vector<Point> croots;
  int ic;
  for (ic = 0; ic < fc->curvs.size(); ic++)
    if (fc->curvs[ic] == gc)
      break;
  if (ic != fc->curvs.size()) {
    croots = fc->croots[ic];
    curveA = fc;
    curveB = gc;
  }
  else {
    for (ic = 0; ic < gc->curvs.size(); ic++)
      if (gc->curvs[ic] == fc)
	break;
    if (ic != gc->curvs.size()) {
      croots = gc->croots[ic];
      curveA = gc;
      curveB = fc;
    }
    else
      return;
  }
  
  for (int ir = 0; ir < croots.size(); ir++) {
    Point p = croots[ir];
    Point p0(p.x - 0.1, p.y);
    Line *line = Line::alloc(p0, p);
    segs.push_back(Segment(p0, p, line));
  }
  
  extern void segdump (const char *filename, const vector<Segment> &segs);
  
  Psystem psys(curveA->f, curveB->f);
  double rootsxy[4*psys.degree()];
  // double rootsxy[100];
  int nr = psys.roots(rootsxy);
  cout << nr << "roots" << endl;
  for (int ir = 0; ir < nr; ir++) {
    Point p(rootsxy[4*ir], rootsxy[4*ir+2]);
    cout << p << endl;
    Point p1(p.x + 0.1, p.y);
    Line *line = Line::alloc(p, p1);
    segs.push_back(Segment(p, p1, line));
  }
  
  segdump("missing.fig", segs);
  cout << curveA->f << endl;
  cout << curveB->f << endl;
  cout << "num FG " << Timing::numFG << endl;
  // roots were generated by a solving
  // Psystem psys(curveA->f, curveB->f);
#endif
}

void checkRootList (Function *f, Function *g, const vector<Point> roots) {
  // return; //debug
  DOUBLE minX = MAX(f->getMinX(), g->getMinX());
  DOUBLE maxX = MIN(f->getMaxX(), g->getMaxX());

  double BBOX = 1;

  if (minX == -Function::FINFINITY)
    minX = roots.size() == 0 ? -BBOX : -2 * fabs(roots[0].x) - BBOX;

  if (maxX == Function::FINFINITY)
    maxX = roots.size() == 0 ? BBOX : 2 * fabs(roots[roots.size()-1].x) + BBOX;

  static DOUBLE maxYY;

  for (int i = 0; i < roots.size(); i++) {
    assert(minX < roots[i].x && roots[i].x < maxX);
    if (fabs(f->getY(roots[i].x) - g->getY(roots[i].x)) > maxYY) {
      maxYY = fabs(f->getY(roots[i].x) - g->getY(roots[i].x));
      cout << "maxYY " << maxYY << endl;
    }
  }
  // return;

  static DOUBLE maxfg = 0;
  bool dumped = false;

  for (int i = 0; i <= roots.size(); i++) {
    DOUBLE a = i == 0 ? minX : roots[i-1].x;
    DOUBLE b = i == roots.size() ? maxX : roots[i].x;
    for (int j = 1; j < 100; j++) {
      DOUBLE c = a + (b - a) * j / 100.0;
      DOUBLE fY = f->getY(c);
      DOUBLE gY = g->getY(c);
      if (i % 2 == 0) {
	// assert(fY < gY);
	if (fY - gY > maxfg) {
	  maxfg = fY - gY;
	  cout << "maxfg " << maxfg << endl;
	}
      }
      else {
	// assert(fY > gY);
	if (gY - fY > maxfg) {
	  maxfg = gY - fY;
	  cout << "maxfg " << maxfg << endl;
	}
      }
      if (maxfg > 0.001) {
#ifdef SEGMENT_TEST
	if (!dumped) {
	  Acurve *fc = ((Abranch *)f)->curve;
	  Acurve *gc = ((Abranch *)g)->curve;
	  vector<Segment> segs;
	  fc->segments(segs, -2, -1, 1, 2);
	  gc->segments(segs, -2, -1, 1, 2);
	  
	  Acurve *curveA, *curveB;
	  vector<Point> croots;
	  int ic;
	  for (ic = 0; ic < fc->curvs.size(); ic++)
	    if (fc->curvs[ic] == gc)
	      break;
	  if (ic != fc->curvs.size()) {
	    croots = fc->croots[ic];
	    curveA = fc;
	    curveB = gc;
	  }
	  else {
	    for (ic = 0; ic < gc->curvs.size(); ic++)
	      if (gc->curvs[ic] == fc)
		break;
	    if (ic != gc->curvs.size()) {
	      croots = gc->croots[ic];
	      curveA = gc;
	      curveB = fc;
	    }
	    else
	      return;
	  }
	  
	  for (int ir = 0; ir < croots.size(); ir++) {
	    Point p = croots[ir];
	    Point p0(p.x - 0.1, p.y);
	    Line *line = Line::alloc(p0, p);
	    segs.push_back(Segment(p0, p, line));
	  }
	  
	  extern void segdump (const char *filename, const vector<Segment> &segs);

	  Psystem psys(curveA->f, curveB->f);
	  double rootsxy[4*psys.degree()];
	  // double rootsxy[100];
	  int nr = psys.roots(rootsxy);
	  cout << nr << "roots" << endl;
	  for (int ir = 0; ir < nr; ir++) {
	    Point p(rootsxy[4*ir], rootsxy[4*ir+2]);
	    cout << p << endl;
	    Point p1(p.x + 0.1, p.y);
	    Line *line = Line::alloc(p, p1);
	    segs.push_back(Segment(p, p1, line));
	  }

	  segdump("missing.fig", segs);
	  cout << curveA->f << endl;
	  cout << curveB->f << endl;
	  cout << "num FG " << Timing::numFG << endl;
	  dumped = true;
	  // roots were generated by a solving
	  // Psystem psys(curveA->f, curveB->f);
	}
#endif
	cout << "this is it!" << endl;
	maxfg = 0;
      }
    }
  }
}

bool inrootlistSimple (Function *f, Function *g, const vector<Point> &roots,
		       int &sense)
{
  if (roots.empty()) {
    sense = g->bboxOrder(f);
    return sense != 0;
  }
  sense = f->slopeOrder(g, roots[0]);
  if (sense == 0) return false;
  int os = sense;
  for (int i = 1; i < roots.size(); ++i) {
    int s = f->slopeOrder(g, roots[i]);
    if (s != - os) return false;
    os = s;
  }
  return true;
}

#ifdef SEGMENT_TEST
Point findroot (Function *f, Function *g, double a, double b) {
  double da = f->getY(a) - g->getY(a);
  double db = f->getY(b) - g->getY(b);
  assert(da * db < 0);
  double c, dc;
  int n = 0;

  while (true) {
    if (n++ % 2 == 0)
      if (fabs(da) < fabs(db))
	c = a + (b - a) * (da / (da - db));
      else
	c = b + (a - b) * (db / (db - da));
    else
      c = (a + b) / 2;
    if (c == a || c == b)
      break;
    dc = f->getY(c) - g->getY(c);
    if (dc == 0)
      break;
    if (da * dc < 0) {
      b = c;
      db = dc;
    }
    else {
      a = c;
      da = dc;
    }
  }

  double ab[2] = { a, b };
  Function *fg[2] = { f, g };
  Point p[4];
  double dist[4];
  int k = 0;
  double distMin = Function::FINFINITY;
  int iMin = 0;

  for (int i = 0; i < 2; i++)
    for (int j = 0; j < 2; j++) {
      p[k] = Point(ab[i], fg[j]->getY(ab[i]));
      dist[k] = ((Abranch *) fg[!j])->curve->distance(p[k]);
      if (dist[k] < distMin) {
	distMin = dist[k];
	iMin = k;
      }
      k++;
    }

  static double maxFindRoot = 0;
  if (dist[iMin] > maxFindRoot) {
    maxFindRoot = dist[iMin];
    cout << "maxFindRoot " << maxFindRoot << endl;
  }

  return p[iMin];
}    
#endif

bool check4 = false;

void inrootlist (Function *f, Function *g, const vector<Point> &roots_,
		 vector<Point> &roots, int &sense) {
  bool check = false, check2 = false, check3 = false;

#ifdef BLEEN
  if (f == badF && g == badG || f == badG && g == badF) {
    cout << "bleen bad f and g" << endl;
    for (int i = 0; i < roots_.size(); i++)
      cout << roots_[i];
    check3 = true;
  }
#endif

  double saveX0, saveX1;

#ifdef BLEEN
  for (int i = 0; i < roots_.size(); i++)
    if (i == 0 || roots_[i].x != roots_[i-1].x)
      roots.push_back(roots_[i]);
#endif

  DOUBLE minX = MAX(f->getMinX(), g->getMinX());
  DOUBLE maxX = MIN(f->getMaxX(), g->getMaxX());

  roots = roots_;
  sort(roots.begin(), roots.end(), xLessThan);
  int nroots = 0;
  for (int i = 0; i < roots.size(); i++)
    if ((i == 0 || roots[i].x != roots[i-1].x) &&
	roots[i].x > minX && roots[i].x < maxX)
      roots[nroots++] = roots[i];
  while (roots.size() > nroots)
    roots.pop_back();  
    
  //if (inrootlistSimple(f, g, roots, sense))
  //return;  
  
  /* old version
  if (minX == -Function::FINFINITY)
    minX = roots.size() == 0 ? MIN(0,maxX)-100 : -2 * fabs(roots[0].x) - 100;
  if (maxX == Function::FINFINITY)
    maxX = roots.size() == 0 ? MAX(0,minX)+100 : 2 * fabs(roots[roots.size()-1].x) + 100;
  */
  double xstep = 1.0;
  if (minX == -Function::FINFINITY && maxX == Function::FINFINITY)
    minX = roots.size() == 0 ? - xstep : -2 * fabs(roots[0].x) - xstep;
  else if (minX == -Function::FINFINITY)
    minX = roots.size() == 0 ? maxX - xstep : -2 * fabs(roots[0].x) - xstep;
  if (maxX == Function::FINFINITY)
    maxX = roots.size() == 0 ? minX + xstep : 2 * fabs(roots[roots.size()-1].x) + xstep;
  assert(minX < maxX);

#ifdef BLEEN
  double fgminX = f->getY(minX) - g->getY(minX);
  double fgmaxX = f->getY(maxX) - g->getY(maxX);
  if (fabs(fgminX) > 1e-3 && fabs(fgmaxX) > 1e-3) {
    if (fgminX * fgmaxX < 0 && roots.size() % 2 == 0 ||
	fgminX * fgmaxX > 0 && roots.size() % 2 != 0)
      check2 = true;
  }
#endif

  static DOUBLE minmaxX = 1;
  if (minmaxX > maxX - minX) {
    minmaxX = maxX - minX;
    // cout << "minmaxX " << minmaxX << endl;
    check = true;
  }

  if (roots.size() > 0) {
    static DOUBLE minr = 1;
    if (minr > roots[0].x - minX) {
      minr = roots[0].x - minX;
      // cout << "minr " << minr << endl;
      check = true;
    }
    static DOUBLE rmax = 1;
    if (rmax > maxX - roots[roots.size()-1].x) {
      rmax = maxX - roots[roots.size()-1].x;
      // cout << "rmax " << rmax << endl;
      check = true;
    }
  }

  while (roots.size() > 0 && roots[0].x == minX)
    roots.erase(roots.begin());
  while (roots.size() > 0 && roots.back().x == maxX)
    roots.pop_back();

  if (roots.size() > 0)
    assert(roots[0].x > minX);

  bool isCircle = (f->getType() == Function::LOWER_CIRCLE ||
		   f->getType() == Function::UPPER_CIRCLE ||
		   g->getType() == Function::LOWER_CIRCLE ||
		   g->getType() == Function::UPPER_CIRCLE);

  DOUBLE x0 = roots.size() == 0 ? (minX + maxX) / 2 : (minX + roots[0].x) / 2;
  DOUBLE f0 = f->getY(x0);
  DOUBLE g0 = g->getY(x0);
  DOUBLE f0g0 = f0 - g0;
  if (isCircle) {
    DOUBLE x0c = roots.size() == 0 ?
      (3 * minX + maxX) / 4 : (3 * minX + roots[0].x) / 4;
    DOUBLE f0c = f->getY(x0c);
    DOUBLE g0c = g->getY(x0c);
    DOUBLE f0g0c = f0c - g0c;
    if (fabs(f0g0c) > fabs(f0g0)) {
      x0 = x0c;
      f0 = f0c;
      g0 = g0c;
      f0g0 = f0g0c;
    }
  }

  saveX1 = saveX0 = x0;

  int n = 0;
  DOUBLE fprev = f0;
  DOUBLE gprev = g0;
  DOUBLE fprevgprev = f0g0;
  //debug
  //n = roots.size();
  //debug
  for (int i = 0; i < roots.size(); i++) {
    DOUBLE rii = i < roots.size() - 1 ? roots[i + 1].x : maxX;
#ifdef SEGMENT_TEST
    if (!((rii > roots[i].x))) {
      Acurve *curveF = ((Abranch *) f)->curve;
      Acurve *curveG = ((Abranch *) g)->curve;
      cout << curveF->f << endl;
      cout << curveG->f << endl;
    }
#endif
    assert(rii > roots[i].x);
    DOUBLE x = (roots[i].x + rii) / 2;
    DOUBLE fi = f->getY(x);
    DOUBLE gi = g->getY(x);
    DOUBLE figi = fi - gi;
    if (isCircle) {
      DOUBLE xc = (3 * roots[i].x + rii) / 4;
      DOUBLE fic = f->getY(xc);
      DOUBLE gic = g->getY(xc);
      DOUBLE figic = fic - gic;
      if (fabs(figic) > fabs(figi)) {
	x = xc;
	fi = fic;
	gi = gic;
	figi = figic;
      }
    }

    if (f0g0 == 0) {
      f0 = fi;
      g0 = gi;
      f0g0 = figi;
      assert(n == 0);
      assert(fprevgprev == 0);

      saveX1 = saveX0 = x;
    }
    
    if (fprevgprev < 0 && figi > 0 || fprevgprev > 0 && figi < 0)
      roots[n++] = roots[i];
    else {
      // cout << "Skipping root!" << endl;
      check = true;
    }

    if (figi != 0) {
      fprev = fi;
      gprev = gi;
      fprevgprev = figi;

      saveX1 = x;
    }
  }
  
  while(roots.size() > n)
    roots.pop_back();
  //
  //debug end
  if (f0g0 == 0) {
    assert(n == 0);

    if (f->getType() == Function::TELOMERE_CURVE &&
	g->getType() == Function::TELOMERE_CURVE) {
#ifdef SEGMENT_TEST
      if (((TCurve *) f)->index < ((TCurve *) g)->index)
	sense = 1;
      else
	sense = -1;
#else
      assert(0);
#endif
    }
    else
      sense = 1;
  }
  else {
    sense = f0g0 < 0 ? 1 : -1;

#ifdef SEARCH_ENDPOINTS
    double maxmin = max(f->getMinX(), g->getMinX());
    if (maxmin > -Function::FINFINITY) {
      double fminx = f->getY(maxmin);
      double gminx = g->getY(maxmin);
      double fminxgminx = fminx - gminx;
      double mid = (maxmin + saveX0) / 2;

      if (fminxgminx * f0g0 < 0) {
	Point p = findroot(f, g, maxmin, saveX0);
	if (p.x != maxmin) {
	  roots.insert(roots.begin(), p);
	  sense = -sense;
	}
      }
    }

    double minmax = min(f->getMaxX(), g->getMaxX());
    if (minmax < Function::FINFINITY) {
      double fmaxx = f->getY(minmax);
      double gmaxx = g->getY(minmax);
      double fmaxxgmaxx = fmaxx - gmaxx;
      double mid = (minmax + saveX1) / 2;
      
      if (fmaxxgmaxx * fprevgprev < 0) {
	Point p = findroot(f, g, saveX1, minmax);
	if (p.x != minmax)
	  roots.push_back(p);
      }
    }
#endif
  }

#ifdef BLEEN
  {
    double maxmin = max(f->getMinX(), g->getMinX());
    double minmax = min(f->getMaxX(), g->getMaxX());
    
    if (maxmin > -Function::FINFINITY) {
      double fminx = f->getY(maxmin);
      double gminx = g->getY(maxmin);
      double dMin = sense * (fminx - gminx);
      static double maxDMin = 0;
      if (dMin > maxDMin) {
	maxDMin = dMin;
	cout << "maxDMin " << maxDMin << endl;
#ifdef BLEEN
	if (sense < 0)
	  checkRootList(g, f, roots);
	else if (sense > 0)
	  checkRootList(f, g, roots);
#endif
      }
    }
    
    if (minmax < Function::FINFINITY) {
      double fmaxx = f->getY(minmax);
      double gmaxx = g->getY(minmax);
      double dMax =
	sense * (roots.size() % 2 == 0 ? fmaxx - gmaxx : gmaxx - fmaxx);
      static double maxDMax = 0;
      if (dMax > maxDMax) {
	maxDMax = dMax;
	cout << "maxDMax " << maxDMax << endl;
#ifdef BLEEN
	if (sense < 0)
	  checkRootList(g, f, roots);
	else if (sense > 0)
	  checkRootList(f, g, roots);
#endif
      }
    }
  }
#endif

#ifdef CONSISTENT_MAXMIN
  double maxmin = max(f->getMinX(), g->getMinX());
  double minmax = min(f->getMaxX(), g->getMaxX());
  
  if (maxmin > -Function::FINFINITY) {
    double fminx = f->getY(maxmin);
    double gminx = g->getY(maxmin);
    
    if (sense * (fminx - gminx) > 0) {
      double newx = nextafter(maxmin, maxmin + 1);
      assert(maxmin < newx && newx <= minmax);
      if (newx == minmax)
	assert(roots.size() == 0);
      else if (roots.size() > 0 && roots[0].x == newx)
	roots.erase(roots.begin());
      else {
	double fnewx = f->getY(newx);
	double gnewx = g->getY(newx);
	double fgnewx = fabs(fnewx - gnewx);
	static double maxFGnewX;
	if (maxFGnewX < fgnewx) {
	  maxFGnewX = fgnewx;
	  cout << "maxFGnewX " << maxFGnewX << endl;
	}
	Timing::newroot++;
	
	double dfnewx = f->getDY(newx);
	double dgnewx = g->getDY(newx);
	double newy = fabs(dfnewx) < fabs(dgnewx) ? fnewx : gnewx;
	roots.insert(roots.begin(), Point(newx, newy));
      }
      sense = -sense;
    }
  }
#endif  

  if (maxX - minX > 1e-3)
    check = true;
  else
    check = false;
  check2 = false;
  // check2 = true;
  if (check && check4 || check2 || check3) {
    if (sense < 0)
      checkRootList(g, f, roots);
    else if (sense > 0)
      checkRootList(f, g, roots);
    else
      assert(n == 0);
  }
}

RootList::RootList (Function *f, Function *g) {
  func[0] = f;
  func[1] = g;
  sense = 666;
}  

void report (RootList *this1, RootList *this2) {
  assert(this1 == this2);
}

RootList::RootList (Function *f, Function *g,
		    const vector<Point> &roots, int sense)
  : roots(roots), sense(sense) {
  func[0] = f;
  func[1] = g;
#ifdef LINKED_LIST_ROOTLISTS
  next[0] = f->head[0];
  next[1] = g->head[1];
  f->head[0] = this;
  g->head[1] = this;
#else
#ifdef SETP_ROOTLISTS
  RootList *this2 = f->lists.insert(this);
  if (this != this2)
    report(this, this2);
#else
  f->lists.insert(this, true);
  g->lists.insert(this, true);
#endif
#endif
}  

RootList::RootList (Function *f, Function *g, const vector<Point> &roots_) {
  Timing::numRootList++;

  sense = 0;
  assert(f < g);
  func[0] = f;
  func[1] = g;

  if (f->getMaxX() <= g->getMinX()) {
    if (f->getY(f->getMaxX()) > g->getY(g->getMinX()))
      sense = 1;
    else
      sense = -1;
    return;
  }
  if (f->getMinX() >= g->getMaxX()) {
    if (f->getY(f->getMinX()) > g->getY(g->getMaxX()))
      sense = 1;
    else
      sense = -1;
    return;
  }

#ifdef LAZY_ROOTLIST
  sense = 17;
  for (int i = 0; i < roots_.size(); i++)
    roots.push_back(roots_[i]);
#else
  inrootlist(f, g, roots_, roots, sense);
#endif

#ifdef LINKED_LIST_ROOTLISTS
  next[0] = f->head[0];
  next[1] = g->head[1];
  f->head[0] = this;
  g->head[1] = this;
#else
#ifdef SETP_ROOTLISTS
  RootList *this2 = f->lists.insert(this);
  if (this != this2)
    report(this, this2);
#else
  f->lists.insert(this, true);
  g->lists.insert(this, true);
#endif
#endif
}

RootList::~RootList () {
  if (func[0] == 0)
    return;
  if (sense == 666)
    return;

  Timing::numRootList--;

#ifdef LINKED_LIST_ROOTLISTS
  RootList *&in0 = func[0]->find(func[1]);
  assert(in0 == this);
  RootList *&in1 = func[1]->find(func[0]);
  assert(in1 == this);
  in0 = in0->next[0];
  in1 = in1->next[1];
#else
#ifdef SETP_ROOTLISTS
  assert(this == func[0]->lists.find(this));
  func[0]->lists.remove();
#else
  func[0]->lists.remove(this, true);
  func[1]->lists.remove(this, true);
#endif
#endif
}

void RootList::build () {
  Function *f = func[0];

  // if (f->getType() == Function::TELOMERE_CURVE)
  // Timing::begin(Timing::compare);

  Function *g = func[1];
  vector<Point> roots_;
  roots_ = roots;
  roots.clear();
  sense = 0;
  inrootlist(f, g, roots_, roots, sense);

  if (f->getType() == Function::TELOMERE_CURVE) {
    assert(g->getType() == Function::TELOMERE_CURVE);
#ifdef SEGMENT_TEST
    TCurve *tf = (TCurve *) f;
    TCurve *tg = (TCurve *) g;
    double maxt = max(tf->tail.x, tg->tail.x);
    for (int i = 0; i < roots.size(); i++)
      if (roots[i].x > maxt) {
	Timing::N += roots.size() - i;
	break;
      }
    // Timing::end(Timing::compare);
#else
    assert(0);
#endif
  }
}

void RootList::makeTangent () {
  roots.clear();
  if (func[0]->getType() == Function::LINE &&
      func[1]->getType() == Function::LINE)
    sense = 17;
  else
    sense = 17;
}

int RootList::compare (DOUBLE x, int ler) {
  assert(sense != -17);
  if (sense == 17)
    build();

  int ix;
  if (ler <= 0)
    for (ix = 0; ix < roots.size() && x > roots[ix].x; ix++);
  else
    for (ix = 0; ix < roots.size() && x >= roots[ix].x; ix++);

  if (ler == 0 && ix < roots.size() && x == roots[ix].x)
    return 0;
  else if (ix % 2 == 0)
    return -sense;
  else
    return sense;
}

DOUBLE RootList::nextRoot (DOUBLE x) {
  assert(sense != -17);
  if (sense == 17)
    build();

  for (int i = 0; i < roots.size(); i++)
    if (x < roots[i].x)
      return roots[i].x;
  return Function::FINFINITY;
}

DOUBLE RootList::nearestRoot (DOUBLE x) {
  assert(sense != -17);
  if (sense == 17)
    build();

  if (roots.size() == 0)
    return Function::FINFINITY;

  int i;
  for (i = 0; i < roots.size(); i++)
    if (x < roots[i].x)
      break;

  if (i == 0)
    return roots[i].x;

  if (i == roots.size())
    return roots[i-1].x;

  if (x - roots[i-1].x < roots[i].x - x)
    return roots[i-1].x;
  else
    return roots[i].x;
}

Point RootList::getRoot (DOUBLE x) {
  assert(sense != -17);
  if (sense == 17)
    build();

  for (int i = 0; i < roots.size(); i++)
    if (roots[i].x == x)
      return roots[i];
  return Point(Function::FINFINITY, Function::FINFINITY);
}

void RootList::flip (Point p, Point q) {
  assert(sense != -17);
  if (sense == 17)
    build();

  assert(nextRoot(p.x) >= q.x);
  vector<Point>::iterator it;
  for (it = roots.begin(); it != roots.end() && p.x > it->x; it++);
  if (it != roots.end() && p.x == it->x)
    roots.erase(it);
  else
    roots.insert(it, p);
  for (it = roots.begin(); it != roots.end() && q.x > it->x; it++);
  if (it != roots.end() && q.x == it->x)
    roots.erase(it);
  else
    roots.insert(it, q);
}

int numFunction;

Function::Function () {
  refcount = 0;
#ifdef LINKED_LIST_ROOTLISTS
  head[0] = head[1] = 0;
#endif
  numFunction++;
  // cout << "new Function " << this << endl;
}

Function::~Function () {
  clearRootLists();
  numFunction--;
  // cout << "del Function " << this << endl;
}

DOUBLE Function::distance (Point p) const {
  DOUBLE dy = fabs(p.y - getY(p.x));
  DOUBLE s = getDY(p.x);
  DOUBLE d = dy / sqrt(1 + s * s);
  return d;
}

RootList *Function::find (Function *that) {
  assert(this != that);
#ifdef LINKED_LIST_ROOTLISTS
  bool ind = (this > that);
  RootList **p;
  Timing::begin(Timing::find);
  for (p = &head[ind]; *p != 0; p = &(*p)->next[ind])
    if ((*p)->func[ind] == this && (*p)->func[!ind] == that)
      break;
  Timing::end(Timing::find);
  return *p;
#else
#ifdef SETP_ROOTLISTS
  Timing::begin(Timing::find);
  Function *f = this < that ? this : that;
  Function *g = this > that ? this : that;
  RootList rl(0, g);
  RootList *found = f->lists.find(&rl);
  assert(found == 0 || found->func[0] == f && found->func[1] == g);
  Timing::end(Timing::find);
  return found;
#else
  Timing::begin(Timing::find);
  Function *f = this < that ? this : that;
  Function *g = this > that ? this : that;
  RootList rl(f, g);
  RootList *found = f->lists.find(&rl);
  assert(found == 0 || found->func[0] == f && found->func[1] == g);
  Timing::end(Timing::find);
  return found;
#endif
#endif
}

void Function::clearRootLists () {
#ifdef LINKED_LIST_ROOTLISTS
  while (head[0] != 0)
    delete head[0];
  while (head[1] != 0)
    delete head[1];
#else
  while (lists.first()) {
    // lists.first()->func[1]->clearRootLists();
    delete lists.first();
  }
#endif
}  

RootList *Function::getRootList (Function *that) {
  assert(this < that);
  RootList *rl = find(that);
  if (rl == 0) {
    if (this->overlaps(that)) {
      Timing::begin(Timing::rootlist);
      rl = makeRootList(that);
      Timing::end(Timing::rootlist);
    }
    else {
      vector<Point> roots;
      if (this < that)
	rl = new RootList(this, that, roots);
      else
	rl = new RootList(that, this, roots);
    }
  }
  return rl;
}  

void Function::getRootLists (vector <RootList *> &res) {
#ifdef LINKED_LIST_ROOTLISTS
  for (RootList *rl = head[0]; rl != 0; rl = rl->next[0])
    res.push_back(rl);
#else
  for (lists.first(); lists.curr(); lists++)
    res.push_back(lists.curr());
#endif
}

void Function::makeTangent (Function *that) {
  if (this == that)
    return;
  if (this > that)
    that->makeTangent(this);
  else {
    RootList *rl = getRootList(that);
    rl->makeTangent();
  }
}

int Function::compare (Function *that, DOUBLE x, int ler) {
  if (this == that)
    return 0;
  if (this > that)
    return -that->compare(this, x, ler);
  
#ifdef BLEEN
  int bo = bboxOrder(that);
  if (bo) return bo;
  
  if (this->getMaxX() <= that->getMinX()) {
    if (this->getY(this->getMaxX()) < that->getY(that->getMinX()))
      return -1;
    else if (this->getY(this->getMaxX()) > that->getY(that->getMinX()))
      return 1;
    else
      return 0;
  }
  if (that->getMaxX() <= this->getMinX()) {
    if (that->getY(that->getMaxX()) < this->getY(this->getMinX()))
      return -1;
    else if (that->getY(that->getMaxX()) > this->getY(this->getMinX()))
      return 1;
    else
      return 0;
  }
#endif

  RootList *rl = getRootList(that);
  return rl->compare(x, ler);
}

int Function::cheap_compare (Function *that, DOUBLE x, int ler) {
  assert(ler == 1);
  assert(this != that);
  assert(this->overlaps(that));
  if (this > that)
    return -that->cheap_compare(this, x, ler);
  
  RootList *rl = find(that);
  if (rl != 0)
    return rl->compare(x, ler);

  DOUBLE thisY = this->getY(x);
  DOUBLE thatY = that->getY(x);
  if (thisY < thatY)
    return -1;
  else if (thisY > thatY)
    return 1;

  DOUBLE thisDY = this->getDY(x);
  DOUBLE thatDY = that->getDY(x);
  DOUBLE thisDY2 = this->getDY(x);
  DOUBLE thatDY2 = that->getDY(x);
  assert(thisDY == thisDY2 && thatDY == thatDY2);
  if (thisDY < thatDY)
    return -1;
  if (thisDY > thatDY)
    return 1;
  return 0;
}

DOUBLE Function::nextRoot (Function *that, DOUBLE x) {
  if (!this->overlaps(that))
    return FINFINITY;
  if (this == that)
    return FINFINITY;
  if (this > that)
    return that->nextRoot(this, x);
  
  RootList *rl = getRootList(that);
  return rl->nextRoot(x);
}

DOUBLE Function::nearestRoot (Function *that, DOUBLE x) {
  if (this == that)
    return FINFINITY;
  if (this > that)
    return that->nearestRoot(this, x);
  
  RootList *rl = getRootList(that);
  return rl->nearestRoot(x);
}

Point Function::getRoot (Function *that, DOUBLE x) {
  if (this == that)
    return Point(Function::FINFINITY, Function::FINFINITY);
  if (this > that)
    return that->getRoot(this, x);
  
  RootList *rl = find(that);
  if (rl != 0)
    return rl->getRoot(x);
  else
    return Point(Function::FINFINITY, Function::FINFINITY);
}

double thismin, thismax, thatmin, thatmax;

bool checkSegment = false;

void Function::check (Point p) {
  if (!checkSegment)
    return;

  DOUBLE temp;
  static DOUBLE maxMinX;
  if (maxMinX < (temp = getMinX() - p.x)) {
    maxMinX = temp;
    cout << "maxMinX " << maxMinX << endl;
  }
  
  static DOUBLE maxMaxX;
  if (maxMaxX < (temp = p.x - getMaxX())) {
    maxMaxX = temp;
    cout << "maxMaxX " << maxMaxX << endl;
  }
  
  static DOUBLE maxDist;
  temp = distance(p);
  if (temp > 1.000001 * maxDist) {
    maxDist = temp;
    cout << "maxDist " << maxDist << endl;
    maxDist = temp;
  }

  if (maxDist > 1e-3) {
#ifdef BLEEN
    extern Function *saveLoF;
    cout << this << " " << saveLoF << endl;
    double y = getY(p.x);
    cout << y << " - " << p.y << " = " << y - p.y << endl;
    cout << getMinX() << " " << getMaxX() << endl;
    thismin = getMinX();
    thismax = getMaxX();
    if (saveLoF != 0) {
      int s = compare(saveLoF, p.x, 1);
      double y2 = saveLoF->getY(p.x);
      cout << s << " " << y2 << endl;
      cout << saveLoF->getMinX() << " " << saveLoF->getMaxX() << endl;
      thatmin = saveLoF->getMinX();
      thatmax = saveLoF->getMaxX();
    }
#endif
    cout << "maxDist:  this is it!" << endl;
    report2(maxDist);
    // assert(0);
    maxDist = 0;
  }
}

void approximate (Function *f, Point tail, Point head, vector<Point> &points) {
  if (tail.x == head.x) {
    assert(f == 0);
    points.push_back(tail);
    points.push_back(head);
    return;
  }

  if (fabs(head.x - tail.x) < 2.0 / 1200) {
    points.push_back(tail);
    points.push_back(head);
    return;
  }

  Point mid = (tail + head) / 2;
  DOUBLE y = f->getY(mid.x);
  if (fabs(y - mid.y) < 0.5 / 1200) {
    points.push_back(tail);
    points.push_back(head);
    return;
  }

  mid.y = y;
  approximate(f, tail, mid, points);
  points.pop_back();
  approximate(f, mid, head, points);
}

void drawLine (ostream &out, int depth, Point tail, Point head) {
  out << "2 1 0 1 0 7 " << depth
      << " 0 -1 0.000 0 0 -1 0 0 2" << endl;
  out << "\t " << IP(tail.conjugate() * 1200)
      << " " << IP(head.conjugate() * 1200) << endl;
}

void drawSpline (ostream &out, int depth, const vector<Point> &points) {
  out << "3 2 0 1 0 7 " << depth
      << " 0 -1 0.000 0 0 0 " << points.size() << endl;
  out << "\t";
  for (int i = 0; i < points.size(); i++)
    out << " " << IP(points[i].conjugate() * 1200);
  out << endl;
  out << "\t 0.000";
  for (int i = 0; i < points.size() - 2; i++)
    out << " -1.000";
  out << " 0.000" << endl;
}

void segdump (const char *filename, const vector<Segment> &segs) {
#ifdef NODUMP
  return;
#endif
  // cout << "start segdump " << filename << endl;
  ofstream out;
  out.open(filename);

  out << "#FIG 3.2" << endl
      << "Landscape" << endl
      << "Center" << endl
      << "Inches" << endl
      << "Letter  " << endl
      << "100.00" << endl
      << "Single" << endl
      << "-2" << endl
      << "1200 2" << endl;

  vector<Point> points;
  for (int i = 0; i < segs.size(); i++) {
    const Segment &seg = segs[i];
    points.clear();
    approximate(seg.getF(), seg.getTail(), seg.getHead(), points);
    out << "2 1 0 1 0 7 50 0 -1 0.000 0 0 -1 0 0 " << points.size()
	<< endl << "\t";
    for (int j = 0; j < points.size(); j++)
      out << " " << IP(points[j].conjugate() * 1200);
    out << endl;
  }

  out.close();
  // cout << "end segdump " << filename << endl;
}  

