char copyright1[] = "copyright (c) 2008 Victor J. Milenkovic, not for commercial use without permission";

#include "Embedding.h"
#include "Circle.h"
#include "curve.h"
#include <string>
#include <fstream>

using namespace std;

// Smart pointer class.
typedef Ptr<Embedding> RegionPtr;

// Read xfig file containing single closed loop polygon and create
// regions from it.
RegionPtr readPolyFig (const char *filename) {
  ifstream in;
  in.open(filename);
  char buf[1024];
  for (int i = 0; i < 26; i++)
    in >> buf;
  int n;
  in >> n;
  Point points[n];
  for (int i = 0; i < n; i++) {
    in >> points[i];
    cout << points[i] << endl;
    points[i] = points[i] / 1200;
  }

  vector<Segment> segs;
  n--;
  for (int i = 0; i < n; i++)
    segs.push_back(Segment(points[i], points[i+1],
			   Line::alloc(points[i], points[i+1])));
  
  return new Embedding(segs);
}  

// Calculate bounding box of regions.
double maxXY = 0;

// Read shape bounded by a loop of arcs and line segments and create
// region from it.
// First line in file is number of segments.
// Each subsequent line describes a line segment or circular arc.
// Line segment format is
//     0 tailX tailY headX headY insideOnLeft
// where insideOnLeft is 0/1 if inside is on right/left.
// Arc segment format is
//     0 tailX tailY headX headY insideOnLeft centerX centerY radius
RegionPtr readPolyArc (const char *filename) {
  fstream in;
  in.open(filename);
  if (!in) {
    cout << "could not open " << filename << endl;
    return 0;
  }

  vector<Segment> segs;
  int n;
  in >> n;
  for (int i = 0; i < n; i++) {
    int linecircle, rightleft;
    Point tail, head;
    in >> linecircle >> tail >> head >> rightleft;

    if (maxXY < fabs(tail.x))
      maxXY = fabs(tail.x);
    if (maxXY < fabs(tail.y))
      maxXY = fabs(tail.y);

    if (linecircle == 0) {
      Line *line = tail.x == head.x ? 0 : Line::alloc(tail, head);
      if (rightleft)
	segs.push_back(Segment(tail, head, line));
      else
	segs.push_back(Segment(head, tail, line));
      continue;
    }

    Point center;
    double radius;
    in >> center >> radius;
    center = center;
    radius = radius;

    // improve accuracy
    // (mc + s th) * th = 0
    // s = -mc * th / th * th
    Point mid = (tail + head) / 2;
    Point mc = center - mid;
    Point th = head - tail;
    double s = -(mc * th) / (th * th);
    center = center + th * s;
    radius = (center - tail).length();

    Circle *circle = Circle::alloc(center, radius);
    if (tail.y < center.y || tail.y == center.y && tail.x < center.x) {
      if (head.y <= center.y) {
	assert(tail.x < head.x);
	if (rightleft)
	  segs.push_back(Segment(tail, head, circle->getBranch(0)));
	else
	  segs.push_back(Segment(head, tail, circle->getBranch(0)));
      }
      else {
	Point right(center.x + radius, center.y);
	if (rightleft) {
	  segs.push_back(Segment(tail, right, circle->getBranch(0)));
	  segs.push_back(Segment(right, head, circle->getBranch(1)));
	}
	else {
	  segs.push_back(Segment(head, right, circle->getBranch(1)));
	  segs.push_back(Segment(right, tail, circle->getBranch(0)));
	}
      }
    }
    else {
      if (head.y >= center.y) {
	assert(tail.x > head.x);
	if (rightleft)
	  segs.push_back(Segment(tail, head, circle->getBranch(1)));
	else
	  segs.push_back(Segment(head, tail, circle->getBranch(1)));
      }
      else {
	Point left(center.x - radius, center.y);
	if (rightleft) {
	  segs.push_back(Segment(tail, left, circle->getBranch(1)));
	  segs.push_back(Segment(left, head, circle->getBranch(0)));
	}
	else {
	  segs.push_back(Segment(head, left, circle->getBranch(0)));
	  segs.push_back(Segment(left, tail, circle->getBranch(1)));
	}
#ifdef BLEEN
	if (rightleft)
	  segs.push_back(Segment(tail, left, circle->getBranch(1)));
	else
	  segs.push_back(Segment(left, tail, circle->getBranch(1)));
	if (rightleft)
	  segs.push_back(Segment(left, head, circle->getBranch(0)));
	else
	  segs.push_back(Segment(head, left, circle->getBranch(0)));
#endif
      }
    }
  }

  for (int i = 0; i < segs.size(); i++) {
    Segment &s = segs[i];
    Segment &t = segs[(i+1)%segs.size()];
    assert(s.getHead() == t.getTail());
  }

  cout << "data maxXY " << maxXY << endl;

  return new Embedding(segs);
}

RegionPtr makeCircle (double r) {
  Circle *circle = Circle::alloc(Point(0, 0), r);
  vector<Segment> segs;
  segs.push_back(Segment(Point(-r, 0), Point(r, 0), circle->getBranch(0)));
  segs.push_back(Segment(Point(r, 0), Point(-r, 0), circle->getBranch(1)));
  return new Embedding(segs);
}  

RegionPtr makeSquare (double r) {
  Point ps[] = { Point(-r, -r), Point(r, -r), Point(r, r), Point(-r, r) };
  vector<Segment> segs;
  for (int i = 0; i < 4; i++) {
    Point a = ps[i];
    Point b = ps[(i+1)%4];
    Line *line = a.x == b.x ? 0 : Line::alloc(a, b);
    segs.push_back(Segment(a, b, line));
  }
  return new Embedding(segs);
}    

RegionPtr makeRectangle (Point bbox[2]) {
  Point ps[] = { bbox[0], Point(bbox[1].x, bbox[0].y),
		 bbox[1], Point(bbox[0].x, bbox[1].y) };
  vector<Segment> segs;
  for (int i = 0; i < 4; i++) {
    Point a = ps[i];
    Point b = ps[(i+1)%4];
    Line *line = a.x == b.x ? 0 : Line::alloc(a, b);
    segs.push_back(Segment(a, b, line));
  }
  return new Embedding(segs);
}    

// Calculate a point inside a region.
Point pointInside (RegionPtr s) {
  Timing::begin(Timing::pointInside);
  Point p = s->pointInside();
  Timing::end(Timing::pointInside);
  return p;
}

// Complement of a (+) -b.
RegionPtr calcU (RegionPtr a, RegionPtr b) {
  Transformation rot180(Point(0, 0), Point(-1, 0));
  b = new Embedding(b, rot180);
  RegionPtr u = a->sum(b);
  u->complement();
  return u;
}

// Union of a and b.
RegionPtr doUnion (RegionPtr a, RegionPtr b) {
  vector<Embedding *> sets(2);
  sets[0] = a;
  sets[1] = b;
  vector<int> cn(2);
  cn[0] = a->cells.at(0).state;
  cn[1] = b->cells.at(0).state;
  return new Embedding(sets, cn, Embedding::Union());
}

// Intersection of a and b.
RegionPtr doIntersect (RegionPtr a, RegionPtr b) {
  vector<Embedding *> sets(2);
  sets[0] = a;
  sets[1] = b;
  vector<int> cn(2);
  cn[0] = a->cells.at(0).state;
  cn[1] = b->cells.at(0).state;
  return new Embedding(sets, cn, Embedding::Intersection());
}

// Place regions[0] and regions[1] without overlapping.
// The idea is that region[0] is an unbounded complement of a
// container region, and hence the name ``fit1''.
// Return 1 if it cannot be done.
// Return 0 if it can.
// tr[i] is set to valid translation of region[i], i=0,1
int fit1 (RegionPtr regions[2], Point tr[2]) {
  RegionPtr u01 = calcU(regions[0], regions[1]);
  if (u01->isNull())
    return 1;
  tr[1] = pointInside(u01);
  return 0;
}

// Root of file name used by drawAll.
char convName[1024] = "conv";

// Draw regions placed at translations in tr to file with root in convName
// and unique integer id.
// Output is in xfig format.
void drawall (RegionPtr *regions, Point *tr, int n) {
  static int count;
  vector<Embedding *> vec(n);
  vector<int> cn(n);
  for (int i = 0; i < n; i++) {
    vec[i] = new Embedding(regions[i], Transformation(tr[i]));
    // sprintf(buf, "%s%02d-%d.fig", convName, count, i);
    // vec[i]->edgedraw(buf);
    cn[i] = vec[i]->cells.at(0).state;
  }
  Embedding all(vec, cn, Embedding::Union());
  char buf[1024] = "foo";
  sprintf(buf, "%s%02d.fig", convName, count++);
  cout << buf << endl;
  all.edgedraw(buf);
}

// Place 3 regions (2 regions into a container) without overlapping.
// Return 0 if successful.
// Return 1, 2, or 3 depending on where (exact) algorithm failed.
// Return -1 for failure because of numerical inconsistency.
int fit2 (RegionPtr regions[3], Point tr[3]) {
  RegionPtr u01 = calcU(regions[0], regions[1]);
  if (u01->isNull())
    return 1;
  // u01->edgedraw("u01.fig");
  RegionPtr u02 = calcU(regions[0], regions[2]);
  if (u02->isNull())
    return 2;
  // u02->edgedraw("u02.fig");
  RegionPtr u21 = calcU(regions[2], regions[1]);
  // u21->edgedraw("u21.fig");
  RegionPtr u02u21 = u02->sum(u21);
  // u02u21->edgedraw("u02u21.fig");
  u01 = doIntersect(u01, u02u21);
  if (u01->isNull())
    return 3;
  // u01->edgedraw("u01b.fig");
  tr[1] = pointInside(u01);
  Transformation t1r(tr[1], Point(-1, 0));
  RegionPtr t1u12 = new Embedding(u21, t1r);
  u02 = doIntersect(u02, t1u12);
  // u02->edgedraw("u02.fig");
  if (u02->isNull())
    return -1;
  tr[2] = pointInside(u02);
  return 0;
}

// Place 4 regions (3 regions into a container) without overlapping.
// region[0] must be the complement of a rectangle.
// Return 0 if successful.
// Return positive integer for failure of exact algorithm.
// Return negative integer for numerical inconsistency.
int fit3 (RegionPtr regions[4], Point tr[4]) {
  Point bboxj[4][2];
  for (int j = 0; j < 4; j++)
    regions[j]->getBBox(bboxj[j]);

  Point bbox0j[4][2];
  for (int j = 1; j < 4; j++) {
    bbox0j[j][0] = bboxj[0][0] - bboxj[j][0];
    bbox0j[j][1] = bboxj[0][1] - bboxj[j][1];
    if (bbox0j[j][0].x >= bbox0j[j][1].x || bbox0j[j][0].y >= bbox0j[j][1].y)
      return j;
  }

  RegionPtr u[4][4];
  for (int i = 1; i < 3; i++)
    for (int j = i+1; j < 4; j++) {
      Point bbox[2] = { -bbox0j[i][1] + bbox0j[j][0],
			-bbox0j[i][0] + bbox0j[j][1] };
      RegionPtr rect = makeRectangle(bbox);
      u[i][j] = calcU(regions[i], regions[j]);
      u[i][j] = doIntersect(u[i][j], rect);
      if (u[i][j]->isNull())
	return 1 + i + j;
    }

  // u[1][2]->edgedraw("u12.fig");
  // u[2][3]->edgedraw("u23.fig");
  RegionPtr u12u23 = u[1][2]->sum(u[2][3]);
  // u12u23->edgedraw("u12u23.fig");
  RegionPtr u13old = u[1][3];
  u[1][3] = doIntersect(u[1][3], u12u23);
  if (u[1][3]->isNull())
    return 7;
  
  // u[1][3]->edgedraw("u13.fig");
  Point tr13 = pointInside(u[1][3]);

  {
    Point bbox[2] = { bbox0j[3][0] - tr13, bbox0j[3][1] - tr13 };
    if (bbox0j[1][0].x < bbox[0].x)
      bbox0j[1][0].x = bbox[0].x;
    if (bbox0j[1][0].y < bbox[0].y)
      bbox0j[1][0].y = bbox[0].y;
    if (bbox0j[1][1].x > bbox[1].x)
      bbox0j[1][1].x = bbox[1].x;
    if (bbox0j[1][1].y > bbox[1].y)
      bbox0j[1][1].y = bbox[1].y;
  }
  if (bbox0j[1][0].x > bbox0j[1][1].x || bbox0j[1][0].y > bbox0j[1][1].y) {
    Point bbox13old[2];
    u13old->getBBox(bbox13old);
    cout << "bbox13old "
	 << bbox13old[0].x << " " << bbox13old[0].y << " "
	 << bbox13old[1].x << " " << bbox13old[1].y << endl;
    // u13old->edgedraw("u13old.fig");
    // cout << "u13old outermult " << u13old->getOuterState() << endl;

    Point bbox12u23[2];
    u12u23->getBBox(bbox12u23);
    // cout << "bbox12u23 " << bbox12u23[0] << " " << bbox12u23[1] << endl;
    // u12u23->edgedraw("u12u23.fig");
    // cout << "u12u23 outermult " << u12u23->getOuterState() << endl;

    Point bbox13[2];
    u[1][3]->getBBox(bbox13);
    // cout << "bbox13 " << bbox13[0] << " " << bbox13[1] << endl;
    // u[1][3]->edgedraw("u13.fig");

    cout << "tr13 " << tr13 << endl;
    u[1][3] = doIntersect(u13old, u12u23);
    u[1][3]->getBBox(bbox13);
    return -1;
  }

  {
    Transformation tr13r(tr13, Point(-1, 0));
    RegionPtr tr13u32 = new Embedding(u[2][3], tr13r);
    u[1][2] = doIntersect(u[1][2], tr13u32);
  }
  if (u[1][2]->isNull())
    return -2;
  
  {
    Point bbox12[2] = { -bbox0j[1][1] + bbox0j[2][0],
			-bbox0j[1][0] + bbox0j[2][1] };
    RegionPtr rect12 = makeRectangle(bbox12);
    u[1][2] = doIntersect(u[1][2], rect12);
  }
  if (u[1][2]->isNull())
    return -3;

  // u[1][2]->edgedraw("u12.fig");
  Point tr12 = pointInside(u[1][2]);
  {
    Point bbox[2] = { bbox0j[2][0] - tr12, bbox0j[2][1] - tr12 };
    if (bbox0j[1][0].x < bbox[0].x)
      bbox0j[1][0].x = bbox[0].x;
    if (bbox0j[1][0].y < bbox[0].y)
      bbox0j[1][0].y = bbox[0].y;
    if (bbox0j[1][1].x > bbox[1].x)
      bbox0j[1][1].x = bbox[1].x;
    if (bbox0j[1][1].y > bbox[1].y)
      bbox0j[1][1].y = bbox[1].y;
  }
  if (bbox0j[1][0].x > bbox0j[1][1].x || bbox0j[1][0].y > bbox0j[1][1].y)
    return -4;
    
  tr[1] = (bbox0j[1][0] + bbox0j[1][1]) / 2;
  tr[2] = tr[1] + tr12;
  tr[3] = tr[1] + tr13;

  return 0;
}

// 0: kinetic, 1: monotonic
int Embedding::convolutionType = 1;

int main (int argc, char *argv[]) {
  // For extreme error checking
  // extern bool check4;
  // check4 = false;

  map<string,string> parms;
  for (int i = 1; i < argc; i++) {
    string arg(argv[i]);
    int iEQ = arg.find("=");
    if (iEQ == -1)
      parms[arg] = "yes";
    else
      parms[arg.substr(0, iEQ)] = arg.substr(iEQ+1);
  }

  vector<Segment> segs;

  // square:  Generate a square.
  if (parms["square"] != "") {
    segs.clear();
    Point ps[] = { Point(-10, -10), Point(10, -10), Point(10, 10), Point(-10, 10) };
    for (int i = 0; i < 4; i++) {
      Point a = ps[i];
      Point b = ps[(i+1)%4];
      Line *line = a.x == b.x ? 0 : Line::alloc(a, b);
      segs.push_back(Segment(a, b, line));
    }
  }

  // circle:  Generate a square-like shape with arcs in its boundary.
  if (parms["circle"] != "") {
    segs.clear();
    vector<Segment> side[4];
    side[0].push_back(Segment(Point(5, -1), Point(3, 1),
			      Circle::alloc(Point(3, -1), 2)->getBranch(1)));
    side[0].push_back(Segment(Point(3, 1), Point(1, 3),
			      Circle::alloc(Point(3, 3), 2)->getBranch(0)));
    side[0].push_back(Segment(Point(1, 3), Point(1, 5), 0));
    
    Transformation rot90(Point(0, 0), Point(0, 1));
    for (int i = 0; i < 3; i++)
      for (int j = 0; j < side[i].size(); j++)
	side[i][j].transform(rot90, side[i+1]);
    
    for (int i = 0; i < 4; i++)
      for (int j = 0; j < side[i].size(); j++)
	segs.push_back(side[i][j]);
    
    Embedding csquare(segs);
    csquare.draw("csquare.fig");
    cout << "csquare area " << csquare.area() << endl;
  }

  // implicit:  Read a polygon from standard input and use it as the
  // side of a ``square''.  The format is:
  //
  // minX maxX minY maxY
  // numVariables numTerms
  // coefficient xExponent yExponent
  // coefficient xExponent yExponent
  // ...
  //
  // where numVariables had better be two!  The first line gives the
  // bounding box.  See cubic.in for an example.
  if (parms["implicit"] != "") {
    segs.clear();
    double bbox[4];
    for (int i = 0; i < 4; i++)
      cin >> bbox[i];
    Polynomial f = read_polynomial(cin);
    Acurve *c = new Acurve(f, bbox);
    c->form_branches(false);
    vector<Segment> side;
    c->segments(side);

    // Rotations to make the sides of the ``square'', roughly 0, 90,
    // 180, and 270.
    DOUBLE q0 = 2 * M_PI / 180;
    DOUBLE q1 = 93 * M_PI / 180;
    DOUBLE q2 = 179 * M_PI / 180;
    DOUBLE q3 = 274 * M_PI / 180;
  
    // Translations, roughly (2,0), (0,2), (-2,0), and (0,-2)
    Point p0 = Point(2.23934, 0.0392847);
    Point p1 = Point(0.34372, 2.98734);
    Point p2 = Point(-2.34289374, 0.0521555);
    Point p3 = Point(-0.03847234, -2.48343434);
    
    // Create transformations
    Transformation t0(p0, q0);
    Transformation t1(p1, q1);
    Transformation t2(p2, q2);
    Transformation t3(p3, q3);
    
    // Create sides by transforming every segment in side by each transformation.
    for (int i = 0; i < side.size(); i++) {
      side[i].transform(t0, segs);
      side[i].transform(t1, segs);
      side[i].transform(t2, segs);
      side[i].transform(t3, segs);
    }
    
    segdump("segs.fig", segs);
  }

  // simple:  Simple test of Euclidean and Boolean operations.
  if (parms["simple"] != "") {
    Embedding square(segs, true);
    square.draw("square.fig");
    Transformation ts(Point(0, 0), Point(0.8, 0.6));
    Embedding squaret(&square, ts);
    squaret.draw("squaret.fig");
    vector<Embedding *> in(2);
    in[0] = (&square);
    in[1] = (&squaret);
    vector<int> cn(2);
    cn[0] = square.isUnbounded();
    cn[1] = squaret.isUnbounded();
    Embedding diff(in, cn, Embedding::XOR());
    diff.draw("diff.fig");
    Embedding setunion(in, cn, Embedding::Union());
    setunion.draw("union.fig");
    Embedding intersection(in, cn, Embedding::Intersection());
    intersection.draw("intersection.fig");
  }

  // torture=n: Repeatedly (n times, default is 6) rotate and XOR with
  // itself, halving the angle each time.
  // angle=d:  Use starting angle of 0.765*2^{-d} radians, default d=0.
  if (parms["torture"] != "") {
    Embedding square(segs, true);
    square.draw("square.fig");
    int n = atoi(parms["torture"].c_str());
    if (n == 0) n = 6;
    cout << "torture = " << n << endl;
    Point t(0.0123, 0.0321);
    double q = 0.765;
    int d = atoi(parms["angle"].c_str());
    cout << "angle = " << d << endl;
    for (int i = 0; i < d; i++) {
      t = t / 2;
      q /= 2;
    }

    Transformation tr(t, Point(cos(q), sin(q)));
    char buf[1024];
    Ptr<Embedding> a = new Embedding(segs, true);
    vector<Embedding *> in(2);
    vector<int> cn(2);
    for (int i = 1; i <= n; i++) {
      a->draw("a.fig");
      Ptr<Embedding> b = new Embedding(a, tr);
      b->draw("b.fig");
      in[0] = a;
      in[1] = b;
      Ptr<Embedding> c = new Embedding(in, cn, Embedding::XOR());
      sprintf(buf, "diff%d.fig", i);
      c->draw(buf);
      cout << "embeddedEndpoints " << Timing::embeddedEndpoints << endl;
      cout << "embeddingViolations " << Timing::embeddingViolations << endl;
      cout << "splitAtInconsistencies " << Timing::splitAtInconsistencies << endl;
      a = c;

      // evil:  For each iteration, XOR region with its ``evil twin''.  
      // Evil twin is created by rotating region and then rotating it
      // back numerically, resulting in a twin that is identical up to
      // rounding error.
      if (parms["evil"] != "") {
	c = new Embedding(a, tr);
	c->draw("c.fig");
	b = new Embedding(c, tr.inverse());
	b->draw("b.fig");
	in[0] = a;
	in[1] = b;
	c = new Embedding(in, cn, Embedding::XOR());
	sprintf(buf, "evil%d.fig", i);
	c->edgedraw(buf);
      }

      t = t / 2;
      q = q / 2;
      tr = Transformation(t, Point(cos(q), sin(q)));
    }

    cout << "numrealroots " << Timing::numrealroots << " "
	 << "timerealroots " << Timing::timerealroots << endl;
  }

  // convolution=kinetic, convolution=monotonic: Type of convolution.
  // Default is monotonic.
  if (parms["convolution"] == "kinetic")
    Embedding::convolutionType = 0;

  // mink1:  Simple test of Minkowski sum.
  if (parms["mink1"] != "") {
    Embedding square(segs);
    square.draw("square.fig");
    Point t(0.0123, 0.0321);
    double q = 0.765;
    Transformation tr(t, q);
    Embedding square2(&square, tr);
    square2.draw("square2.fig");
    Embedding *mink = square.sum(&square2, 1);
    mink->edgedraw("mink.fig");
    cout << "ncells " << mink->cells.size() << endl;
  }

  // poly: Read polygons from poly1.fig and poly2.fig and write their
  // sum to mink.fig.
  if (parms["poly"] != "") {
    Embedding *poly1 = readPolyFig("poly1.fig");
    Embedding *poly2 = readPolyFig("poly2.fig");
    poly2->complement();
    poly1->draw("poly1b.fig");
    poly2->draw("poly2b.fig");
    Embedding *mink = poly1->sum(poly2);
    mink->draw("mink.fig");
    cout << "ncells " << mink->cells.size() << endl;
  }

  // polyarc=in.txt: read polyarc from in.txt and sum with itself.
  if (parms["polyarc"] != "") {
    Embedding *poly = readPolyArc(parms["polyarc"].c_str());
    poly->draw("polyarc.fig");
    Embedding *mink = poly->sum(poly);
    mink->edgedraw("mink.fig");
    cout << "ncells " << mink->cells.size() << endl;
    vector<Embedding::Cell> &cells = mink->cells;
    vector<Embedding::Loop> &loops = mink->loops;
    for (int i = 0; i < cells.size(); i++) {
      Embedding::Cell &cell = cells[i];
      int maxInfo = 0;
      for (int j = 0; j < cell.loops.size(); j++) {
	Embedding::DEdge head = loops[cell.loops[j]].d;
	Embedding::DEdge d = head;
	do {
	  if (maxInfo < d.edge()->info)
	    maxInfo = d.edge()->info;
	} while (++d != head);
      }
      cout << "depth " << cell.depth << " "
	   << "maxInfo " << maxInfo << endl;
    }
  }

  // mtest=n:  read polyarcs from files mtest1, mtest2, ..., mtestn
  // Sum all pairs.
  if (parms["mtests"] != "") {
    char buf[1024];
    Embedding *mtests[6];
    for (int i = 0; i < 6; i++) {
      sprintf(buf, "mtest%d", i+1);
      mtests[i] = readPolyArc(buf);
      sprintf(buf, "mtest%d.fig", i+1);
      mtests[i]->edgedraw(buf);
    }

    for (int i = 0; i < 6; i++)
      for (int j = 0; j < 6; j++) {
	sprintf(convName, "mtest%d%d", i+1, j+1);
	Embedding *mink = mtests[i]->sum(mtests[j]);
	for (int k = 1; k < mink->cells.size(); k++)
	  // assert(mink->cells[k].state == Embedding::INSIDE);
	  if (mink->cells[k].state == Embedding::OUTSIDE)
	    cout << "bubble??" << endl;
	sprintf(buf, "%sS.fig", convName);
	mink->edgedraw(buf);
	if (mink->cells.size() > 2)
	  cout << "this is it" << endl;
	cout << "data " << buf << " s " << mink->edges.size() << endl;
	cout << "time sum " << Timing::sum
	     << " loops " << Timing::sumloops
	     << " arrange " << Timing::sumarrange
	     << " inside " << Timing::pointInside
	     << " adapt " << Timing::adapt
	     << endl;
	delete mink;
      }
  }

  // fit1:  Test fitting the ``square'' into a copy of itself twice as big.
  if (parms["fit1"] != "") {
    RegionPtr regions[2];
    regions[1] = new Embedding(segs);
    Transformation t(Point(0, 0), Point(2, 0));
    regions[0] = new Embedding(regions[1], t);
    regions[0]->complement();
    Point tr[2];
    fit1(regions, tr);
    drawall(regions, tr, 2);
  }

  // fit2: Test fitting two copies of the ``square'' into a copy of
  // itself three times as big.
  if (parms["fit2"] != "") {
    RegionPtr regions[3];
    regions[1] = new Embedding(segs);
    regions[2] = new Embedding(segs);
    Transformation t(Point(0, 0), Point(3, 0));
    regions[0] = new Embedding(regions[1], t);
    regions[0]->complement();
    Point tr[3];
    fit2(regions, tr);
    drawall(regions, tr, 3);
  }

  // fit2l: Fit two copies of ``square'' into smallest scaled copy of
  // the same ``square''.
  // fit2l=polyarc.txt: Fit two copies of polyarc into smallest scaled
  // copy of the same polyarc.
  if (parms["fit2l"] != "") {
    RegionPtr regions[3];
    if (parms["fit2l"] != "") {
      regions[1] = readPolyArc(parms["fit2l"].c_str());
      regions[2] = readPolyArc(parms["fit2l"].c_str());
      sprintf(convName, "%s", parms["fit2l"].c_str());
    }
    else {
      regions[1] = new Embedding(segs);
      regions[2] = new Embedding(segs);
      sprintf(convName, "%s", "segs");
    }
    double lo = 1;
    double hi = 4.9;
    double n = 0;
    double goodLo = 1, goodHi = 0;
    for (int i = 0; i < 2; i++) {
      while (true) {
	cout << "lo " << lo << " hi " << hi << " - " << hi - lo << endl;
	double scale = (lo + hi) / 2;
	if (i == 1)
	  cout << setprecision(20) << "data scale " << scale << endl;
	Transformation t(Point(0, 0), Point(scale, 0));
	RegionPtr regions2[3];
	regions2[0] = new Embedding(regions[1], t);
	regions2[0]->complement();
	regions2[1] = regions[1];
	regions2[2] = regions[2];
	Point tr[3];
	Timing::begin(Timing::fit);
	int itFit = fit2(regions2, tr);
	Timing::end(Timing::fit);
	n += 1;
	if (i == 0) {
	  cout << "data n " << n << " "
	       << "ncells " << Timing::ncells << " "
	       << "nbubbles " << Timing::nbubbles << " "
	       << "fit " << Timing::fit << " "
	       << "sum " << Timing::sum << " "
	       << "loops " << Timing::sumloops << " "
	       << "arrange " << Timing::sumarrange << " "
	       << "inside " << Timing::pointInside << " "
	       << " adapt " << Timing::adapt
	       << endl;
	  // cout << "pointInside calls " << pointInsideCall << " "
	  // << "loops " << pointInsideLoop << endl;
	  cout << "bubble_test " << Timing::timeFG << endl;
	}
	if (itFit < 0) {
	  cout << "bad fit2" << endl;
	  break;
	}
	else if (itFit == 0) {
	  cout << "success yes " << n << endl;
	  goodLo = lo;
	  goodHi = hi;
	  drawall(regions2, tr, 3);
	  if (hi == scale || i == 1)
	    break;
	  hi = scale;
	}
	else {
	  cout << "success no " << n << endl;
	  // assert(itFit == 3);
	  assert(i == 0);
	  if (lo == scale)
	    break;
	  lo = scale;
	}
      }
      extern bool checkSegment;
      checkSegment = true;
      lo = goodLo;
      hi = goodHi;
    }
  }

  // fit3: Test fitting three copies of the ``square'' into a 10 by 10
  // square.
  if (parms["fit3"] != "") {
    RegionPtr regions[4];
    regions[1] = new Embedding(segs);
    regions[2] = new Embedding(segs);
    regions[3] = new Embedding(segs);
    regions[0] = makeSquare(10);
    regions[0]->complement();
    Point tr[4];
    fit3(regions, tr);
    drawall(regions, tr, 4);
  }

  // fit3l: Fit two copies of ``square'' into smallest square.
  // fit3l=polyarc.txt: Fit two copies of polyarc into smallest square.
  if (parms["fit3l"] != "") {
    RegionPtr regions[4];
    if (parms["fit3l"] != "") {
      regions[1] = readPolyArc(parms["fit3l"].c_str());
      regions[2] = readPolyArc(parms["fit3l"].c_str());
      regions[3] = readPolyArc(parms["fit3l"].c_str());
      sprintf(convName, "%s", parms["fit3l"].c_str());
    }
    else {
      regions[1] = new Embedding(segs);
      regions[2] = new Embedding(segs);
      regions[3] = new Embedding(segs);
      sprintf(convName, "%s", "segs");
    }
    
    char buf[1024];
    sprintf(buf, "%s.fig", convName);
    regions[1]->edgedraw(buf);
    
    Point bbox[2];
    regions[1]->getBBox(bbox);
    double d = min(bbox[1].x - bbox[0].x, bbox[1].y - bbox[0].y); 
    double lo = d / 2;
    double hi = 3 * d;
    double n = 0;
    double goodLo = 1, goodHi = 0;
    for (int i = 0; i < 2; i++) {
      while (true) {
	cout << "lo " << lo << " hi " << hi << " - " << hi - lo << endl;
	double scale = (lo + hi) / 2;
	if (i == 1)
	  cout << setprecision(20) << "data scale " << scale << endl;
	Transformation t(Point(0, 0), Point(scale, 0));
	RegionPtr regions2[4];
	regions2[0] = makeSquare(scale);
	regions2[0]->complement();
	regions2[1] = regions[1];
	regions2[2] = regions[2];
	regions2[3] = regions[3];
	Point tr[4];
	Timing::begin(Timing::fit);
	int itFit = fit3(regions2, tr);
	Timing::end(Timing::fit);
	n += 1;
	if (i == 0) {
	  cout << "data n " << n << " "
	       << "ncells " << Timing::ncells << " "
	       << "nbubbles " << Timing::nbubbles << " "
	       << "fit " << Timing::fit << " "
	       << "sum " << Timing::sum << " "
	       << "loops " << Timing::sumloops << " "
	       << "arrange " << Timing::sumarrange << " "
	       << "inside " << Timing::pointInside << " "
	       << " adapt " << Timing::adapt
	       << endl;
	  // cout << "pointInside calls " << pointInsideCall << " "
	  // << "loops " << pointInsideLoop << endl;
	  cout << "bubble_test " << Timing::timeFG << endl;
	}
	if (itFit < 0) {
	  cout << "bad fit3 " << itFit << endl;
	  break;
	}
	else if (itFit == 0) {
	  cout << "success yes " << n << endl;
	  goodLo = lo;
	  goodHi = hi;
	  drawall(regions2, tr, 4);
	  if (hi == scale || i == 1)
	    break;
	  hi = scale;
	}
	else {
	  cout << "success no " << n << endl;
	  assert(i == 0);
	  cout << "failure " << itFit << endl;
	  if (lo == scale)
	    break;
	  lo = scale;
	}
      }
      extern bool checkSegment;
      checkSegment = true;
      lo = goodLo;
      hi = goodHi;
    }
  }

#ifdef NTESTS
  Embedding *ntest6 = readPolyArc("ntest6");
  ntest6->edgedraw("ntest6.fig");
#endif
}
