#include "plot.h"

Curves branches;
Points crossings;
double bbox[4], sbbox[4];
int wwidth, wheight;
bool ZoomState = false;
Point ZoomStart, ZoomEnd;

void graphics_init (int argc, char **argv, int wsize, Curves &br, Points &cr,
		    double *bb)
{
  wwidth = wheight = wsize;
  branches = br;
  crossings = cr;
  for (int i = 0; i < 4; ++i)
    sbbox[i] = bbox[i] = bb[i];
  glutInit(&argc, argv);
  glutInitWindowSize(wsize, wsize);
  glutInitDisplayMode(GLUT_SINGLE | GLUT_RGB);
  glutCreateWindow("curves");
  glutReshapeFunc(reshape);
  glutKeyboardFunc(keyboard);	    
  glutMouseFunc(mouse);
  glutMotionFunc(motion);
  glutDisplayFunc(display);
  glutMainLoop();
}

void reshape (int w, int h)
{
  glViewport(0, 0, w, h);
  wwidth = w;
  wheight = h;
}

void keyboard (unsigned char key, int x, int y)
{
  switch(key) {
  case 'r': {
    for (int i = 0; i < 4; ++i)
      bbox[i] = sbbox[i];
    break;
  }
  case 'q': 
    exit(0);
  }
  glutPostRedisplay();
}

void mouse (int button, int buttonState, int x, int y)
{
  Point p = point(x, y);
  mouse_zoom(buttonState, p);
}

Point point (int x, int y)
{
  double u = x/(double) wwidth, v = 1.0 - y/(double) wheight;
  double px = (1.0 - u)*bbox[0] + u*bbox[1];
  double py = (1.0 - v)*bbox[2] + v*bbox[3];
  return Point(px, py);
}

void mouse_zoom (int buttonState, const Point &p)
{
  if (buttonState == GLUT_DOWN) {
    ZoomStart = p;
    ZoomEnd = p;
    ZoomState = true;
  }
  else {
    const Point &q = ZoomStart;
    bbox[0] = min(p.x, q.x);
    bbox[1] = max(p.x, q.x);
    bbox[2] = min(p.y, q.y);
    bbox[3] = max(p.y, q.y);
    ZoomState = false;
  }
  glutPostRedisplay();
}

void motion (int x, int y)
{
  if (!ZoomState) return;
  ZoomEnd = point(x, y);
  glutPostRedisplay();
}

void display ()
{
  float white[3] = {1.0f, 1.0f, 1.0f};
  glPointSize(5.0);
  set_scale(bbox[0], bbox[1], bbox[2], bbox[3]);
  clear_screen(white);
  glColor3f(0.0f, 0.0f, 0.0f);
  draw_axes(bbox, bbox + 2);
  for (int i = 0; i < branches.size(); ++i)
    draw_polyline(branches[i]);
  glColor3f(1.0f, 0.0f, 0.0f);
  for (int i = 0; i < branches.size(); ++i) {
    const Points &br = branches[i];
    draw_point(br[0]);
    draw_point(*(br.end()-1));
  }
  glColor3f(0.0f, 1.0f, 0.0f);
  for (int i = 0; i < crossings.size(); ++i)
    draw_point(crossings[i]);
  if (ZoomState)
      display_zoom();
  glFlush();
}

void display_zoom ()
{
  const Point &p = ZoomStart, &q = ZoomEnd;
  glColor3f(0.0, 0.0, 1.0);
  glLineWidth(1);
  glBegin(GL_LINE_LOOP);
  glVertex2f(p.x, p.y);
  glVertex2f(p.x, q.y);
  glVertex2f(q.x, q.y);
  glVertex2f(q.x, p.y);
  glEnd();
}

void set_scale (double x1, double x2, double y1, double y2)
{
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluOrtho2D(x1, x2, y1, y2);
  glMatrixMode(GL_MODELVIEW);
}

void clear_screen (float *color)
{
  glClearColor(color[0], color[1], color[2], color[3]);
  glClear(GL_COLOR_BUFFER_BIT);
}

void draw_axes (const double *xr, const double *yr)
{
  double x0, y0;
  x0 = xr[0] < 0 && xr[1] > 0 ? 0.0 : xr[0];
  y0 = yr[0] < 0 && yr[1] > 0 ? 0.0 : yr[0];
  glBegin(GL_LINES);
  glVertex2d(xr[0], y0);
  glVertex2d(xr[1], y0);
  glVertex2d(x0, yr[0]);
  glVertex2d(x0, yr[1]);
  glEnd();
}

void draw_polyline (const Points &points)
{
  glBegin(GL_LINE_STRIP);
  for (int i = 0; i < points.size(); ++i)
    glVertex2d(points[i].x, points[i].y);
  glEnd();
}

void draw_point (const Point &p)
{
  glBegin(GL_POINTS);
  glVertex2d(p.x, p.y);
  glEnd();
}

